/*
 *                      Yeppp! library build framework
 *
 * This file is part of Yeppp! library infrastructure and licensed under 2-clause BSD license.
 * See library/LICENSE.txt for details.
 *
 */

package info.yeppp.Build;

import info.yeppp.Build.internals.*;
import info.yeppp.Build.internals.Process;
import info.yeppp.Build.sdk.Windows8Sdk;
import info.yeppp.Build.sdk.WindowsSdk;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.SystemUtils;

import java.io.*;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public abstract class VisualStudio {

	VisualStudio(WindowsSdk windowsSdk) {
		this.windowsSdk = windowsSdk;
	}

	public boolean isSupportedArchitecture(Architecture architecture) {
		switch (architecture) {
			case X86:
				return cppToolsPathX86 != null && cppLibraryPathX86 != null;
			case X64:
				return cppToolsPathX64 != null && cppLibraryPathX64 != null;
			case IA64:
				return cppToolsPathIA64 != null && cppLibraryPathIA64 != null;
			case ARM:
				return cppToolsPathArm != null && cppLibraryPathArm != null;
			default:
				return false;
		}
	}

	public String getCppToolsPath(Architecture architecture) {
		if (isSupportedArchitecture(architecture)) {
			switch (architecture) {
				case X86:
					return cppToolsPathX86;
				case X64:
					return cppToolsPathX64;
				case IA64:
					return cppToolsPathIA64;
				case ARM:
					return cppToolsPathArm;
				default:
					assert false;
					throw new IllegalArgumentException(String.format("Architecture %s is not supported in %s", architecture.toString(), getName()));
			}
		} else {
			throw new IllegalArgumentException(String.format("Architecture %s is not supported in %s", architecture.toString(), getName()));
		}
	}

	public String getCppIncludePath() {
		if (cppIncludePath != null) {
			return cppIncludePath;
		} else {
			throw new UnsupportedOperationException("Visual Studio 2010 include directory was not found on disk");
		}
	}

	public String getCppLibraryPath(Architecture architecture) {
		if (isSupportedArchitecture(architecture)) {
			switch (architecture) {
				case X86:
					return cppLibraryPathX86;
				case X64:
					return cppLibraryPathX64;
				case IA64:
					return cppLibraryPathIA64;
				case ARM:
					return cppLibraryPathArm;
				default:
					assert false;
					throw new IllegalArgumentException(String.format("Architecture %s is not supported in %s", architecture.toString(), getName()));
			}
		} else {
			throw new IllegalArgumentException(String.format("Architecture %s is not supported in %s", architecture.toString(), getName()));
		}
	}

	public String getIdePath() {
		return idePath;
	}

	public String getCommonToolsPath() {
		return commonToolsPath;
	}
	
	public abstract String getName();
	
	public abstract String getShortName();
	
	protected String lookupCppToolchain(String directoryPath, String assemblerName) {
		final String absoluteDirectoryPath = FilenameUtils.concat(installPath, directoryPath);
		File directory = new File(absoluteDirectoryPath);
		if (directory.exists() && directory.isDirectory()) {
			final File compiler = new File(FilenameUtils.concat(absoluteDirectoryPath, "cl.exe"));
			if (compiler.exists() && compiler.isFile()) {
				final File linker = new File(FilenameUtils.concat(absoluteDirectoryPath, "link.exe"));
				if (linker.exists() && linker.isFile()) {
					final File librarian = new File(FilenameUtils.concat(absoluteDirectoryPath, "lib.exe"));
					if (librarian.exists() && librarian.isFile()) {
						final File editBin = new File(FilenameUtils.concat(absoluteDirectoryPath, "editbin.exe"));
						if (editBin.exists() && editBin.isFile()) {
							final File dumpBin = new File(FilenameUtils.concat(absoluteDirectoryPath, "dumpbin.exe"));
							if (dumpBin.exists() && dumpBin.isFile()) {
								if (assemblerName != null) {
									final File assembler = new File(FilenameUtils.concat(absoluteDirectoryPath, assemblerName + ".exe"));
									if (assembler.exists() && assembler.isFile()) {
										return absoluteDirectoryPath;
									}
								} else {
									return absoluteDirectoryPath;
								}
							}
						}
					}
				}
			}
		}
		return null;
	}

	protected final WindowsSdk windowsSdk;
	protected String installPath = null;
	protected String commonToolsPath = null;
	protected String idePath = null;
	protected String cppIncludePath = null;
	protected String cppToolsPathX86 = null;
	protected String cppLibraryPathX86 = null;
	protected String cppToolsPathX64 = null;
	protected String cppLibraryPathX64 = null;
	protected String cppToolsPathIA64 = null;
	protected String cppLibraryPathIA64 = null;
	protected String cppToolsPathArm = null;
	protected String cppLibraryPathArm = null;
	protected boolean isCrossChainX64 = false;
	protected boolean isCrossChainIA64 = false;
	protected boolean isCrossChainArm = false;
}
