/*
 *                      Yeppp! library build framework
 *
 * This file is part of Yeppp! library infrastructure and licensed under 2-clause BSD license.
 * See library/LICENSE.txt for details.
 *
 */

package info.yeppp.Build.sdk;

import info.yeppp.Build.AbstractJavaCompiler;
import info.yeppp.Build.BuildMessage;
import info.yeppp.Build.BuildMessages;
import info.yeppp.Build.internals.CommandLineTool;
import info.yeppp.Build.internals.FileSystem;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.SystemUtils;

import java.io.File;
import java.io.IOException;
import java.util.*;

public abstract class JavaSdk {

	public class JavaCompiler extends AbstractJavaCompiler {

		public JavaCompiler() {
			if (SystemUtils.IS_OS_WINDOWS) {
				this.compilerPath = FilenameUtils.concat(JavaSdk.this.toolPath, "javac.exe");
			} else {
				this.compilerPath = "javac";
			}
		}

		public BuildMessages compile(String sourcePath) {
			if (sourceDirectory == null) {
				throw new IllegalStateException("Source directory was not assigned any value");
			}
			if (classDirectory == null) {
				throw new IllegalStateException("Object directory was not assigned any value");
			}
			final BuildMessages buildMessages = new BuildMessages();
			final String relativeSourcePath = FileSystem.getRelativePath(sourcePath, sourceDirectory);
			if (relativeSourcePath == null) {
				throw new IllegalArgumentException("Source path is not within source directory");
			}
			final String classPath = FilenameUtils.concat(classDirectory,
					FilenameUtils.removeExtension(relativeSourcePath) + ".class");

			createDirectories(classPath);
			FileUtils.deleteQuietly(new File(classPath));
			try {
				final List<String> commandLine = composeCommandLine(sourcePath);

				final CommandLineTool compiler = new CommandLineTool(commandLine);
				for (String line : compiler.getError()) {
					System.err.println(line);
				}
				buildMessages.add(parseCommandLineOutput(compiler.getOutput()));
			} catch (IOException exception) {
				buildMessages.add(new BuildMessage(BuildMessage.Type.Exception, relativeSourcePath, null, null, exception.getMessage()));
			} catch (Exception e) {
				e.printStackTrace();
			}
			return buildMessages;
		}

		protected List<String> composeCommandLine(String sourcePath) {
			final List<String> commandLine = new ArrayList<String>();
			commandLine.add(compilerPath);
			if (debugEnabled) {
				commandLine.add("-g");
			} else {
				commandLine.add("-g:none");
			}
			if (!classPathDirectories.isEmpty()) {
				commandLine.add("-cp");
				commandLine.add(StringUtils.join(classPathDirectories, SystemUtils.PATH_SEPARATOR));
			}
			commandLine.add("-d");
			commandLine.add(classDirectory);
			commandLine.add(sourcePath);
			return commandLine;
		}

		protected BuildMessages parseCommandLineOutput(List<String> lines) {
			BuildMessages buildMessages = new BuildMessages();
			for (String line : lines) {
				final BuildMessage lastBuildMessage = buildMessages.getLastBuildMessage();
				if (lastBuildMessage == null) {
					buildMessages.add(new BuildMessage(BuildMessage.Type.Error, null, null, null, line));
				} else {
					lastBuildMessage.appendMessage(SystemUtils.LINE_SEPARATOR + line);
				}
			}
			return buildMessages;
		}

		private void createDirectories(String path) {
			File file = new File(path);
			if (!file.isDirectory()) {
				file = file.getParentFile();
			}
			file.mkdirs();
		}


	}

	public String getToolsPath() {
		return toolPath;
	}

	public String getJniIncludePath() {
		if (jniIncludePath != null) {
			return jniIncludePath;
		} else {
			throw new UnsupportedOperationException("JNI include directory was not found on disk");
		}
	}

	public String getJniPlatformIncludePath() {
		if (jniPlatformIncludePath != null) {
			return jniPlatformIncludePath;
		} else {
			throw new UnsupportedOperationException("JNI platform-specific include directory was not found on disk");
		}
	}

	public abstract String getName();

	protected String installPath = null;
	protected String jniIncludePath = null;
	protected String jniPlatformIncludePath = null;
	protected String toolPath = null;

}
