/*
 *                          Yeppp! library header
 *
 * This file is part of Yeppp! library and licensed under 2-clause BSD license.
 *
 * Copyright (C) 2010-2012 Marat Dukhan
 * Copyright (C) 2012-2013 Georgia Institute of Technology
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are those
 * of the authors and should not be interpreted as representing official policies,
 * either expressed or implied, of the Yeppp! Project.
 */

 
 #pragma once

#ifndef __cplusplus
	#error "This intrinsics should only be used in C++ code"
#endif

#include <yepPredefines.h>
#include <yepTypes.h>
#include <math.h>

#if defined(YEP_MICROSOFT_COMPILER)
#include <intrin.h>
#endif

static YEP_INLINE YepSize yepBuiltin_GetPointerMisalignment(const void* pointer, YepSize alignment) {
	return YepSize(pointer) % alignment;
}

static YEP_INLINE Yep64u yepBuiltin_CombineParts_32u32u_64u(Yep32u hi, Yep32u lo) {
	return (Yep64u(hi) << 32) | Yep64u(lo);
}

static YEP_INLINE Yep64u yepBuiltin_GetHighPart_64u_32u(Yep64u n) {
	return Yep32u(n >> 32);
}

static YEP_INLINE Yep64u yepBuiltin_GetLowPart_64u_32u(Yep64u n) {
	return Yep32u(n);
}

static YEP_INLINE Yep16u yepBuiltin_ByteSwap_16u_16u(Yep16u n) {
#if defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_LINUX)
	/* Contrary to GCC documentation, this intrinsic is not supported in gcc/clang/icc */
	/* return __builtin_bswap16(n); */
	return Yep16u(__builtin_bswap32(n << 16));
#elif defined(YEP_MICROSOFT_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_WINDOWS)
	return _byteswap_ushort(n);
#else
	return Yep16u(n >> 8) | Yep16u(n << 8);
#endif
}

static YEP_INLINE Yep32u yepBuiltin_ByteSwap_32u_32u(Yep32u n) {
#if defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_LINUX)
	return __builtin_bswap32(n);
#elif defined(YEP_MICROSOFT_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_WINDOWS)
	return _byteswap_ulong(n);
#else
	return (n >> 24) | ((n >> 8) & 0x0000FF00u) | ((n << 8) & 0x00FF0000u) | (n << 24);
#endif
}

static YEP_INLINE Yep64u yepBuiltin_ByteSwap_64u_64u(Yep64u n) {
#if defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_LINUX)
	return __builtin_bswap64(n);
#elif defined(YEP_MICROSOFT_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_WINDOWS)
	return _byteswap_uint64(n);
#else
	const Yep32u nLo = yepBuiltin_GetLowPart_64u_32u(n);
	const Yep32u nHi = yepBuiltin_GetHighPart_64u_32u(n);
	const Yep32u nLoSwapped = yepBuiltin_ByteSwap_32u_32u(nLo);
	const Yep32u nHiSwapped = yepBuiltin_ByteSwap_32u_32u(nHi);
	return yepBuiltin_CombineParts_32u32u_64u(nLoSwapped, nHiSwapped);
#endif
}

static YEP_INLINE Yep32f yepBuiltin_Abs_32f_32f(Yep32f x) {
#if defined(YEP_MICROSOFT_COMPILER)
	return abs(x);
#elif defined(YEP_INTEL_COMPILER)
	return fabsf(x);
#elif defined(YEP_ARM_COMPILER)
	return __fabsf(x);
#elif defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER)
	return __builtin_fabsf(x);
#else
	return x >= 0.0f ? x : -x;
#endif
}

static YEP_INLINE Yep64f yepBuiltin_Abs_64f_64f(Yep64f x) {
#if defined(YEP_MICROSOFT_COMPILER)
	return abs(x);
#elif defined(YEP_INTEL_COMPILER)
	return fabs(x);
#elif defined(YEP_ARM_COMPILER)
	return __fabs(x);
#elif defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER)
	return __builtin_fabs(x);
#else
	return x >= 0.0 ? x : -x;
#endif
}

static YEP_INLINE Yep32f yepBuiltin_Convert_8s_32f(Yep8s number) {
	return Yep32f(Yep32s(number));
}

static YEP_INLINE Yep32f yepBuiltin_Convert_16s_32f(Yep16s number) {
	return Yep32f(Yep32s(number));
}

static YEP_INLINE Yep32f yepBuiltin_Convert_32s_32f(Yep32s number) {
	return Yep32f(number);
}

static YEP_INLINE Yep32f yepBuiltin_Convert_64s_32f(Yep64s number) {
	return Yep32f(number);
}

static YEP_INLINE Yep64f yepBuiltin_Convert_8s_64f(Yep8s number) {
	return Yep64f(Yep32s(number));
}

static YEP_INLINE Yep64f yepBuiltin_Convert_16s_64f(Yep16s number) {
	return Yep64f(Yep32s(number));
}

static YEP_INLINE Yep64f yepBuiltin_Convert_32s_64f(Yep32s number) {
	return Yep64f(number);
}

static YEP_INLINE Yep64f yepBuiltin_Convert_64s_64f(Yep64s number) {
	return Yep64f(number);
}

static YEP_INLINE Yep32f yepBuiltin_Convert_8u_32f(Yep8u number) {
	return Yep32f(Yep32s(Yep32u(number)));
}

static YEP_INLINE Yep32f yepBuiltin_Convert_16u_32f(Yep16u number) {
	return Yep32f(Yep32s(Yep32u(number)));
}

static YEP_INLINE Yep32f yepBuiltin_Convert_32u_32f(Yep32u number) {
	return Yep32f(Yep64s(Yep64u(number)));
}

static YEP_INLINE Yep32f yepBuiltin_Convert_64u_32f(Yep64u number) {
	return Yep32f(Yep64s(number & 0x7FFFFFFFFFFFFFFFull)) - Yep32f(Yep64s(number & 0x8000000000000000ull));
}

static YEP_INLINE Yep64f yepBuiltin_Convert_8u_64f(Yep8u number) {
	return Yep64f(Yep32s(Yep32u(number)));
}

static YEP_INLINE Yep64f yepBuiltin_Convert_16u_64f(Yep16u number) {
	return Yep64f(Yep32s(Yep32u(number)));
}

static YEP_INLINE Yep64f yepBuiltin_Convert_32u_64f(Yep32u number) {
	return Yep64f(Yep64s(Yep64u(number)));
}

static YEP_INLINE Yep64f yepBuiltin_Convert_64u_64f(Yep64u number) {
	return Yep64f(Yep64s(number & 0x7FFFFFFFFFFFFFFFull)) - Yep64f(Yep64s(number & 0x8000000000000000ull));
}

static YEP_INLINE Yep8u yepBuiltin_Min_8u8u_8u(Yep8u a, Yep8u b) {
	return (a > b) ? b : a;
}

static YEP_INLINE Yep8s yepBuiltin_Min_8s8s_8s(Yep8s a, Yep8s b) {
	return (a > b) ? b : a;
}

static YEP_INLINE Yep16u yepBuiltin_Min_16u16u_16u(Yep16u a, Yep16u b) {
	return (a > b) ? b : a;
}

static YEP_INLINE Yep16s yepBuiltin_Min_16s16s_16s(Yep16s a, Yep16s b) {
	return (a > b) ? b : a;
}

static YEP_INLINE Yep32u yepBuiltin_Min_32u32u_32u(Yep32u a, Yep32u b) {
	return (a > b) ? b : a;
}

static YEP_INLINE Yep32s yepBuiltin_Min_32s32s_32s(Yep32s a, Yep32s b) {
	return (a > b) ? b : a;
}

static YEP_INLINE Yep64u yepBuiltin_Min_64u64u_64u(Yep64u a, Yep64u b) {
	return (a > b) ? b : a;
}

static YEP_INLINE Yep64s yepBuiltin_Min_64s64s_64s(Yep64s a, Yep64s b) {
	return (a > b) ? b : a;
}

static YEP_INLINE Yep32f yepBuiltin_Min_32f32f_32f(Yep32f a, Yep32f b) {
#if defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER) || defined(YEP_INTEL_COMPILER)
	return fminf(a, b);
#else
	if (a != a) {
		return b;
	} else if (b != b) {
		return a;
	} else {
		return (a > b) ? b : a;
	}
#endif
}

static YEP_INLINE Yep64f yepBuiltin_Min_64f64f_64f(Yep64f a, Yep64f b) {
#if defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER) || defined(YEP_INTEL_COMPILER)
	return fmin(a, b);
#else
	if (a != a) {
		return b;
	} else if (b != b) {
		return a;
	} else {
		return (a > b) ? b : a;
	}
#endif
}

static YEP_INLINE Yep8u yepBuiltin_Max_8u8u_8u(Yep8u a, Yep8u b) {
	return (a > b) ? a : b;
}

static YEP_INLINE Yep8s yepBuiltin_Max_8s8s_8s(Yep8s a, Yep8s b) {
	return (a > b) ? a : b;
}

static YEP_INLINE Yep16u yepBuiltin_Max_16u16u_16u(Yep16u a, Yep16u b) {
	return (a > b) ? a : b;
}

static YEP_INLINE Yep16s yepBuiltin_Max_16s16s_16s(Yep16s a, Yep16s b) {
	return (a > b) ? a : b;
}

static YEP_INLINE Yep32u yepBuiltin_Max_32u32u_32u(Yep32u a, Yep32u b) {
	return (a > b) ? a : b;
}

static YEP_INLINE Yep32s yepBuiltin_Max_32s32s_32s(Yep32s a, Yep32s b) {
	return (a > b) ? a : b;
}

static YEP_INLINE Yep64u yepBuiltin_Max_64u64u_64u(Yep64u a, Yep64u b) {
	return (a > b) ? a : b;
}

static YEP_INLINE Yep64s yepBuiltin_Max_64s64s_64s(Yep64s a, Yep64s b) {
	return (a > b) ? a : b;
}

static YEP_INLINE Yep32f yepBuiltin_Max_32f32f_32f(Yep32f a, Yep32f b) {
#if defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER) || defined(YEP_INTEL_COMPILER)
	return fmaxf(a, b);
#else
	if (a != a) {
		return b;
	} else if (b != b) {
		return a;
	} else {
		return (a > b) ? a : b;
	}
#endif
}

static YEP_INLINE Yep64f yepBuiltin_Max_64f64f_64f(Yep64f a, Yep64f b) {
#if defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER) || defined(YEP_INTEL_COMPILER)
	return fmax(a, b);
#else
	if (a != a) {
		return b;
	} else if (b != b) {
		return a;
	} else {
		return (a > b) ? a : b;
	}
#endif
}

static YEP_INLINE YepBoolean yepBuiltin_IsNaN_64f(Yep64f n) {
	return !(n == n);
}

static YEP_INLINE YepBoolean yepBuiltin_IsNaN_32f(Yep32f n) {
	return !(n == n);
}

static YEP_INLINE Yep32f yepBuiltin_PositiveInfinity_32f() {
#if defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER)
	return __builtin_inff();
#else
	const static Yep64f one = 1.0f;
	const static Yep64f zero = 0.0f;
	return one / zero;
#endif
}

static YEP_INLINE Yep64f yepBuiltin_PositiveInfinity_64f() {
#if defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER)
	return __builtin_inf();
#else
	const static Yep64f one = 1.0;
	const static Yep64f zero = 0.0;
	return one / zero;
#endif
}

static YEP_INLINE Yep32f yepBuiltin_NegativeInfinity_32f() {
	return -yepBuiltin_PositiveInfinity_32f();
}

static YEP_INLINE Yep64f yepBuiltin_NegativeInfinity_64f() {
	return -yepBuiltin_PositiveInfinity_64f();
}

static YEP_INLINE Yep32f yepBuiltin_NaN_32f() {
#if defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER)
	return __builtin_nanf("");
#else
	const static Yep32f zero = 0.0f;
	return zero / zero;
#endif
}

static YEP_INLINE Yep64f yepBuiltin_NaN_64f() {
#if defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER)
	return __builtin_nan("");
#else
	const static Yep64f zero = 0.0;
	return zero / zero;
#endif
}

static YEP_INLINE Yep64u yepBuiltin_Cast_64f_64u(Yep64f x) {
	union {
		Yep64f float64;
		Yep64u word64;
	} float64_word64;
	float64_word64.float64 = x;
	return float64_word64.word64;
}

static YEP_INLINE Yep64f yepBuiltin_Cast_64u_64f(Yep64u x) {
	union {
		Yep64f float64;
		Yep64u word64;
	} float64_word64;
	float64_word64.word64 = x;
	return float64_word64.float64;
}

static YEP_INLINE Yep32u yepBuiltin_Nlz_64u_32u(Yep64u x) {
#if defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER)
	return __builtin_clzl(x);
#elif (defined(YEP_MICROSOFT_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_WINDOWS)) && (defined(YEP_IA64_CPU) || defined(YEP_X64_CPU))
	if (x == 0ull) {
		return 64u;
	} else {
		unsigned long bitPosition;
		_BitScanReverse64(&bitPosition, x);
		return 63u - bitPosition;
	}
#elif (defined(YEP_MICROSOFT_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_WINDOWS)) && defined(YEP_X86_CPU)
	const Yep32u xHi = yepBuiltin_GetHighPart_64u_32u(x);
	const Yep32u xLo = yepBuiltin_GetLowPart_64u_32u(x);
	unsigned long bitPositionHi, bitPositionLo;
	_BitScanReverse(&bitPositionLo, xLo);
	_BitScanReverse(&bitPositionHi, xHi);
	if YEP_UNLIKELY(xHi == 0u) {
		if YEP_UNLIKELY(xLo == 0u) {
			return 64u;
		} else {
			return 63u - bitPositionLo;
		}
	} else {
		return 31u - bitPositionHi;
	}
#endif
}

//~ static YEP_INLINE Yep32f yepBuiltin_Log_32f_32f(Yep32f x) {
	//~ return logf(x);
//~ }

static YEP_INLINE Yep64f yepBuiltin_Log_64f_64f(Yep64f x) {
	const Yep64u defaultExponent = 0x3FF0000000000000ull;
#if defined(YEP_COMPILER_SUPPORTS_HEXADECIMAL_FLOATING_POINT_CONSTANTS)
	const Yep64f sqrt2 = 0x1.6A09E667F3BCDp0;
	const Yep64f c2  = -0x1.0000000000004p-1;
	const Yep64f c3  =  0x1.5555555555524p-2;
	const Yep64f c4  = -0x1.FFFFFFFFFD99Cp-3;
	const Yep64f c5  =  0x1.99999999A037Bp-3;
	const Yep64f c6  = -0x1.5555555707EC0p-3;
	const Yep64f c7  =  0x1.24924922E0177p-3;
	const Yep64f c8  = -0x1.FFFFFEE6E9A74p-4;
	const Yep64f c9  =  0x1.C71C7190489D4p-4;
	const Yep64f c10 = -0x1.9999CAC62E0ADp-4;
	const Yep64f c11 =  0x1.745D4E8088327p-4;
	const Yep64f c12 = -0x1.55505B48AD061p-4;
	const Yep64f c13 =  0x1.3B0A137482D51p-4;
	const Yep64f c14 = -0x1.24DCC4EF07DB3p-4;
	const Yep64f c15 =  0x1.11DFFAEDD082Ep-4;
	const Yep64f c16 = -0x1.FB780993249D2p-5;
	const Yep64f c17 =  0x1.CF5EFF4973CFAp-5;
	const Yep64f c18 = -0x1.D0CACFA711D8Cp-5;
	const Yep64f c19 =  0x1.0BA1125E4CABDp-4;
	const Yep64f c20 = -0x1.ED49EDACFD1C0p-5;
	const Yep64f c21 =  0x1.AE7E3413896B4p-6;
	const Yep64f ln2_lo = 0x1.EF35793C76730p-45;
	const Yep64f ln2_hi = 0x1.62E42FEFA3800p-1;
#else
	const Yep64f sqrt2 = 1.4142135623730951;
	const Yep64f c2  = -0.5000000000000004;
	const Yep64f c3  =  0.3333333333333306;
	const Yep64f c4  = -0.24999999999972722;
	const Yep64f c5  =  0.20000000000075233;
	const Yep64f c6  = -0.16666666671607366;
	const Yep64f c7  =  0.1428571428077807;
	const Yep64f c8  = -0.12499999590964156;
	const Yep64f c9  =  0.11111111031326654;
	const Yep64f c10 = -0.10000018318794686;
	const Yep64f c11 =  0.09090929665445611;
	const Yep64f c12 = -0.08332858712930281;
	const Yep64f c13 =  0.07691390609413885;
	const Yep64f c14 = -0.07149960448486219;
	const Yep64f c15 =  0.06686399478078961;
	const Yep64f c16 = -0.06194688673089711;
	const Yep64f c17 =  0.0565638528893206;
	const Yep64f c18 = -0.05673733288909508;
	const Yep64f c19 =  0.06533915686740506;
	const Yep64f c20 = -0.06021591588060771;
	const Yep64f c21 =  0.026275206428315115;
	const Yep64f ln2_lo = 5.497923018708371e-14;
	const Yep64f ln2_hi = 0.6931471805598903;
#endif
	if YEP_LIKELY(x == x) {
		const Yep64s xWord = yepBuiltin_Cast_64f_64u(x);
		if YEP_UNLIKELY(xWord < 0ll) {
			// sign(x) == -1
			return yepBuiltin_NaN_64f();
		} else if YEP_UNLIKELY(xWord == 0ll) {
			// x == +0.0
			return yepBuiltin_NegativeInfinity_64f();
		} else if YEP_UNLIKELY(xWord == 0x7FF0000000000000ull) {
			// x == +inf
			return x;
		}
		Yep32s exponent;
		Yep64u mantissa;
		if YEP_LIKELY(xWord >= 0x0010000000000000ull) {
			// Normalized number
			exponent = Yep32s((yepBuiltin_GetHighPart_64u_32u(Yep64u(xWord)) >> 20) & 0x7FFu) - 1023;
			mantissa = xWord & 0x000FFFFFFFFFFFFFull;
		} else {
			// Denormalized number
			const Yep32u pointOffset = yepBuiltin_Nlz_64u_32u(xWord) - 11u;
			exponent = -1022 - pointOffset;
			mantissa = (xWord << pointOffset) & 0x000FFFFFFFFFFFFFull;
		}
		x = yepBuiltin_Cast_64u_64f(defaultExponent | mantissa);
		if (x >= sqrt2) {
			exponent += 1;
			x = x * 0.5;
		}
		const Yep64f t = x - 1.0;
		const Yep64f dexp = Yep64f(exponent);
		return (t + t * (t * (c2 + t * (c3 + t * (c4 + t * (c5 + t * (c6 + t * (c7 + t * (c8 + t * (c9 + t * (c10 + t * (c11 + t * (c12 + t * (c13 + t * (c14 + t * (c15 + t * (c16 + t * (c17 + t * (c18 + t * (c19 + t * (c20 + t * c21)))))))))))))))))))) +
			dexp * ln2_lo) + dexp * ln2_hi;
	} else {
		// x is NaN
		return x;
	}
}

//~ static YEP_INLINE Yep32f yepBuiltin_Exp_32f_32f(Yep32f x) {
	//~ return expf(x);
//~ }

static YEP_INLINE Yep64f yepBuiltin_Exp_64f_64f(Yep64f x) {
#if defined(YEP_COMPILER_SUPPORTS_HEXADECIMAL_FLOATING_POINT_CONSTANTS)
	const Yep64f magicBias = 0x1.00000000003FFp+52;
	const Yep64f specialMagicBias = 0x1.00000000007FEp+52;
	const Yep64f zeroCutoff = -0x1.74385446D71C3p+9;
	const Yep64f infCutoff = 0x1.62E42FEFA39EFp+9;
	const Yep64f log2e = 0x1.71547652B82FEp+0;
	const Yep64f ln2_hi = 0x1.62E42FEFA3800p-1; // The lowest 10 bits are zeros.
	const Yep64f ln2_lo = 0x1.EF35793C76730p-45;
	
	const Yep64f c2 = 0x1.000000000000Cp-1;
	const Yep64f c3 = 0x1.5555555555522p-3;
	const Yep64f c4 = 0x1.555555554FE5Cp-5;
	const Yep64f c5 = 0x1.111111111DC0Ap-7;
	const Yep64f c6 = 0x1.6C16C185EC9E1p-10;
	const Yep64f c7 = 0x1.A01A016020E85p-13;
	const Yep64f c8 = 0x1.A019960BCB82Ap-16;
	const Yep64f c9 = 0x1.71DEADCE6F1F1p-19;
	const Yep64f c10 = 0x1.28B0A7EA5C532p-22;
	const Yep64f c11 = 0x1.AE3BB8191C002p-26;
#else
	const Yep64f magicBias = 4503599627371519.0;
	const Yep64f specialMagicBias = 4503599627372542.0;
	const Yep64f zeroCutoff = -744.4400719213812;
	const Yep64f infCutoff = 709.782712893384;
	const Yep64f log2e = 1.4426950408889634;
	const Yep64f ln2_hi = 0.6931471805598903; // The lowest 10 bits are zeros.
	const Yep64f ln2_lo = 5.497923018708371e-14;
	
	const Yep64f c2 = 0.5000000000000013;
	const Yep64f c3 = 0.16666666666666524;
	const Yep64f c4 = 0.04166666666651217;
	const Yep64f c5 = 0.008333333333423471;
	const Yep64f c6 = 0.0013888888947631783;
	const Yep64f c7 = 0.0001984126965944752;
	const Yep64f c8 = 2.4801489458763235e-05;
	const Yep64f c9 = 2.7557450501243553e-06;
	const Yep64f c10 = 2.7631415257859634e-07;
	const Yep64f c11 = 2.5042872751587012e-08;
#endif
	
	const Yep32s x_hi = yepBuiltin_GetHighPart_64u_32u(yepBuiltin_Cast_64f_64u(x));
	if YEP_LIKELY(x == x) {
		Yep64f t = x * log2e + specialMagicBias;
		Yep32u e1 = yepBuiltin_GetLowPart_64u_32u(yepBuiltin_Cast_64f_64u(t)) << 20;
		Yep32u e2 = e1;
		e1 = yepBuiltin_Max_32u32u_32u(e1, 0x80100000u);
		e1 = yepBuiltin_Min_32u32u_32u(e1, 0xFFD00000u);
		e2 -= e1;
		e1 -= 0x7FF00000;
		const Yep64f s1 = yepBuiltin_Cast_64u_64f(yepBuiltin_CombineParts_32u32u_64u(e1, 0u));
		const Yep64f s2 = yepBuiltin_Cast_64u_64f(yepBuiltin_CombineParts_32u32u_64u(e2, 0u));
		t -= specialMagicBias;
		const Yep64f rx = (x - t * ln2_hi) - t * ln2_lo;
		const Yep64f rf = rx  + rx * rx * (c2 + rx * (c3 + rx * (c4 + rx * (c5 + rx * (c6 + rx * (c7 + rx * (c8 + rx * (c9 + rx * (c10 + rx * c11)))))))));
		Yep64f f = s2 * (s1 * rf + s1);
		if YEP_UNLIKELY(x > infCutoff) {
			f = yepBuiltin_PositiveInfinity_64f();
		}
		if YEP_UNLIKELY(x < zeroCutoff) {
			f = 0.0;
		}
		return f;
	} else {
		// x is NaN
		return x;
	}
}

//~ static YEP_INLINE Yep32f yepBuiltin_Sqrt_32f_32f(Yep32f x) {
	//~ return sqrtf(x);
//~ }

#if defined(YEP_MICROSOFT_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_WINDOWS)
	#pragma intrinsic(sqrt)
#endif

static YEP_INLINE Yep64f yepBuiltin_Sqrt_64f_64f(Yep64f x) {
#if defined(YEP_MICROSOFT_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_WINDOWS)
	return sqrt(x);
#else
	return __builtin_sqrt(x);
#endif
}

//~ static YEP_INLINE Yep32f yepBuiltin_Power_32f32f_32f(Yep32f x, Yep32f y) {
	//~ return powf(x, y);
//~ }

//~ static YEP_INLINE Yep64f yepBuiltin_Power_64f64f_64f(Yep64f x, Yep64f y) {
	//~ return pow(x, y);
//~ }

static YEP_INLINE Yep64u yepBuiltin_Multiply_32u32u_64u(Yep32u x, Yep32u y) {
#if defined(YEP_MICROSOFT_COMPILER) && defined(YEP_X86_CPU)
	return __emulu(x, y);
#else
	return Yep64u(x) * Yep64u(y);
#endif
}

static YEP_INLINE Yep64s yepBuiltin_Multiply_32s32s_64s(Yep32s x, Yep32s y) {
#if defined(YEP_MICROSOFT_COMPILER) && defined(YEP_X86_CPU)
	return __emul(x, y);
#else
	return Yep64u(x) * Yep64u(y);
#endif
}

static YEP_INLINE Yep32u yepBuiltin_MultiplyHigh_32u32u_32u(Yep32u x, Yep32u y) {
#if defined(YEP_MICROSOFT_COMPILER) && defined(YEP_X86_CPU)
	return Yep32u(__emulu(x, y) >> 32);
#else
	return Yep32u(Yep64u(x) * Yep64u(y) >> 32);
#endif
}

static YEP_INLINE Yep32s yepBuiltin_MultiplyHigh_32s32s_32s(Yep32s x, Yep32s y) {
#if defined(YEP_MICROSOFT_COMPILER) && defined(YEP_X86_CPU)
	return Yep32s(Yep32u(Yep64u(__emul(x, y)) >> 32));
#else
	return Yep32s(Yep32u(Yep64u(Yep64s(x) * Yep64s(y)) >> 32));
#endif
}

static YEP_INLINE Yep32u yepBuiltin_AtomicCAS_32u_32u32u(Yep32u newValue, Yep32u lastValue, Yep32u volatile *currentValue) {
#if defined(YEP_MICROSOFT_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_WINDOWS)
	return static_cast<Yep32u>(_InterlockedCompareExchange(reinterpret_cast<long volatile*>(currentValue), static_cast<long>(newValue), static_cast<long>(lastValue)));
#elif defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_LINUX)
	return __sync_val_compare_and_swap(currentValue, lastValue, newValue);
#else
	#error "Unsupported compiler"
#endif
}

static YEP_INLINE Yep32u yepBuiltin_AtomicCAS_64u_64u64u(Yep64u newValue, Yep64u lastValue, Yep64u volatile *currentValue) {
#if defined(YEP_MICROSOFT_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_WINDOWS)
	return static_cast<Yep64u>(_InterlockedCompareExchange64(reinterpret_cast<__int64 volatile*>(currentValue), static_cast<__int64>(newValue), static_cast<__int64>(lastValue)));
#elif defined(YEP_GNU_COMPILER) || defined(YEP_CLANG_COMPILER) || defined(YEP_INTEL_COMPILER_FOR_LINUX)
	return __sync_val_compare_and_swap(currentValue, lastValue, newValue);
#else
	#error "Unsupported compiler"
#endif
}



