#
#       Portable Efficient Assembly Code-generator in Higher-level Python (PEACH-Py)
#
# This file is part of Yeppp! library infrastructure and licensed under 2-clause BSD license.
# See codegen/LICENSE.txt for details.
#

__author__ = 'Marat'

import yeppp.codegen.codegen
import yeppp.codegen.c
import copy

current_function = None
supported_isa_extensions = ['CMOV', 'MMX', 'MMX+', '3dnow!', '3dnow!+', '3dnow! Prefetch', '3dnow! Geode',
							'SSE', 'SSE2', 'SSE3', 'SSSE3', 'SSE4A', 'SSE4.1', 'SSE4.2',
							'AVX', 'AVX2', 'FMA3', 'FMA4', 'F16C',
							'MOVBE', 'POPCNT', 'LZCNT', 'TZCNT', 'BMI', 'BMI2', 'TBM',
							'AES']

class Parameter(yeppp.codegen.c.Parameter):
	def __init__(self, parameter):
		if isinstance(parameter, yeppp.codegen.c.Parameter):
			super(Parameter, self).__init__(parameter.name, parameter.type)
			self.register_location = None
			self.stack_location = None
			self.save_on_stack = False
		else:
			raise TypeError("Invalid parameter type {0}".format(type(parameter)))

class Microarchitecture(object):
	supported_microarchitectures = ['Prescott', 'Conroe', 'Penryn', 'Bonnell',
	                                'Nehalem', 'SandyBridge', 'Saltwell', 'IvyBridge', 'Haswell', 'KnightsFerry', 'KnightsCorner',
	                                'K8', 'K10', 'Bobcat', 'Bulldozer', 'Piledriver', 'Jaguar', 'Steamroller']

	abbreviations = {'Prescott': 'P4P', 'Conroe': 'CNR', 'Penryn': 'PNR', 'Bonnell': 'BNL', 'Nehalem': 'NHM', 'SandyBridge': 'SNB',
	                 'Saltwell': 'SLW', 'IvyBridge': 'IVB', 'Haswell': 'HSW', 'KnightsFerry': 'KNF', 'KnightsCorner': 'KNC',
	                 'K8': 'K8', 'K10': 'K10', 'Bobcat': 'BCT', 'Bulldozer': 'BLD', 'Piledriver': 'PLD', 'Jaguar': 'JAG', 'Steamroller': 'STR'}

	descriptions = {'Prescott': 'Intel Prescott', 'Conroe': 'Intel Conroe', 'Penryn': 'Intel Penryn', 'Bonnell': 'Intel Bonnell',
	                'Nehalem': 'Intel Nehalem', 'SandyBridge': 'Intel Sandy Bridge', 'Saltwell': 'Intel Saltwell',
	                'IvyBridge': 'Intel Ivy Bridge', 'Haswell': 'Intel Haswell', 'KnightsFerry': 'Intel Knights Ferry',
	                'KnightsCorner': 'Intel Knights Corner', 'K8': 'AMD K8', 'K10': 'AMD K10', 'Bobcat': 'AMD Bobcat',
	                'Bulldozer': 'AMD Bulldozer', 'Piledriver': 'AMD Piledriver', 'Jaguar': 'AMD Jaguar',
	                'Steamroller': 'AMD Steamroller'}

	def __init__(self, name):
		super(Microarchitecture, self).__init__()
		if name in Microarchitecture.supported_microarchitectures:
			self.name = name
		else:
			raise ValueError('Unsupported microarchitecture {0}: only ({1}) microarchitectures are supported on this architecture'.format(name, ', '.join(Microarchitecture.supported_microarchitectures)))

	def __str__(self):
		return Microarchitecture.descriptions[self.name]

	def get_name(self):
		return self.name

	def get_number(self):
		return Microarchitecture.supported_microarchitectures.index(self.get_name())

	def get_yeppp_name(self):
		return "YepCpuMicroarchitecture" + self.name

class Assembler(yeppp.codegen.codegen.CodeGenerator):
	def __init__(self, abi):
		super(Assembler, self).__init__()
		if isinstance(abi, yeppp.codegen.c.ABI):
			if abi.get_name() in ["x64-ms", "x64-sysv", "x32"]:
				self.abi = abi
			else:
				raise ValueError('Unsupported abi {0}: only "x64-ms", "x64-sysv" and "x32" ABIs are supported on this architecture'.format(abi))
		else:
			raise TypeError('Wrong type of ABI object')
		self.functions = list()

	def begin_function(self, name, arguments, microarchitecture):
		global current_function
		if current_function is not None:
			raise ValueError('Function {0} was not finalized'.format(current_function.name))
		arguments = [yeppp.codegen.x64.Parameter(argument) for argument in arguments]
		current_function = Function(name, arguments, Microarchitecture(microarchitecture), self.abi)

	def end_function(self):
		global current_function
		if current_function is None:
			raise ValueError('Trying to finalize a function while no function is active')
		function = current_function
		current_function = None

		function.generate_labels()
		function.determine_available_registers()
		function.determine_live_registers()
		function.determine_register_relations()

		# for instruction in function.instructions:
		# 	if isinstance(instruction, Instruction):
		# 		consumed_registers = ", ".join(sorted(map(str, list(instruction.get_input_registers_list()))))
		# 		produced_registers = ", ".join(sorted(map(str, list(instruction.get_output_registers_list()))))
		# 		available_registers = ", ".join(sorted(map(str, list(instruction.available_registers))))
		# 		live_registers = ", ".join(sorted(map(str, list(instruction.live_registers))))
		# 		print "{0:70s} ; {1}".format(instruction, live_registers)
		# 	else:
		# 		print str(instruction)

		function.allocate_registers()
		function.determine_live_registers()
		function.update_stack_frame()
#		function.determine_stack_offsets(0, 0)
		function.generate_prolog_and_epilog()
		function.generate_parameter_loads()
		function.generate_constant_loads()
		self.functions.append(function)

		instructions = function.instructions
		constants = function.constants

		function_label = function.abi.function_prefix + function.symbol_name
		constants_label = function.symbol_name + "_constants"
		if len(constants) > 0:
			if self.abi.name == 'x64-ms':
				self.add_line('section .rdata${0} rdata align={1}'.format(function.microarchitecture.get_number(), function.constants[0].capacity))
			else:
				self.add_line('section .rodata.{0} progbits alloc noexec nowrite align={1}'.format(function.microarchitecture.get_name(), function.constants[0].capacity))
			self.add_line(constants_label + ':')
			data_declaration_map = {8: "DB", 16: "DW", 32: "DD", 64: "DQ", 128: "DO"}
			self.indent()
			need_alignment = False
			for constant_bucket in constants:
				if need_alignment:
					self.add_line("ALIGN {0}".format(constant_bucket.capacity))
				for constant in constant_bucket.constants:
					self.add_line(".{0} {1} {2}".format(constant.label, data_declaration_map[constant.size], ", ".join([str(constant)] * constant.repeats)))
				need_alignment = not constant_bucket.is_full()
			self.dedent()
			self.add_line()
			self.add_line()

		if self.abi.name == 'x64-ms':
			self.add_line('section .text${0} code align=16'.format(function.microarchitecture.get_number()))
		else:
			self.add_line('section .text.{0} progbits alloc exec nowrite align=16'.format(function.microarchitecture.get_name()))
		self.add_line("global " + function_label)
		self.add_line(function_label + ':')
		self.indent()
		for instruction in instructions:
			if isinstance(instruction, Instruction):
				constant = instruction.get_constant()
				if constant is not None:
					constant.prefix = constants_label
				consumed_registers = ", ".join(sorted(map(str, list(instruction.get_input_registers_list()))))
				produced_registers = ", ".join(sorted(map(str, list(instruction.get_output_registers_list()))))
				available_registers = ", ".join(sorted(map(str, list(instruction.available_registers))))
				live_registers = ", ".join(sorted(map(str, list(instruction.live_registers))))
				if hasattr(instruction, "stack_offset"):
					stack_offset = instruction.stack_offset
				else:
					stack_offset = 0
				self.add_line("{0:70s} ; {1}".format(instruction, live_registers))
			elif isinstance(instruction, Label):
				self.add_line(str(instruction), indent = 0)
			else:
				self.add_line(str(instruction))
		self.dedent()
		self.add_line()
		return function

	def __str__(self):
		return self.get_code()

class Function(object):
	def __init__(self, name, arguments, microarchitecture, abi):
		super(Function, self).__init__()
		self.name = name
		assert abi.name in ['x64-sysv', 'x64-ms']
		if abi.name == 'x64-ms':
			floating_point_parameter_registers = [xmm0, xmm1, xmm2, xmm3]
			integer_parameter_registers = [rcx, rdx, r8, r9]
			for (index, argument) in enumerate(arguments):
				if index < 4:
					if argument.get_type().is_floating_point():
						argument.register_location = floating_point_parameter_registers[index]
					else:
						argument.register_location = integer_parameter_registers[index]
				argument.stack_location = (index + 1) * 8
		elif abi.name == 'x64-sysv':
			available_floating_point_registers = [xmm0, xmm1, xmm2, xmm3, xmm4, xmm5, xmm6, xmm7]
			available_integer_registers = [rdi, rsi, rdx, rcx, r8, r9]
			parameter_offset = 8
			for argument in arguments:
				if argument.get_type().is_floating_point():
					if len(available_floating_point_registers) > 0:
						argument.register_location = available_floating_point_registers.pop(0)
				else:
					if len(available_integer_registers) > 0:
						argument.register_location = available_integer_registers.pop(0)
		self.arguments = arguments
		self.microarchitecture = microarchitecture
		self.symbol_name = "_" + name + "_" + microarchitecture.get_name()
		self.abi = abi
		self.stack_offset = 0
		self.instructions = list()
		self.constants = list()
		self.stack_frame = StackFrame(abi.stack_alignment, abi.pointer_size)
		self.local_variables_count = 0
		self.virtual_registers_count = 0x40
		self.uses_avx = False
		self.conflicting_registers = dict()
		self.possible_register_allocations = dict()

	def find_parameter(self, parameter):
		for argument in self.arguments:
			if argument.name == parameter.name:
				return argument
		return None

	def add_instruction(self, instruction):
		self.instructions.append(instruction)

	def generate_prolog_and_epilog(self):
		prologue_instructions = self.stack_frame.generate_prologue(self.abi, self.uses_avx)
		epilogue_instructions = self.stack_frame.generate_epilogue(self.abi, self.uses_avx)
		new_instructions = list()
		for instruction in self.instructions:
			if isinstance(instruction, Label):
				new_instructions.append(instruction)
				if instruction.name == 'ENTRY':
					new_instructions.extend(prologue_instructions)
			elif isinstance(instruction, ReturnInstruction):
				new_instructions.extend(epilogue_instructions)
				new_instructions.extend(instruction.to_instruction_list())
			elif isinstance(instruction, RetInstruction):
				new_instructions.extend(epilogue_instructions)
				new_instructions.append(instruction)
			else:
				new_instructions.append(instruction)
		self.instructions = new_instructions

	def generate_labels(self):
		for instruction in self.instructions:
			if isinstance(instruction, Label):
				if instruction.name == 'ENTRY':
					break
		else:
			self.instructions.insert(0, Label("ENTRY"))

#		follows_branch_instruction = False
#		label_number = 0
#		new_instructions = list()
#		for index, instruction in enumerate(self.instructions):
#			if isinstance(instruction, ConditionalJumpInstruction):
#				follows_branch_instruction = True
#			else:
#				if not isinstance(instruction, Label) and follows_branch_instruction:
#					new_instructions.append(Label("L" + str(label_number)))
#					label_number += 1
#				follows_branch_instruction = False
#			new_instructions.append(instruction)
#		self.instructions = new_instructions

	def get_label_table(self):
		label_table = dict()
		for index, instruction in enumerate(self.instructions):
			if isinstance(instruction, Label):
				label_table[instruction.name] = index
		return label_table

	def find_entry_label(self):
		for index, instruction in enumerate(self.instructions):
			if isinstance(instruction, Label):
				if instruction.name == 'ENTRY':
					return index
		raise ValueError('Instruction stream does not contain the ENTRY label')

	def find_ret_instructions(self):
		ret_instructions = list()
		for index, instruction in enumerate(self.instructions):
			if isinstance(instruction, RetInstruction) or isinstance(instruction, ReturnInstruction):
				ret_instructions.append(index)
		return ret_instructions

	def determine_available_registers(self):
		def mark_available_registers(instructions, start, initial_available_registers, label_table):
			available_registers = set(initial_available_registers)
			for i in range(start, len(instructions)):
				instruction = instructions[i]
				if isinstance(instruction, Instruction):
					instruction.available_registers = set(available_registers)
					if isinstance(instruction, ConditionalJumpInstruction):
						target_label = instruction.destination
						target_index = label_table[target_label]
						if not i in instructions[target_index].input_branches:
							instructions[target_index].input_branches.add(i)
							mark_available_registers(instructions, target_index, available_registers, label_table)
					elif isinstance(instruction, JmpInstruction):
						target_label = instruction.destination
						target_index = label_table[target_label]
						if not i in instructions[target_index].input_branches:
							instructions[target_index].input_branches.add(i)
							mark_available_registers(instructions, target_index, available_registers, label_table)
						return
					else:
						available_registers |= set(instruction.get_output_registers_list())

		label_table = self.get_label_table()
		current_index = self.find_entry_label()
		mark_available_registers(self.instructions, current_index, set(), label_table)

	def determine_live_registers(self):
		for instruction in self.instructions:
			if hasattr(instruction, 'is_visited'):
				instruction.is_visited = False

		def mark_live_registers(instructions, end, initial_live_registers):
			live_registers = dict(initial_live_registers)
			for i in range(end, -1, -1):
				instruction = instructions[i]
				if isinstance(instruction, JmpInstruction) and i != end:
					return
				elif isinstance(instruction, Instruction):
					# First mark registers which are written to by this instruction as non-live
					# Then mark registers which are read by this instruction as live
					for output_register in instruction.get_output_registers_list():
						# Instructions which write to 32-bit register implicitly zero out its upper 32 bit
						if isinstance(output_register, GeneralPurposeRegister32):
							output_register = output_register.get_qword()
						register_id = output_register.get_id()
						register_mask = output_register.get_mask()
						if register_id in live_registers:
							live_registers[register_id] &= ~register_mask
							if live_registers[register_id] == 0:
								del live_registers[register_id]
					for input_register in instruction.get_input_registers_list():
						register_id = input_register.get_id()
						register_mask = input_register.get_mask()
						if register_id in live_registers:
							live_registers[register_id] |= register_mask
						else:
							live_registers[register_id] = register_mask

					instruction.live_registers = set([Register.from_parts(id, mask) for (id, mask) in live_registers.iteritems()])
				elif isinstance(instruction, Label):
					for target_index in instruction.input_branches:
						if not instructions[target_index].is_visited:
							instructions[target_index].is_visited = True
							mark_live_registers(instructions, target_index, live_registers)

		ret_instructions = self.find_ret_instructions()
		for ret_instruction in ret_instructions:
			mark_live_registers(self.instructions, ret_instruction, set())

	def determine_register_relations(self):
		all_registers = self.abi.volatile_registers + list(reversed(self.abi.argument_registers)) + self.abi.callee_save_registers
		available_registers = { 'general-purpose': list(), 'mmx': list(), 'avx': list() }
		for register in all_registers:
			available_registers[register.type].append(register)
		for instruction in self.instructions:
			if isinstance(instruction, Instruction):
				virtual_live_registers = [register for register in instruction.live_registers if register.is_virtual()]
				for registerX in virtual_live_registers:
					if registerX not in self.conflicting_registers:
						self.conflicting_registers[registerX] = set()
					for registerY in virtual_live_registers:
						if registerX != registerY and registerX.type == registerY.type:
							self.conflicting_registers[registerX].add(registerY)
		for instruction in self.instructions:
			if isinstance(instruction, Instruction):
				for input_register in instruction.get_input_registers_list():
					if input_register.is_virtual() and input_register not in self.conflicting_registers:
						self.conflicting_registers[input_register] = set()
				for output_register in instruction.get_output_registers_list():
					if output_register.is_virtual() and output_register not in self.conflicting_registers:
						self.conflicting_registers[output_register] = set()
		# Mark available physical registers for each virtual register
		for register in self.conflicting_registers.iterkeys():
			self.possible_register_allocations[register] = list(available_registers[register.type])
		for instruction in self.instructions:
			if isinstance(instruction, Instruction):
				virtual_live_registers = [register for register in instruction.live_registers if register.is_virtual()]
				physical_live_registers = [register for register in instruction.live_registers if not register.is_virtual()]
				for virtual_register in virtual_live_registers:
					for physical_register in physical_live_registers:
						if physical_register in self.possible_register_allocations[virtual_register]:
							self.possible_register_allocations[virtual_register].remove(physical_register)


	def allocate_registers(self):
		register_allocation = dict()
		for register in self.possible_register_allocations.iterkeys():
			register_allocation[register] = None

		def bind_register(virtual_register, physical_register):
			# Remove option to allocate any conflicting virtual register to the same physical register
			for conflicting_register in self.conflicting_registers[virtual_register]:
				if physical_register in self.possible_register_allocations[conflicting_register]:
					self.possible_register_allocations[conflicting_register].remove(physical_register)
			register_allocation[virtual_register] = physical_register

		def is_allocated(virtual_register):
			assert isinstance(virtual_register, Register)
			assert virtual_register.is_virtual()
			return bool(register_allocation[virtual_register])

		# First allocate parameters
		for instruction in self.instructions:
			if isinstance(instruction, LoadParameterPseudoInstruction):
				if instruction.parameter.register_location:
					# print instruction
					if instruction.destination.register.is_virtual():
						if not is_allocated(instruction.destination.register):
							if instruction.parameter.register_location in self.possible_register_allocations[instruction.destination.register]:
								bind_register(instruction.destination.register, instruction.parameter.register_location)
		# Now allocate all other register
		for virtual_register in sorted(self.conflicting_registers.iterkeys(), key = lambda register: len(self.conflicting_registers[register]), reverse = True):
			if not is_allocated(virtual_register):
				if not self.possible_register_allocations[virtual_register]:
					print virtual_register
				assert self.possible_register_allocations[virtual_register]
				physical_register = self.possible_register_allocations[virtual_register][0]
				bind_register(virtual_register, physical_register)
		for instruction in self.instructions:
			if isinstance(instruction, Instruction):
				# print instruction
				for input_register in instruction.get_input_registers_list():
					if input_register.is_virtual():
						input_register.bind(register_allocation[input_register].get_id())
				for output_register in instruction.get_output_registers_list():
					if output_register.is_virtual():
						if output_register in register_allocation:
							output_register.bind(register_allocation[output_register].get_id())
		# print "Register allocations:"
		# for register in register_allocation:
		# 	if register_allocation[register]:
		# 		print "\t{0} -> {1}".format(register, register_allocation[register])


	def update_stack_frame(self):
		nonvolatile_registers = set(self.abi.callee_save_registers)
		modified_registers = set()
		local_variables = list()
		for instruction in self.instructions:
			if isinstance(instruction, Instruction):
				if instruction.get_isa_extension() in ['AVX', 'AVX2', 'F16C', 'FMA3', 'FMA4']:
					self.uses_avx = True

				local_variable = instruction.get_local_variable()
				if local_variable is not None:
					if not local_variable in local_variables:
						local_variables.append(local_variable)
				for output_register in instruction.get_output_registers_list():
					if isinstance(output_register, GeneralPurposeRegister):
						modified_registers.add(output_register.get_qword())
					elif isinstance(output_register, SSERegister):
						modified_registers.add(output_register)
					elif isinstance(output_register, AVXRegister):
						modified_registers.add(output_register.get_oword())
			# if isinstance(instruction, LoadParameterPseudoInstruction):
			# 	parameter = instruction.parameter
			# 	for live_register in instruction.live_registers:
			# 		if parameter.register_location in live_register:
			# 			parameter.save_on_stack = True
		preserved_registers = list(nonvolatile_registers & modified_registers)
		self.stack_frame.preserve_registers(preserved_registers)

		if local_variables:
			max_variable_size = 1
			local_variables_map = dict()
			for local_variable in local_variables:
				variable_size = local_variable.get_size()
				max_variable_size = max(max_variable_size, variable_size)
				if not variable_size in local_variables_map:
					local_variables_map[variable_size] = list()
				local_variables_map[variable_size].append(local_variable)
			self.stack_frame.local_variables_alignment = max_variable_size
			while local_variables:
				local_variable = local_variables.pop(0)
				variable_size = local_variable.get_size()
				local_variables_map[variable_size].remove(local_variable)
				bucket_size = variable_size
				self.stack_frame.local_variables.append(local_variable)
				while bucket_size < max_variable_size:
					if local_variables_map[variable_size]:
						local_variable = local_variables_map[variable_size].pop(0)
						local_variables.remove(local_variable)
						assert variable_size == local_variable.get_size()
						bucket_size += variable_size
						self.stack_frame.local_variables.append(local_variable)
					else:
						if bucket_size % (variable_size * 2) != 0:
							local_variable = LocalVariable(variable_size)
							bucket_size += variable_size
							self.stack_frame.local_variables.append(local_variable)
						variable_size *= 2
		for parameter in self.arguments:
			if parameter.save_on_stack:
				self.stack_frame.preserve_parameter(parameter)
		self.stack_frame.allocate_space(self.abi)

	def generate_parameter_loads(self):
		new_instructions = list()
		for instruction in self.instructions:
			if isinstance(instruction, LoadParameterPseudoInstruction):
				parameter = instruction.parameter
				if parameter.save_on_stack:
					# If parameter is on stack, generate load from stack:
					#   MOV( instruction.destination, [rsp + parameter.stack_location] )
					parameter_offset = parameter.stack_location
					new_instruction = MovInstruction(instruction.destination, Operand([rsp + parameter_offset]))
					new_instruction.stack_offset = instruction.stack_offset
					new_instructions.append(new_instruction)
				else:
					# If parameter is in a register, use register-register move:
					if instruction.destination.register != parameter.register_location:
						# Parameter is in a different register than instruction destination, generate move:
						#   MOV( instruction.destination, parameter.register_location )
						new_instruction = MovInstruction(instruction.destination, Operand(parameter.register_location))
						new_instruction.stack_offset = instruction.stack_offset
						new_instruction.live_registers = instruction.live_registers
						new_instruction.available_registers = instruction.available_registers
						new_instructions.append(new_instruction)
					# If parameter is in the same register as instruction destination, no instruction needed:
					#   MOV( instruction.destination == parameter.register_location, parameter.register_location )
					# is a no-op
			else:
				new_instructions.append(instruction)
		self.instructions = new_instructions

	def generate_constant_loads(self):
		max_alignment = 0
		for instruction in self.instructions:
			if isinstance(instruction, Instruction):
				constant = instruction.get_constant()
				if constant is not None:
					constant_alignment = constant.get_alignment()
					constant_size = constant.size * constant.repeats
					max_alignment = max(max_alignment, constant_alignment)

		constant_id = 0
		constant_label_map = dict()
		constant_buckets = dict()
		for instruction in self.instructions:
			if isinstance(instruction, Instruction):
				constant = instruction.get_constant()
				if constant is not None:
					if constant in constant_label_map:
						constant.label = constant_label_map[constant]
					else:
						constant.label = "c" + str(constant_id)
						constant_id += 1
						constant_label_map[constant] = constant.label
						constant_alignment = constant.get_alignment()
						constant_size = constant.size * constant.repeats
						if constant_alignment in constant_buckets:
							constant_buckets[constant_alignment].add(constant)
							if constant_buckets[constant_alignment].is_full():
								del constant_buckets[constant_alignment]
						else:
							constant_bucket = ConstantBucket(max_alignment / 8)
							constant_bucket.add(constant)
							self.constants.append(constant_bucket)
							if not constant_bucket.is_full():
								constant_buckets[constant_alignment] = constant_bucket

		new_instructions = list()
		for instruction in self.instructions:
			if isinstance(instruction, LoadConstantPseudoInstruction):
				constant = instruction.source.constant
				assert constant.size * constant.repeats == 128

				if constant.basic_type == 'float32':
					if self.uses_avx:
						new_instruction = VMOVAPS( instruction.destination.register, instruction.source.constant )
					else:
						new_instruction = MOVAPS( instruction.destination.register, instruction.source.constant )
				elif constant.basic_type == 'float64':
					if self.uses_avx:
						new_instruction = VMOVAPD( instruction.destination.register, instruction.source.constant )
					else:
						new_instruction = MOVAPD( instruction.destination.register, instruction.source.constant )
				else:
					if self.uses_avx:
						new_instruction = VMOVDQA( instruction.destination.register, instruction.source.constant )
					else:
						new_instruction = MOVDQA( instruction.destination.register, instruction.source.constant )
				new_instruction.stack_offset = instruction.stack_offset
				new_instructions.append(new_instruction)
			else:
				new_instructions.append(instruction)
		self.instructions = new_instructions

	def get_isa_extensions(self):
		isa_extensions = set()
		for instruction in self.instructions:
			if isinstance(instruction, Instruction):
				isa_extensions.add(instruction.get_isa_extension())
		return list(isa_extensions)

	def get_yeppp_isa_extensions(self):
		isa_extensions_map = {'CMOV':            ('CMOV',       None,           None),
		                      'MMX':             ( None,       'MMX',          'FPU'),
		                      'MMX+':            ( None,       'MMXPlus',      'FPU'),
		                      '3dnow!':          ( None,       '3dnow',        'FPU'),
		                      '3dnow!+':         ( None,       '3dnowPlus',    'FPU'),
		                      '3dnow! Prefetch': ( None,       '3dnowPrefetch', None),
		                      'SSE':             ( None,       'SSE',          'SSE'),
		                      'SSE2':            ( None,       'SSE2',         'SSE'),
		                      'SSE3':            ( None,       'SSE3',         'SSE'),
		                      'SSSE3':           ( None,       'SSSE3',        'SSE'),
		                      'SSE4A':           ( None,       'SSE4A',        'SSE'),
		                      'SSE4.1':          ( None,       'SSE4_1',       'SSE'),
		                      'SSE4.2':          ( None,       'SSE4_2',       'SSE'),
		                      'AVX':             ( None,       'AVX',          'AVX'),
		                      'AVX2':            ( None,       'AVX2',         'AVX'),
		                      'FMA3':            ( None,       'FMA3',         'AVX'),
		                      'FMA4':            ( None,       'FMA4',         'AVX'),
		                      'F16C':            ( None,       'F16C',         'AVX'),
							  'MOVBE':           ('Movbe',      None,           None),
							  'POPCNT':          ('Popcnt',     None,           None),
							  'LZCNT':           ('Lzcnt',      None,           None),
							  'TZCNT':           ('Tzcnt',      None,           None),
							  'BMI':             ('BMI',        None,           None),
							  'BMI2':            ('BMI2',       None,           None),
							  'TBM':             ('TBM',        None,           None),
							  'AES':             ('AES',        None,          'SSE'),
							  'VAES':            ('AES',        None,          'AVX'),
							  'RDRAND':          ('Rdrand',     None,           None),
							  'RDSEED':          ('Rdseed',     None,           None),
							  'PCLMULQDQ':       ('Pclmulqdq',  None,          'SSE'),
							  'VPCLMULQDQ':      ('Pclmulqdq',  None,          'AVX')}
		(isa_extensions, simd_extensions, system_extensions) = (set(), set(), set())
		for isa_extension in self.get_isa_extensions():
			if isa_extension is not None:
				(isa_extension, simd_extension, system_extension) = isa_extensions_map[isa_extension]
				if isa_extension is not None:
					isa_extensions.add(isa_extension)
				if simd_extension is not None:
					simd_extensions.add(simd_extension)
				if system_extension is not None:
					system_extensions.add(system_extension)
		isa_extensions = map(lambda id: "YepX86IsaFeature" + id, isa_extensions)
		if not isa_extensions:
			isa_extensions = ["YepIsaFeaturesDefault"]
		simd_extensions = map(lambda id: "YepX86SimdFeature" + id, simd_extensions)
		if not simd_extensions:
			simd_extensions = ["YepSimdFeaturesDefault"]
		system_extensions = map(lambda id: "YepX86SystemFeature" + id, system_extensions)
		if not system_extensions:
			system_extensions = ["YepSystemFeaturesDefault"]
		return (isa_extensions, simd_extensions, system_extensions)

	def allocate_local_variable(self):
		self.local_variables_count += 1
		return self.local_variables_count

	def allocate_avx_register(self):
		self.virtual_registers_count += 1
		return (self.virtual_registers_count << 12) | 0x00C0

	def allocate_sse_register(self):
		self.virtual_registers_count += 1
		return (self.virtual_registers_count << 12) | 0x0040

	def allocate_mmx_register(self):
		self.virtual_registers_count += 1
		return (self.virtual_registers_count << 12) | 0x0010

	def allocate_general_purpose_register64(self):
		self.virtual_registers_count += 1
		return (self.virtual_registers_count << 12) | 0x000F

	def allocate_general_purpose_register32(self):
		self.virtual_registers_count += 1
		return (self.virtual_registers_count << 12) | 0x0007

	def allocate_general_purpose_register16(self):
		self.virtual_registers_count += 1
		return (self.virtual_registers_count << 12) | 0x0003

	def allocate_general_purpose_register8(self):
		self.virtual_registers_count += 1
		return (self.virtual_registers_count << 12) | 0x0001

class LocalVariable(object):
	def __init__(self, register_type):
		super(LocalVariable, self).__init__()
		if isinstance(register_type, int):
			self.size = register_type
		elif register_type == GeneralPurposeRegister64:
			self.size = 8
		elif register_type == GeneralPurposeRegister32:
			self.size = 4
		elif register_type == GeneralPurposeRegister16:
			self.size = 2
		elif register_type == GeneralPurposeRegister8:
			self.size = 1
		elif register_type == MMXRegister:
			self.size = 8
		elif register_type == SSERegister:
			self.size = 16
		elif register_type == AVXRegister:
			self.size = 32
		else:
			raise ValueError('Unsupported register type {0}'.format(register_type))
		self.id = current_function.allocate_local_variable()
		self.offset = None
		self.base = None

	def __eq__(self, other):
		return self.id == other.id

	def __hash__(self):
		return hash(self.id)

	def __str__(self):
		if self.offset is not None and self.base is not None:
			return "[{0} + {1}]".format(self.base, self.offset)
		else:
			return "local-variable<{0}>".format(self.id)

	def get_size(self):
		return self.size

class StackFrame(object):
	def __init__(self, stack_alignment, stack_misalignment):
		super(StackFrame, self).__init__()
		self.stack_alignment = stack_alignment
		self.stack_misalignment = stack_misalignment
		self.general_purpose_registers = list()
		self.sse_registers = list()
		self.local_variables = list()
		self.parameters = list()
		self.local_variables_alignment = 0
		self.general_purpose_registers_space = None
		self.sse_registers_space = None
		self.parameters_space = None
		self.local_variables_space = None

	def preserve_parameter(self, parameter):
		if not parameter in self.parameters:
			self.parameters.append(parameter)

	def preserve_registers(self, registers):
		for register in registers:
			self.preserve_register(register)

	def preserve_register(self, register):
		if isinstance(register, GeneralPurposeRegister8):
			register = register.get_qword()
		elif isinstance(register, GeneralPurposeRegister16):
			register = register.get_qword()
		elif isinstance(register, GeneralPurposeRegister32):
			register = register.get_qword()

		if isinstance(register, GeneralPurposeRegister64):
			if not register in self.general_purpose_registers:
				self.general_purpose_registers.append(register)
		elif isinstance(register, SSERegister):
			if not register in self.sse_registers:
				self.sse_registers.append(register)
		else:
			raise TypeError("Unsupported register type {0}".format(type(register)))

	def allocate_space(self, abi):
		self.general_purpose_registers_space = 0
		self.sse_registers_space = 0
		self.parameters_space = 0
		self.local_variables_space = 0

		stack_misalignment = self.stack_misalignment
		if self.general_purpose_registers:
			assert self.stack_alignment >= 8
			if stack_misalignment % 8 != 0:
				self.general_purpose_registers_space += 8 - stack_misalignment % 8
				stack_misalignment = (stack_misalignment + (8 - stack_misalignment % 8)) % self.stack_alignment
			self.general_purpose_registers_space += len(self.general_purpose_registers) * 8
			stack_misalignment = (stack_misalignment + len(self.general_purpose_registers) * 8) % self.stack_alignment
		if self.sse_registers:
			assert self.stack_alignment >= 16
			if stack_misalignment % 16 != 0:
				self.sse_registers_space += 16 - stack_misalignment % 16
				stack_misalignment = (stack_misalignment + (16 - stack_misalignment % 16)) % self.stack_alignment
			self.sse_registers_space += len(self.sse_registers) * 16
			stack_misalignment = (stack_misalignment + len(self.sse_registers) * 16) % self.stack_alignment
		if abi.get_name() != 'x64-ms':
			if self.parameters:
				assert self.stack_alignment >= 8
				if stack_misalignment % 8 != 0:
					self.parameters_space += 8 - stack_misalignment % 8
					stack_misalignment = (stack_misalignment + (8 - stack_misalignment % 8)) % self.stack_alignment
				self.parameters_space += len(self.parameters) * 8
			stack_misalignment = (stack_misalignment + len(self.parameters) * 8) % self.stack_alignment
		if self.local_variables:
			assert self.stack_alignment >= self.local_variables_alignment
			if stack_misalignment % self.local_variables_alignment != 0:
				self.local_variables_space += self.local_variables_alignment - stack_misalignment % self.local_variables_alignment
				stack_misalignment =  (stack_misalignment + (self.local_variables_alignment - stack_misalignment % self.local_variables_alignment)) % self.stack_alignment
			for local_variable in self.local_variables:
				self.local_variables_space += local_variable.get_size()
				stack_misalignment = (stack_misalignment + local_variable.get_size()) % self.stack_alignment

	def generate_prologue(self, abi, use_avx):
		instructions = list()
		if abi.get_name() == 'x64-ms':
			for parameter in self.parameters:
				instructions.append(MovInstruction(Operand([rsp + parameter.stack_location]), Operand(parameter.register_location)))
		assert self.stack_misalignment % 8 == 0
		for register in self.general_purpose_registers:
			# Situation where we have to align stack pointer is not handled
			instructions.append(PushInstruction(Operand(register)))
		stack_frame_size = self.sse_registers_space + self.parameters_space + self.local_variables_space
		if stack_frame_size != 0:
			instructions.append(ArithmeticInstruction("SUB", Operand(rsp), Operand(stack_frame_size)))
			for index, register in enumerate(self.sse_registers):
				offset = stack_frame_size - self.sse_registers_space + index * 16
				if use_avx:
					instructions.append( VMOVAPS([rsp + offset], register) )
				else:
					instructions.append( MOVAPS([rsp + offset], register) )
			if abi.get_name() != 'x64-ms':
				for index, parameter in enumerate(self.parameters):
					assert not parameter.get_type().is_floating_point()
					offset = stack_frame_size - self.sse_registers_space - self.parameters_space + index * 8
					instructions.append(MovInstruction(Operand([rsp + offset]), Operand(parameter.register_location)))
			offset = 0
			for local_variable in self.local_variables:
				local_variable.base = rsp
				local_variable.offset = offset
				offset += local_variable.get_size()


#		assert len(self.avx_registers) == 0
#		if self.sse_registers:
#			instructions.append(ArithmeticInstruction("SUB", Operand(rsp), Operand(additional_frame_size)))
#
#			register_offset = 0
#			for sse_register in self.sse_registers:
#				instructions.append(SseMovInstruction("MOVAPS", Operand([rsp + register_offset]), Operand(sse_register)))
#				register_offset += 16
		return instructions

	def generate_epilogue(self, abi, use_avx):
#		additional_frame_size = self.get_additional_register_save_space()

		instructions = list()

		stack_frame_size = self.sse_registers_space + self.parameters_space + self.local_variables_space
		if stack_frame_size != 0:
			for index, register in enumerate(self.sse_registers):
				offset = stack_frame_size - self.sse_registers_space + index * 16
				if use_avx:
					instructions.append(AvxMovInstruction("VMOVAPS", Operand(register), Operand([rsp + offset])))
				else:
					instructions.append(SseMovInstruction("MOVAPS", Operand(register), Operand([rsp + offset])))
			instructions.append(ArithmeticInstruction("ADD", Operand(rsp), Operand(stack_frame_size)))
		if use_avx:
			instructions.append(AvxZeroInstruction("VZEROUPPER"))

		for register in reversed(self.general_purpose_registers):
			instructions.append(PopInstruction(Operand(register)))

#		assert len(self.avx_registers) == 0
#		if len(self.sse_registers) > 0:
#			register_offset = 0
#			for sse_register in self.sse_registers:
#				instructions.append(SseMovInstruction("MOVAPS", Operand(sse_register), Operand([rsp + register_offset])))
#				register_offset += 16
#
#			instructions.append(ArithmeticInstruction("ADD", Operand(rsp), Operand(additional_frame_size)))
#
#		instructions.extend([PopInstruction(Operand(register)) for register in self.general_purpose_registers])
		return instructions

class ConstantBucket(object):
	supported_capacity = [1, 2, 4, 8, 16, 32, 64, 128]

	def __init__(self, capacity):
		super(ConstantBucket, self).__init__()
		if isinstance(capacity, int):
			if capacity in ConstantBucket.supported_capacity:
				self.capacity = capacity
			else:
				raise ValueError("Capacity value {0} is not among the supports capacities ({1})".format(capacity, ConstantBucket.supported_capacity))
		else:
			raise TypeError("Constant capacity {0} must be an integer".format(capacity))
		self.size = 0
		self.constants = list()

	def add(self, constant):
		if isinstance(constant, Constant):
			if constant.get_alignment() > self.get_capacity() * 8:
				raise ValueError("Constants alignment exceeds constant bucket alignment")
			elif (self.size * 8) % constant.get_alignment() != 0:
				raise ValueError("Constant alignment is not compatible with the internal constant bucket alignment")
			elif self.size + constant.size * constant.repeats / 8 > self.capacity:
				raise ValueError("Constant bucket is overflowed")
			else:
				self.constants.append(constant)
				self.size += constant.size * constant.repeats / 8
		else:
			raise TypeError("Only Constant objects can be added to a constant bucket")

	def get_capacity(self):
		return self.capacity

	def is_empty(self):
		return self.size == 0

	def is_full(self):
		return self.size == self.capacity

	def empty(self):
		constants = self.constants
		self.constants = list()
		return constants

class Constant(object):
	supported_sizes = [8, 16, 32, 64, 128, 256]
	supported_types = ['uint8', 'uint16', 'uint32', 'uint64', 'uint128', 'uint256',
	                   'sint8', 'sint16', 'sint32', 'sint64', 'sint128', 'sint256',
	                   'int8',  'int16',  'int32',  'int64',  'int128',  'int256',
	                   'float16', 'float32', 'float64', 'float80', 'float128']

	def __init__(self, size, repeats, data, basic_type):
		super(Constant, self).__init__()
		if isinstance(size, int):
			if size in Constant.supported_sizes:
				self.size = size
			else:
				raise ValueError("Constant size {0} is not among the supported sizes ({1})".format(size, ", ".join(Constant.supported_sizes)))
		else:
			raise TypeError("Constant size {0} must be an integer".format(size))
		if isinstance(repeats, int):
			if size in Constant.supported_sizes:
				self.repeats = repeats
			else:
				raise ValueError("The number of constant repeats {0} is not among the supported repeat numbers ({1})".format(repeats, ", ".join(Constant.supported_sizes)))
		else:
			raise TypeError("The number of constant repeats {0} must be an integer".format(repeats))
		if isinstance(basic_type, str):
			if basic_type in Constant.supported_types:
				self.basic_type = basic_type
		else:
			raise TypeError("The basic type {0} of a constant must be an integer".format(basic_type))
		self.data = data
		self.label = None
		self.prefix = None

	def __str__(self):
		text = hex(self.data)
		if text.endswith("L"):
			text = text[:-1]
		if text.startswith("0x"):
			text = text[2:]
		if len(text) < self.size * 2:
			text = "0" * (self.size / 4 - len(text)) + text
		text = "0x" + text.upper()
		if self.size == 8:
			return text
		elif self.size == 16:
			return text
		elif self.size == 32:
			return text
		elif self.size == 64:
			return text
		elif self.size == 128:
			return text

	def __hash__(self):
		return hash(self.data) ^ hash(self.size) ^ hash(self.repeats)

	def __eq__(self, other):
		if isinstance(other, Constant):
			if self.size == other.size and self.repeats == other.repeats:
				return self.data == other.data
			else:
				return False
		else:
			return False

	def get_alignment(self):
		if self.size == 80:
			return 128
		else:
			return self.size * self.repeats

	@staticmethod
	def uint64(number):
		if isinstance(number, int) or isinstance(number, long):
			if 0 <= number <= 18446744073709551615L:
				return Constant(64, 1, long(number))
			elif -9223372036854775808L <= number < 0:
				return Constant(64, 1, long(number + 18446744073709551616L))
			else:
				raise ValueError("The number {0} is not a 64-bit integer".format(number))
		else:
			raise TypeError("The number used to construct a 64-bit unsigned integer constant must be an integer")

	@staticmethod
	def uint64x2(number1, number2 = None):
		if isinstance(number1, int) or isinstance(number1, long):
			if 0 <= number1 <= 18446744073709551615L:
				number1 = long(number1)
			elif -9223372036854775808L <= number1 < 0:
				number1 = long(number1 + 18446744073709551616L)
			else:
				raise ValueError("The number {0} is not a 64-bit integer".format(number1))
		else:
			raise TypeError("The number used to construct a 64-bit unsigned integer constant must be an integer")
		if number2 is None:
			number2 = number1
		elif isinstance(number2, int) or isinstance(number2, long):
			if 0 <= number2 <= 18446744073709551615L:
				number2 = long(number2)
			elif -9223372036854775808L <= number2 < 0:
				number2 = long(number2 + 18446744073709551616L)
			else:
				raise ValueError("The number {0} is not a 64-bit integer".format(number2))
		else:
			raise TypeError("The number used to construct a 64-bit unsigned integer constant must be an integer")
		if number1 == number2:
			return Constant(64, 2, number1, 'uint64')
		else:
			return Constant(128, 1, (number1 << 64) + number2, 'uint64')

	@staticmethod
	def uint32(number):
		if isinstance(number, int) or isinstance(number, long):
			if 0 <= number <= 4294967295L:
				return Constant(32, 1, long(number), 'uint32')
			elif -2147483648L <= number < 0:
				return Constant(32, 1, long(number + 4294967296L), 'uint32')
			else:
				raise ValueError("The number {0} is not a 32-bit integer".format(number))
		else:
			raise TypeError("The number used to construct a 32-bit unsigned integer constant must be an integer")

	@staticmethod
	def uint16(number):
		if isinstance(number, int) or isinstance(number, long):
			if 0 <= number <= 65535:
				return Constant(16, 1, long(number), 'uint16')
			elif -32768 <= number < 0:
				return Constant(16, 1, long(number + 65536), 'uint16')
			else:
				raise ValueError("The number {0} is not a 16-bit integer".format(number))
		else:
			raise TypeError("The number used to construct a 16-bit unsigned integer constant must be an integer")

	@staticmethod
	def uint8(number):
		if isinstance(number, int) or isinstance(number, long):
			if 0 <= number <= 255:
				return Constant(8, 1, long(number), 'uint8')
			elif -128 <= number < 0:
				return Constant(8, 1, long(number + 256), 'uint8')
			else:
				raise ValueError("The number {0} is not an 8-bit integer".format(number))
		else:
			raise TypeError("The number used to construct an 8-bit unsigned integer constant must be an integer")

	@staticmethod
	def float64(number):
		return Constant(64, 1, Constant.parse_float64(number), is_floating_point = True)

	@staticmethod
	def float64x2(number1, number2 = None):
		number1 = Constant.parse_float64(number1)
		if number2 is None:
			number2 = number1
		else:
			number2 = Constant.parse_float64(number2)
		if number1 == number2:
			return Constant(64, 2, number1, 'float64')
		else:
			return Constant(128, 1, (number1 << 64) + number2, 'float64')

	@staticmethod
	def parse_float64(number):
		if isinstance(number, float):
			number = float.hex(number)
		elif isinstance(number, int) or isinstance(number, long):
			number = float.hex(float(number))
		elif isinstance(number, str):
			# Validity check
			float.hex(float.fromhex(number))
		is_negative = number.startswith("-")
		point_position = number.index('.')
		exp_position = number.rindex('p')
		number_prefix = number[int(is_negative):point_position]
		assert number_prefix == '0x0' or number_prefix == '0x1'
		mantissa = number[point_position + 1:exp_position]
		if number_prefix == '0x0' and int(mantissa) == 0:
			# Zero
			return long(is_negative) << 63
		else:
			exponent = number[exp_position + 1:]
			mantissa_bits = len(mantissa) * 4
			if mantissa_bits == 52:
				mantissa = long(mantissa, 16)
			elif mantissa_bits < 52:
				mantissa = long(mantissa, 16) << (52 - mantissa_bits)
			else:
				mantissa = long(mantissa, 16) >> (mantissa_bits - 52)
			exponent = int(exponent)
			if exponent <= -1023:
				# Denormals
				mantissa = (mantissa + (1 << 52) ) >> -(exponent + 1022)
				exponent = -1023
			elif exponent > 1023:
				# Infinity
				mantissa = 0
				exponent = 1023
			return mantissa + (long(exponent + 1023) << 52) + (long(is_negative) << 63)

class Register(object):
	def __init__(self):
		super(Register, self).__init__()

	def __lt__(self, other):
		return self.number < other.number

	def __le__(self, other):
		return self.number <= other.number

	def __eq__(self, other):
		return self.number == other.number

	def __ne__(self, other):
		return self.number != other.number

	def __gt__(self, other):
		return self.number > other.number

	def __ge__(self, other):
		return self.number >= other.number

	def __contains__(self, register):
		if self.get_id() == register.get_id():
			register_mask = register.get_mask()
			return ((self.get_mask() & register_mask) == register_mask)
		else:
			return False

	def __hash__(self):
		return hash(self.number)

	def get_id(self):
		return (self.number >> 12)

	def get_mask(self):
		return (self.number & 0xFFF)

	def get_type(self):
		return self.type

	@staticmethod
	def from_parts(id, mask):
		# Mask not necessarily is valid, e.g. it can include only the high part of a 32-bit register
		assert mask != 0
		if (mask & ~0x00F) == 0:
			# General-purpose register
			if (mask & 0x8) == 0x8:
				return GeneralPurposeRegister64((id << 12) | 0x00F)
			elif (mask & 0x4) == 0x4:
				return GeneralPurposeRegister32((id << 12) | 0x007)
			elif (mask & 0x3) == 0x3:
				return GeneralPurposeRegister16((id << 12) | 0x003)
			else:
				return GeneralPurposeRegister8((id << 12) | mask)
		elif (mask & ~0x010) == 0:
			# MMX register
			return MMXRegister((id << 12) | 0x010)
		elif (mask & ~0x0C0) == 0:
			# AVX or SSE register
			if (mask & 0x080) == 0x080:
				return AVXRegister((id << 12) | 0x0C0)
			else:
				return SSERegister((id << 12) | 0x040)
		else:
			print mask
			# Unknown register mask
			assert False


	@staticmethod
	def eliminate_enclosed_registers(registers):
		new_registers = set()
		has_small_registers = False
		for register in registers:
			if isinstance(register, SSERegister):
				new_registers.add(register)
			elif isinstance(register, AVXRegister):
				new_registers.add(register)
			elif isinstance(register, GeneralPurposeRegister64):
				new_registers.add(register)
			elif isinstance(register, GeneralPurposeRegister):
				has_small_registers = True
		if has_small_registers:
			for register in registers:
				if isinstance(register, GeneralPurposeRegister32):
					for new_register in new_registers:
						if register in new_register:
							break
					else:
						new_registers.add(register)
			for register in registers:
				if isinstance(register, GeneralPurposeRegister16):
					for new_register in new_registers:
						if register in new_register:
							break
					else:
						new_registers.add(register)
			for register in registers:
				if isinstance(register, GeneralPurposeRegister8):
					for new_register in new_registers:
						if register in new_register:
							break
					else:
						new_registers.add(register)
		return new_registers

	def is_virtual(self):
		return self.number >= 0x40000

	def bind(self, id):
		assert self.is_virtual()
		self.number = (self.number & 0xFFF) | (id << 12)
		assert not self.is_virtual()

class GeneralPurposeRegister(Register):
	def __init__(self):
		super(GeneralPurposeRegister, self).__init__()

class GeneralPurposeRegister64(GeneralPurposeRegister):
	name_to_number_map = {'rax': 0x2000F,
	                      'rbx': 0x2100F,
	                      'rcx': 0x2200F,
	                      'rdx': 0x2300F,
	                      'rsi': 0x2400F,
	                      'rdi': 0x2500F,
	                      'rbp': 0x2600F,
	                      'rsp': 0x2700F,
	                      'r8' : 0x2800F,
	                      'r9' : 0x2900F,
	                      'r10': 0x2A00F,
	                      'r11': 0x2B00F,
	                      'r12': 0x2C00F,
	                      'r13': 0x2D00F,
	                      'r14': 0x2E00F,
	                      'r15': 0x2F00F}

	number_to_name_map = {0x2000F: 'rax',
	                      0x2100F: 'rbx',
	                      0x2200F: 'rcx',
	                      0x2300F: 'rdx',
	                      0x2400F: 'rsi',
	                      0x2500F: 'rdi',
	                      0x2600F: 'rbp',
	                      0x2700F: 'rsp',
	                      0x2800F: 'r8',
	                      0x2900F: 'r9',
	                      0x2A00F: 'r10',
	                      0x2B00F: 'r11',
	                      0x2C00F: 'r12',
	                      0x2D00F: 'r13',
	                      0x2E00F: 'r14',
	                      0x2F00F: 'r15'}

	def __init__(self, id = None):
		super(GeneralPurposeRegister64, self).__init__()
		if isinstance(id, str):
			if id in GeneralPurposeRegister64.name_to_number_map.iterkeys():
				self.number = GeneralPurposeRegister64.name_to_number_map[id]
				self.type = 'general-purpose'
				self.size = 64
			else:
				raise ValueError('Unknown register name: {0}'.format(id))
		elif isinstance(id, int):
			self.number = id
			self.type = 'general-purpose'
			self.size = 64
		elif isinstance(id, GeneralPurposeRegister64):
			self.number = id.number
			self.type = id.type
			self.size = id.size
		elif id is None:
			self.number = current_function.allocate_general_purpose_register64()
			self.type = 'general-purpose'
			self.size = 64
		else:
			raise TypeError('Register id  is neither a name of an architectural general-purpose 64-bit register, nor an id of a virtual register')

	def __str__(self):
		if self.is_virtual():
			return 'gp64-vreg<{0}>'.format((self.number - 0x40000) >> 12)
		else:
			return GeneralPurposeRegister64.number_to_name_map[self.number]

	def __add__(self, offset):
		if isinstance(offset, int):
			return GeneralPurposeRegister64WithOffset(self, offset)
		else:
			raise TypeError('offset is not a integer number')

	def is_extended(self):
		return self.number >= 0x28000

	def get_low_byte(self):
		return GeneralPurposeRegister8((self.number & -16) | 0x0001)

	def get_word(self):
		return GeneralPurposeRegister16((self.number & -16) | 0x0003)

	def get_dword(self):
		return GeneralPurposeRegister32((self.number & -16) | 0x0007)

	def get_qword(self):
		return GeneralPurposeRegister64((self.number & -16) | 0x000F)

rax = GeneralPurposeRegister64('rax')
rbx = GeneralPurposeRegister64('rbx')
rcx = GeneralPurposeRegister64('rcx')
rdx = GeneralPurposeRegister64('rdx')
rsi = GeneralPurposeRegister64('rsi')
rdi = GeneralPurposeRegister64('rdi')
rbp = GeneralPurposeRegister64('rbp')
rsp = GeneralPurposeRegister64('rsp')
r8  = GeneralPurposeRegister64('r8')
r9  = GeneralPurposeRegister64('r9')
r10 = GeneralPurposeRegister64('r10')
r11 = GeneralPurposeRegister64('r11')
r12 = GeneralPurposeRegister64('r12')
r13 = GeneralPurposeRegister64('r13')
r14 = GeneralPurposeRegister64('r14')
r15 = GeneralPurposeRegister64('r15')

class GeneralPurposeRegister64WithOffset(object):
	def __init__(self, register, offset):
		super(GeneralPurposeRegister64WithOffset, self).__init__()
		self.register = register
		self.offset = offset
		if not isinstance(register, GeneralPurposeRegister64):
			raise TypeError('register is not an instance of the 64-bit general-purpose register clas')
		if not isinstance(offset, int):
			raise TypeError('offset is not a integer number')

	def __add__(self, offset):
		if isinstance(offset, int):
			return GeneralPurposeRegister64WithOffset(self.register, self.offset + offset)
		else:
			raise TypeError('offset is not a integer number')

	def __sub__(self, offset):
		if isinstance(offset, int):
			return GeneralPurposeRegister64WithOffset(self.register, self.offset - offset)
		else:
			raise TypeError('offset is not a integer number')

	def __str__(self):
		if self.offset == 0:
			return str(self.register)
		else:
			return "{0} + {1}".format(self.register, self.offset)

class GeneralPurposeRegister32(GeneralPurposeRegister):
	name_to_number_map = {'eax' : 0x20007,
	                      'ebx' : 0x21007,
	                      'ecx' : 0x22007,
	                      'edx' : 0x23007,
	                      'esi' : 0x24007,
	                      'edi' : 0x25007,
	                      'ebp' : 0x26007,
	                      'esp' : 0x27007,
	                      'r8d' : 0x28007,
	                      'r9d' : 0x29007,
	                      'r10d': 0x2A007,
	                      'r11d': 0x2B007,
	                      'r12d': 0x2C007,
	                      'r13d': 0x2D007,
	                      'r14d': 0x2E007,
	                      'r15d': 0x2F007}

	number_to_name_map = {0x20007: 'eax',
	                      0x21007: 'ebx',
	                      0x22007: 'ecx',
	                      0x23007: 'edx',
	                      0x24007: 'esi',
	                      0x25007: 'edi',
	                      0x26007: 'ebp',
	                      0x27007: 'esp',
	                      0x28007: 'r8d',
	                      0x29007: 'r9d',
	                      0x2A007: 'r10d',
	                      0x2B007: 'r11d',
	                      0x2C007: 'r12d',
	                      0x2D007: 'r13d',
	                      0x2E007: 'r14d',
	                      0x2F007: 'r15d'}

	def __init__(self, id = None):
		super(GeneralPurposeRegister32, self).__init__()
		if isinstance(id, str):
			if id in GeneralPurposeRegister32.name_to_number_map.iterkeys():
				self.number = GeneralPurposeRegister32.name_to_number_map[id]
				self.type = 'general-purpose'
				self.size = 32
			else:
				raise ValueError('Unknown register name: {0}'.format(id))
		elif isinstance(id, int):
			self.number = id
			self.type = 'general-purpose'
			self.size = 32
		elif isinstance(id, GeneralPurposeRegister32):
			self.number = id.number
			self.type = id.type
			self.size = id.size
		elif id is None:
			self.number = current_function.allocate_general_purpose_register32()
			self.type = 'general-purpose'
			self.size = 32
		else:
			raise TypeError('Register id is neither a name of an architectural general-purpose 32-bit register, nor an id of a virtual register')

	def __str__(self):
		if self.is_virtual():
			return 'gp32-vreg<{0}>'.format((self.number - 0x40000) >> 12)
		else:
			return GeneralPurposeRegister32.number_to_name_map[self.number]

	def is_extended(self):
		return self.number >= 0x28000

	def get_low_byte(self):
		return GeneralPurposeRegister8((self.number & -16) | 0x0001)

	def get_word(self):
		return GeneralPurposeRegister16((self.number & -16) | 0x0003)

	def get_dword(self):
		return GeneralPurposeRegister32((self.number & -16) | 0x0007)

	def get_qword(self):
		return GeneralPurposeRegister64((self.number & -16) | 0x000F)

eax  = GeneralPurposeRegister32('eax')
ebx  = GeneralPurposeRegister32('ebx')
ecx  = GeneralPurposeRegister32('ecx')
edx  = GeneralPurposeRegister32('edx')
esi  = GeneralPurposeRegister32('esi')
edi  = GeneralPurposeRegister32('edi')
ebp  = GeneralPurposeRegister32('ebp')
r8d  = GeneralPurposeRegister32('r8d')
r9d  = GeneralPurposeRegister32('r9d')
r10d = GeneralPurposeRegister32('r10d')
r11d = GeneralPurposeRegister32('r11d')
r12d = GeneralPurposeRegister32('r12d')
r13d = GeneralPurposeRegister32('r13d')
r14d = GeneralPurposeRegister32('r14d')
r15d = GeneralPurposeRegister32('r15d')

class GeneralPurposeRegister16(GeneralPurposeRegister):
	name_to_number_map = {'ax'  : 0x20003,
	                      'bx'  : 0x21003,
	                      'cx'  : 0x22003,
	                      'dx'  : 0x23003,
	                      'si'  : 0x24003,
	                      'di'  : 0x25003,
	                      'bp'  : 0x26003,
	                      'sp'  : 0x27003,
	                      'r8w' : 0x28003,
	                      'r9w' : 0x29003,
	                      'r10w': 0x2A003,
	                      'r11w': 0x2B003,
	                      'r12w': 0x2C003,
	                      'r13w': 0x2D003,
	                      'r14w': 0x2E003,
	                      'r15w': 0x2F003}

	number_to_name_map = {0x20003: 'ax',
	                      0x21003: 'bx',
	                      0x22003: 'cx',
	                      0x23003: 'dx',
	                      0x24003: 'si',
	                      0x25003: 'di',
	                      0x26003: 'bp',
	                      0x27003: 'sp',
	                      0x28003: 'r8w',
	                      0x29003: 'r9w',
	                      0x2A003: 'r10w',
	                      0x2B003: 'r11w',
	                      0x2C003: 'r12w',
	                      0x2D003: 'r13w',
	                      0x2E003: 'r14w',
	                      0x2F003: 'r15w'}

	def __init__(self, id = None):
		super(GeneralPurposeRegister16, self).__init__()
		if isinstance(id, str):
			if id in GeneralPurposeRegister16.name_to_number_map.iterkeys():
				self.number = GeneralPurposeRegister16.name_to_number_map[id]
				self.type = 'general-purpose'
				self.size = 16
			else:
				raise ValueError('Unknown register name: {0}'.format(id))
		elif isinstance(id, int):
			self.number = id
			self.type = 'general-purpose'
			self.size = 16
		elif isinstance(id, GeneralPurposeRegister16):
			self.number = id.number
			self.type = id.type
			self.size = id.size
		elif id is None:
			self.number = current_function.allocate_general_purpose_register16()
			self.type = 'general-purpose'
			self.size = 16
		else:
			raise TypeError('Register id is neither a name of an architectural general-purpose 16-bit register, nor an id of a virtual register')

	def __str__(self):
		if self.is_virtual():
			return 'gp16-vreg<{0}>'.format((self.number - 0x40000) >> 12)
		else:
			return GeneralPurposeRegister16.number_to_name_map[self.number]

	def is_extended(self):
		return self.number >= 0x28000

	def get_low_byte(self):
		return GeneralPurposeRegister8((self.number & -16) | 0x0001)

	def get_word(self):
		return GeneralPurposeRegister16((self.number & -16) | 0x0003)

	def get_dword(self):
		return GeneralPurposeRegister32((self.number & -16) | 0x0007)

	def get_qword(self):
		return GeneralPurposeRegister64((self.number & -16) | 0x000F)

ax   = GeneralPurposeRegister16('ax')
bx   = GeneralPurposeRegister16('bx')
cx   = GeneralPurposeRegister16('cx')
dx   = GeneralPurposeRegister16('dx')
si   = GeneralPurposeRegister16('si')
di   = GeneralPurposeRegister16('di')
bp   = GeneralPurposeRegister16('bp')
r8w  = GeneralPurposeRegister16('r8w')
r9w  = GeneralPurposeRegister16('r9w')
r10w = GeneralPurposeRegister16('r10w')
r11w = GeneralPurposeRegister16('r11w')
r12w = GeneralPurposeRegister16('r12w')
r13w = GeneralPurposeRegister16('r13w')
r14w = GeneralPurposeRegister16('r14w')
r15w = GeneralPurposeRegister16('r15w')

class GeneralPurposeRegister8(GeneralPurposeRegister):
	name_to_number_map = {'al'  : 0x20001,
	                      'ah'  : 0x20002,
	                      'bl'  : 0x21001,
	                      'bh'  : 0x21002,
	                      'cl'  : 0x22001,
	                      'ch'  : 0x22002,
	                      'dl'  : 0x23001,
	                      'dh'  : 0x23002,
	                      'sil' : 0x24001,
	                      'dil' : 0x25001,
	                      'bpl' : 0x26001,
	                      'spl' : 0x27001,
	                      'r8b' : 0x28001,
	                      'r9b' : 0x29001,
	                      'r10b': 0x2A001,
	                      'r11b': 0x2B001,
	                      'r12b': 0x2C001,
	                      'r13b': 0x2D001,
	                      'r14b': 0x2E001,
	                      'r15b': 0x2F001}

	number_to_name_map = {0x20001: 'al',
	                      0x20002: 'ah',
	                      0x21001: 'bl',
	                      0x21002: 'bh',
	                      0x22001: 'cl',
	                      0x22002: 'ch',
	                      0x23001: 'dl',
	                      0x23002: 'dh',
	                      0x24001: 'sil',
	                      0x25001: 'dil',
	                      0x26001: 'bpl',
	                      0x27001: 'spl',
	                      0x28001: 'r8b',
	                      0x29001: 'r9b',
	                      0x2A001: 'r10b',
	                      0x2B001: 'r11b',
	                      0x2C001: 'r12b',
	                      0x2D001: 'r13b',
	                      0x2E001: 'r14b',
	                      0x2F001: 'r15b'}

	def __init__(self, id = None):
		super(GeneralPurposeRegister8, self).__init__()
		if isinstance(id, str):
			if id in GeneralPurposeRegister8.name_to_number_map.iterkeys():
				self.number = GeneralPurposeRegister8.name_to_number_map[id]
				self.type = 'general-purpose'
				self.size = 8
			else:
				raise ValueError('Unknown register name: {0}'.format(id))
		elif isinstance(id, int):
			self.number = id
			self.type = 'general-purpose'
			self.size = 8
		elif isinstance(id, GeneralPurposeRegister8):
			self.number = id.number
			self.type = id.type
			self.size = id.size
		elif id is None:
			self.number = current_function.allocate_general_purpose_register8()
			self.type = 'general-purpose'
			self.size = 8
		else:
			raise TypeError('Register id is neither a name of an architectural general-purpose 8-bit register, nor an id of a virtual register')

	def __str__(self):
		if self.is_virtual():
			return 'gp8-vreg<{0}>'.format((self.number - 0x40000) >> 12)
		else:
			return GeneralPurposeRegister8.number_to_name_map[self.number]

	def is_extended(self):
		return self.number >= 0x28000

	def get_low_byte(self):
		return GeneralPurposeRegister8((self.number & -16) | 0x0001)

	def get_word(self):
		return GeneralPurposeRegister16((self.number & -16) | 0x0003)

	def get_dword(self):
		return GeneralPurposeRegister32((self.number & -16) | 0x0007)

	def get_qword(self):
		return GeneralPurposeRegister64((self.number & -16) | 0x000F)

al   = GeneralPurposeRegister8('al')
ah   = GeneralPurposeRegister8('ah')
bl   = GeneralPurposeRegister8('bl')
bh   = GeneralPurposeRegister8('bh')
cl   = GeneralPurposeRegister8('cl')
ch   = GeneralPurposeRegister8('ch')
dl   = GeneralPurposeRegister8('dl')
dh   = GeneralPurposeRegister8('dh')
sil  = GeneralPurposeRegister8('sil')
dil  = GeneralPurposeRegister8('dil')
bpl  = GeneralPurposeRegister8('bpl')
r8b  = GeneralPurposeRegister8('r8b')
r9b  = GeneralPurposeRegister8('r9b')
r10b = GeneralPurposeRegister8('r10b')
r11b = GeneralPurposeRegister8('r11b')
r12b = GeneralPurposeRegister8('r12b')
r13b = GeneralPurposeRegister8('r13b')
r14b = GeneralPurposeRegister8('r14b')
r15b = GeneralPurposeRegister8('r15b')

class MMXRegister(Register):
	name_to_number_map = {'mm0': 0x10010,
	                      'mm1': 0x11010,
	                      'mm2': 0x12010,
	                      'mm3': 0x13010,
	                      'mm4': 0x14010,
	                      'mm5': 0x15010,
	                      'mm6': 0x16010,
	                      'mm7': 0x17010}

	number_to_name_map = {0x10010: 'mm0',
	                      0x11010: 'mm1',
	                      0x12010: 'mm2',
	                      0x13010: 'mm3',
	                      0x14010: 'mm4',
	                      0x15010: 'mm5',
	                      0x16010: 'mm6',
	                      0x17010: 'mm7'}

	def __init__(self, id = None):
		super(MMXRegister, self).__init__()
		if isinstance(id, str):
			if id in MMXRegister.name_to_number_map.iterkeys():
				self.number = MMXRegister.name_to_number_map[id]
				self.type = 'mmx'
				self.size = 64
			else:
				raise ValueError('Unknown register name: {0}'.format(id))
		elif isinstance(id, int):
			self.number = id
			self.type = 'mmx'
			self.size = 64
		elif isinstance(id, MMXRegister):
			self.number = id.number
			self.type = id.type
			self.size = id.size
		elif id is None:
			self.number = current_function.allocate_mmx_register()
			self.type = 'mmx'
			self.size = 64
		else:
			raise TypeError('Register id is neither a name of an architectural mmx register, nor an id of a virtual register')

	def __str__(self):
		if self.is_virtual():
			return 'mm-vreg<{0}>'.format((self.number - 0x40000) >> 12)
		else:
			return MMXRegister.number_to_name_map[self.number]

mm0 = MMXRegister('mm0')
mm1 = MMXRegister('mm1')
mm2 = MMXRegister('mm2')
mm3 = MMXRegister('mm3')
mm4 = MMXRegister('mm4')
mm5 = MMXRegister('mm5')
mm6 = MMXRegister('mm6')
mm7 = MMXRegister('mm7')

class SSERegister(Register):
	name_to_number_map = {'xmm0' : 0x00040,
	                      'xmm1' : 0x01040,
	                      'xmm2' : 0x02040,
	                      'xmm3' : 0x03040,
	                      'xmm4' : 0x04040,
	                      'xmm5' : 0x05040,
	                      'xmm6' : 0x06040,
	                      'xmm7' : 0x07040,
	                      'xmm8' : 0x08040,
	                      'xmm9' : 0x09040,
	                      'xmm10': 0x0A040,
	                      'xmm11': 0x0B040,
	                      'xmm12': 0x0C040,
	                      'xmm13': 0x0D040,
	                      'xmm14': 0x0E040,
	                      'xmm15': 0x0F040}

	number_to_name_map = {0x00040: 'xmm0',
	                      0x01040: 'xmm1',
	                      0x02040: 'xmm2',
	                      0x03040: 'xmm3',
	                      0x04040: 'xmm4',
	                      0x05040: 'xmm5',
	                      0x06040: 'xmm6',
	                      0x07040: 'xmm7',
	                      0x08040: 'xmm8',
	                      0x09040: 'xmm9',
	                      0x0A040: 'xmm10',
	                      0x0B040: 'xmm11',
	                      0x0C040: 'xmm12',
	                      0x0D040: 'xmm13',
	                      0x0E040: 'xmm14',
	                      0x0F040: 'xmm15'}

	def __init__(self, id = None):
		super(SSERegister, self).__init__()
		if isinstance(id, str):
			if id in SSERegister.name_to_number_map.iterkeys():
				self.number = SSERegister.name_to_number_map[id]
				self.type = 'avx'
				self.size = 128
			else:
				raise ValueError('Unknown register name: {0}'.format(id))
		elif isinstance(id, int):
			self.number = id
			self.type = 'avx'
			self.size = 128
		elif isinstance(id, SSERegister):
			self.number = id.number
			self.type = id.type
			self.size = id.size
		elif id is None:
			self.number = current_function.allocate_sse_register()
			self.type = 'avx'
			self.size = 128
		else:
			raise TypeError('Register id is neither a name of an architectural sse register, nor an id of a virtual register')

	def is_extended(self):
		return self.number >= 0x08000

	def __str__(self):
		if self.is_virtual():
			return 'xmm-vreg<{0}>'.format((self.number - 0x40000) >> 12)
		else:
			return SSERegister.number_to_name_map[self.number]

	def get_oword(self):
		return SSERegister((self.number & -256) | 0x0040)

	def get_hword(self):
		return AVXRegister((self.number & -256) | 0x00C0)

xmm0  = SSERegister('xmm0')
xmm1  = SSERegister('xmm1')
xmm2  = SSERegister('xmm2')
xmm3  = SSERegister('xmm3')
xmm4  = SSERegister('xmm4')
xmm5  = SSERegister('xmm5')
xmm6  = SSERegister('xmm6')
xmm7  = SSERegister('xmm7')
xmm8  = SSERegister('xmm8')
xmm9  = SSERegister('xmm9')
xmm10 = SSERegister('xmm10')
xmm11 = SSERegister('xmm11')
xmm12 = SSERegister('xmm12')
xmm13 = SSERegister('xmm13')
xmm14 = SSERegister('xmm14')
xmm15 = SSERegister('xmm15')

class AVXRegister(Register):
	name_to_number_map = {'ymm0' : 0x000C0,
	                      'ymm1' : 0x010C0,
	                      'ymm2' : 0x020C0,
	                      'ymm3' : 0x030C0,
	                      'ymm4' : 0x040C0,
	                      'ymm5' : 0x050C0,
	                      'ymm6' : 0x060C0,
	                      'ymm7' : 0x070C0,
	                      'ymm8' : 0x080C0,
	                      'ymm9' : 0x090C0,
	                      'ymm10': 0x0A0C0,
	                      'ymm11': 0x0B0C0,
	                      'ymm12': 0x0C0C0,
	                      'ymm13': 0x0D0C0,
	                      'ymm14': 0x0E0C0,
	                      'ymm15': 0x0F0C0}

	number_to_name_map = {0x000C0: 'ymm0',
	                      0x010C0: 'ymm1',
	                      0x020C0: 'ymm2',
	                      0x030C0: 'ymm3',
	                      0x040C0: 'ymm4',
	                      0x050C0: 'ymm5',
	                      0x060C0: 'ymm6',
	                      0x070C0: 'ymm7',
	                      0x080C0: 'ymm8',
	                      0x090C0: 'ymm9',
	                      0x0A0C0: 'ymm10',
	                      0x0B0C0: 'ymm11',
	                      0x0C0C0: 'ymm12',
	                      0x0D0C0: 'ymm13',
	                      0x0E0C0: 'ymm14',
	                      0x0F0C0: 'ymm15'}

	def __init__(self, id = None):
		super(AVXRegister, self).__init__()
		if isinstance(id, str):
			if id in AVXRegister.name_to_number_map.iterkeys():
				self.number = AVXRegister.name_to_number_map[id]
				self.type = 'avx'
				self.size = 256
			else:
				raise ValueError('Unknown register name: {0}'.format(id))
		elif isinstance(id, int):
			self.number = id
			self.type = 'avx'
			self.size = 256
		elif isinstance(id, AVXRegister):
			self.number = id.number
			self.type = id.type
			self.size = id.size
		elif id is None:
			self.number = current_function.allocate_avx_register()
			self.type = 'avx'
			self.size = 256
		else:
			raise TypeError('Register id is neither a name of an architectural sse register, nor an id of a virtual register')

	def is_extended(self):
		return self.number >= 0x08000

	def __str__(self):
		if self.is_virtual():
			return 'ymm-vreg<{0}>'.format((self.number - 0x40000) >> 12)
		else:
			return AVXRegister.number_to_name_map[self.number]

	def get_oword(self):
		return SSERegister((self.number & -256) | 0x0040)

	def get_hword(self):
		return AVXRegister((self.number & -256) | 0x00C0)

ymm0  = AVXRegister('ymm0')
ymm1  = AVXRegister('ymm1')
ymm2  = AVXRegister('ymm2')
ymm3  = AVXRegister('ymm3')
ymm4  = AVXRegister('ymm4')
ymm5  = AVXRegister('ymm5')
ymm6  = AVXRegister('ymm6')
ymm7  = AVXRegister('ymm7')
ymm8  = AVXRegister('ymm8')
ymm9  = AVXRegister('ymm9')
ymm10 = AVXRegister('ymm10')
ymm11 = AVXRegister('ymm11')
ymm12 = AVXRegister('ymm12')
ymm13 = AVXRegister('ymm13')
ymm14 = AVXRegister('ymm14')
ymm15 = AVXRegister('ymm15')

class Operand(object):
	def __init__(self, operand):
		super(Operand, self).__init__()
		if isinstance(operand, Register):
			self.type = 'register'
			self.register = operand
		elif isinstance(operand, int) or isinstance(operand, long):
			if -9223372036854775808 <= operand <= 9223372036854775807:
				self.type = 'immediate'
				self.immediate = operand
			else:
				raise ValueError('The immediate operand {0} is not a 32-bit value'.format(operand))
		elif isinstance(operand, list):
			if len(operand) == 1:
				address = operand[0]
				self.type = 'memory'
				self.size = None
				if isinstance(address, GeneralPurposeRegister64):
					self.base = address
					self.scale = 0
					self.index = None
					self.offset = 0
				elif isinstance(operand[0], GeneralPurposeRegister64WithOffset):
					self.base = address.register
					self.scale = 0
					self.index = None
					self.offset = address.offset
				else:
					raise TypeError('Memory operand must be a list with register or register + offset')
			else:
				raise ValueError('Memory operand must be a list with only one element')
		elif isinstance(operand, MemoryOperand):
			address = operand.address
			self.type = 'memory'
			self.size = operand.size
			if isinstance(address, GeneralPurposeRegister64):
				self.base = address
				self.scale = 0
				self.index = None
				self.offset = 0
			elif isinstance(address, GeneralPurposeRegister64WithOffset):
				self.base = address.register
				self.scale = 0
				self.index = None
				self.offset = address.offset
			else:
				raise TypeError('Memory operand must be a list with register or register + offset')
		elif isinstance(operand, Constant):
			self.type = 'constant'
			self.constant = operand
			self.size = operand.size * operand.repeats
		elif isinstance(operand, LocalVariable):
			self.type = 'variable'
			self.variable = operand
			self.size = operand.size * 8
		elif isinstance(operand, str):
			self.type = 'label'
			self.label = operand
		elif operand is None:
			self.type = 'none'
		else:
			raise TypeError('The operand {0} is not a valid assembly instruction operand'.format(operand))

	def __str__(self):
		if self.is_constant():
			size_specifier = MemoryOperandSizeSpecifier.size_to_name_map[self.size]
			if self.constant.prefix is None:
				return size_specifier + "[rel {0}]".format(self.constant.label)
			else:
				return size_specifier + "[rel {1}.{0}]".format(self.constant.label, self.constant.prefix)
		elif self.is_local_variable():
			size_specifier = MemoryOperandSizeSpecifier.size_to_name_map[self.size] + " "
			if self.variable.offset == 0:
				return size_specifier + "[{0}]".format(self.variable.base)
			else:
				return size_specifier + "[{0} + {1}]".format(self.variable.base, self.variable.offset)
		elif self.is_memory_address():
			size_specifier = ""
			if not self.size is None:
				size_specifier = MemoryOperandSizeSpecifier.size_to_name_map[self.size] + " "
			if self.offset == 0:
				return size_specifier + "[{0}]".format(self.base)
			elif -128 <= self.offset <= 127:
				return size_specifier + "[byte {0} + {1}]".format(self.base, self.offset)
			else:
				return size_specifier + "[dword {0} + {1}]".format(self.base, self.offset)
		elif self.is_register():
			return str(self.register)
		elif self.is_label():
			return self.label
		elif self.is_immediate():
			return str(self.immediate)
		elif self.is_none():
			return ""
		else:
			raise TypeError('Unsupported operand type')

	def __eq__(self, other):
		if isinstance(other, Operand):
			if self.type == other.type:
				if self.is_immediate():
					return self.immediate == other.immediate
				elif self.is_register():
					return self.register == other.register
				elif self.is_memory_address():
					return self.base == other.base and self.offset == other.offset
				elif self.is_label():
					return self.label == other.label
				else:
					return False
			else:
				return False
		else:
			return False

	def is_none(self):
		return self.type == 'none'

	def is_immediate(self):
		return self.type == 'immediate'

	def is_int64(self):
		return self.type == 'immediate'

	def is_int32(self):
		if self.type == 'immediate':
			if -2147483648 <= self.immediate <= 4294967295:
				return True
			else:
				return False
		else:
			return False

	def is_sint32(self):
		if self.type == 'immediate':
			if -2147483648 <= self.immediate <= 2147483647:
				return True
			else:
				return False
		else:
			return False

	def is_int16(self):
		if self.type == 'immediate':
			if -32768 <= self.immediate <= 65535:
				return True
			else:
				return False
		else:
			return False

	def is_uint16(self):
		if self.type == 'immediate':
			if 0 <= self.immediate <= 65535:
				return True
			else:
				return False
		else:
			return False

	def is_int8(self):
		if self.type == 'immediate':
			if -128 <= self.immediate <= 255:
				return True
			else:
				return False
		else:
			return False

	def is_uint8(self):
		if self.type == 'immediate':
			if 0 <= self.immediate <= 255:
				return True
			else:
				return False
		else:
			return False

	def is_sint8(self):
		if self.type == 'immediate':
			if -128 <= self.immediate <= 127:
				return True
			else:
				return False
		else:
			return False

	def is_label(self):
		return self.type == 'label'

	def is_constant(self):
		return self.type == 'constant'

	def is_local_variable(self):
		return self.type == 'variable'

	def is_register(self):
		return self.type == 'register'

	def is_general_purpose_register(self):
		if self.type == 'register':
			if isinstance(self.register, GeneralPurposeRegister):
				return True
			else:
				return False
		else:
			return False

	def is_general_purpose_register64(self):
		if self.type == 'register':
			if isinstance(self.register, GeneralPurposeRegister64):
				return True
			else:
				return False
		else:
			return False

	def is_general_purpose_register32(self):
		if self.type == 'register':
			if isinstance(self.register, GeneralPurposeRegister32):
				return True
			else:
				return False
		else:
			return False

	def is_general_purpose_register16(self):
		if self.type == 'register':
			if isinstance(self.register, GeneralPurposeRegister16):
				return True
			else:
				return False
		else:
			return False

	def is_general_purpose_register8(self):
		if self.type == 'register':
			if isinstance(self.register, GeneralPurposeRegister8):
				return True
			else:
				return False
		else:
			return False

	def is_mmx_register(self):
		if self.type == 'register':
			if isinstance(self.register, MMXRegister):
				return True
			else:
				return False
		else:
			return False

	def is_sse_register(self):
		if self.type == 'register':
			if isinstance(self.register, SSERegister):
				return True
			else:
				return False
		else:
			return False

	def is_avx_register(self):
		if self.type == 'register':
			if isinstance(self.register, AVXRegister):
				return True
			else:
				return False
		else:
			return False

	def is_memory_address(self):
		return self.type == 'memory' or self.type == 'constant' or self.type == 'variable'

	def is_memory_address256(self, strict_size = False):
		if self.type == 'memory':
			if self.size is None:
				if strict_size:
					return False
				else:
					return True
			else:
				return self.size == 256
		elif self.type == 'constant':
			return self.size == 256
		elif self.type == 'variable':
			return self.size == 256
		else:
			return False

	def is_memory_address128(self, strict_size = False):
		if self.type == 'memory':
			if self.size is None:
				if strict_size:
					return False
				else:
					return True
			else:
				return self.size == 128
		elif self.type == 'constant':
			return self.size == 128
		elif self.type == 'variable':
			return self.size == 128
		else:
			return False

	def is_memory_address80(self, strict_size = False):
		if self.type == 'memory':
			if self.size is None:
				if strict_size:
					return False
				else:
					return True
			else:
				return self.size == 80
		elif self.type == 'constant':
			return self.size == 80
		else:
			return False

	def is_memory_address64(self, strict_size = False):
		if self.type == 'memory':
			if self.size is None:
				if strict_size:
					return False
				else:
					return True
			else:
				return self.size == 64
		elif self.type == 'constant':
			if strict_size:
				return self.size == 64
			else:
				return True
		elif self.type == 'variable':
			if strict_size:
				return self.size == 64
			else:
				return True
		else:
			return False

	def is_memory_address32(self, strict_size = False):
		if self.type == 'memory':
			if self.size is None:
				if strict_size:
					return False
				else:
					return True
			else:
				return self.size == 32
		elif self.type == 'constant':
			if strict_size:
				return self.size == 32
			else:
				return True
		elif self.type == 'variable':
			if strict_size:
				return self.size == 32
			else:
				return True
		else:
			return False

	def is_memory_address16(self, strict_size = False):
		if self.type == 'memory':
			if self.size is None:
				if strict_size:
					return False
				else:
					return True
			else:
				return self.size == 16
		elif self.type == 'constant':
			if strict_size:
				return self.size == 16
			else:
				return True
		elif self.type == 'variable':
			if strict_size:
				return self.size == 16
			else:
				return True
		else:
			return False

	def is_memory_address8(self, strict_size = False):
		if self.type == 'memory':
			if self.size is None:
				if strict_size:
					return False
				else:
					return True
			else:
				return self.size == 8
		elif self.type == 'constant':
			if strict_size:
				return self.size == 8
			else:
				return True
		elif self.type == 'variable':
			if strict_size:
				return self.size == 8
			else:
				return True
		else:
			return False

	def get_modrm_extra_length(self):
		if self.is_register():
			return 0 # encoded in ModR/M, no extra bytes needed
		elif self.is_constant():
			return 4 # 4 bytes for offset
		elif self.is_local_variable():
			return 1 + 4 # rsp-based 32-bit offset
		elif self.is_memory_address():
			if self.offset == 0:
				if self.base == rbp or self.base == r13:
					return 1 # rbp + 0 or r13 + 0 is encoded
				elif self.base == rsp or self.base == r12:
					return 1 # rsp and r12 are encoded in SIB
				else:
					return 0 # encoded in ModR/M, no extra bytes needed
			elif -128 <= self.offset <= 127:
				if self.base == rsp or self.base == r12:
					return 1 + 1 # rsp and r12 Are encoded in SIB + 1 byte for offset
				else:
					return 1 # 1 byte for offset
			else:
				if self.base == rsp or self.base == r12:
					return 4 + 1 # rsp and r12 are encoded in SIB + 4 bytes for offset
				else:
					return 4 # 4 bytes for offset
		else:
			raise ValueError('The operand can not be encoded in ModR/M')

	def get_vex_extra_length(self):
		return 2

	def get_registers_list(self):
		if self.is_register():
			return [self.register]
		elif self.is_constant():
			return list()
		elif self.is_local_variable():
			return [rsp]
		elif self.is_memory_address():
			return [self.base]
		else:
			return list()

class MemoryOperand(object):
	def __init__(self, address, size):
		super(MemoryOperand, self).__init__()
		check_generic_memory_operand(address)
		self.address = address
		if isinstance(size, int):
			self.size = size
		else:
			raise TypeError('The size of the memory operand must be an integer')

	def __str__(self):
		return str(Operand(self))

class MemoryOperandSizeSpecifier(object):
	def __init__(self, size):
		super(MemoryOperandSizeSpecifier, self).__init__()
		if isinstance(size, int):
			if size in self.size_to_name_map.iterkeys():
				self.size = size
			else:
				raise ValueError('The size of memory operand size specifier ({0}) is not in the list of supported memory operand size specifiers ({1})',
					size, ", ".join(map(str, self.size_to_name_map.iterkeys())))
		else:
			raise TypeError('The size of memory operand size specifier must be an integer')

	def __str__(self):
		return self.size_to_name_map[self.size]

	def __getitem__(self, address):
		if isinstance(address, GeneralPurposeRegister64):
			return MemoryOperand(address, self.size)
		elif isinstance(address, GeneralPurposeRegister64WithOffset):
			return MemoryOperand(address, self.size)
		else:
			raise TypeError('Memory address must be either a register or register + offset')

	size_to_name_map = {8: 'byte', 16: 'word', 32: 'dword', 64: 'qword', 80: 'tword', 128: 'oword', 256: 'hword'}


byte = MemoryOperandSizeSpecifier(8)
word16 = MemoryOperandSizeSpecifier(16)
word = MemoryOperandSizeSpecifier(16)
word32 = MemoryOperandSizeSpecifier(32)
dword = MemoryOperandSizeSpecifier(32)
word64 = MemoryOperandSizeSpecifier(64)
qword = MemoryOperandSizeSpecifier(64)
word80 = MemoryOperandSizeSpecifier(80)
tword = MemoryOperandSizeSpecifier(80)
word128 = MemoryOperandSizeSpecifier(128)
oword = MemoryOperandSizeSpecifier(128)
word256 = MemoryOperandSizeSpecifier(256)
hword = MemoryOperandSizeSpecifier(256)

def check_generic_memory_operand(operand):
	if isinstance(operand, list):
		if len(operand) == 1:
			if isinstance(operand[0], GeneralPurposeRegister64):
				return
			elif isinstance(operand[0], GeneralPurposeRegister64WithOffset):
				return
			else:
				raise TypeError('Memory operand must be a list with register or register + offset')
		else:
			raise ValueError('Memory operand must be a list with only one element')
	else:
		return

class QuasiInstruction(object):
	def __init__(self, name):
		super(QuasiInstruction, self).__init__()
		self.name = name
		self.stack_offset = None

class Instruction(QuasiInstruction):
	def __init__(self, name, isa_extension = None):
		super(Instruction, self).__init__(name)
		if(not isa_extension is None) and (not isa_extension in supported_isa_extensions):
			raise ValueError('Instruction ISA extension {0} in not in the supported ISA extensions list ({1})'.format(isa_extension, ", ".join(supported_isa_extensions)))
		self.isa_extension = isa_extension
		self.short_eax_form = False
		self.always_has_rex_prefix = False
		self.available_registers = set()
		self.live_registers = set()

	def __len__(self):
		return self.size

	def __str__(self):
		return str(self.name)

	def get_isa_extension(self):
		return self.isa_extension

class BinaryInstruction(Instruction):
	def __init__(self, name, destination, source, isa_extension = None):
		super(BinaryInstruction, self).__init__(name, isa_extension)
		self.name = name
		self.destination = destination
		self.source = source

	def __str__(self):
		return "{0} {1}, {2}".format(self.name, self.destination, self.source)

class Label(QuasiInstruction):
	def __init__(self, name):
		super(Label, self).__init__('<LABEL>')
		self.name = name
		self.input_branches = set()

	def __str__(self):
		return "." + self.name + ':'

class AlignQuasiInstruction(QuasiInstruction):
	supported_alignments = [2, 4, 8, 16, 32]

	def __init__(self, alignment):
		super(AlignQuasiInstruction, self).__init__('<ALIGN>')
		if isinstance(alignment, int):
			if alignment in AlignQuasiInstruction.supported_alignments:
				self.alignment = alignment
			else:
				raise ValueError("The alignment value {0} is not in the list of supported alignments ({1})".format(alignment, ", ".join(AlignQuasiInstruction.supported_alignments)))
		else:
			raise TypeError("The alignment value must be an integer")

	def __str__(self):
		return "align {0}".format(self.alignment)

class LoadConstantPseudoInstruction(Instruction):
	def __init__(self, destination, source):
		super(LoadConstantPseudoInstruction, self).__init__('<LOAD-CONSTANT>')
		if destination.is_register():
			self.destination = destination
		else:
			raise ValueError('Load constant pseudo-instruction expects a register as a destination')
		if source.is_constant():
			if destination.register.size == source.constant.size * source.constant.repeats:
				self.source = source
			else:
				raise ValueError('The size of constant should be the same as the size of register')
		else:
			raise ValueError('Load constant pseudo-instruction expects a Constant instance as a source')
		self.size = 4 + 4

	def __str__(self):
		return "LOAD.CONSTANT {0} = {1}".format(self.destination, self.source)

	def get_input_registers_list(self):
		return list()

	def get_output_registers_list(self):
		return [self.destination.register]

	def get_constant(self):
		return self.source.constant

	def get_local_variable(self):
		return None

class LoadParameterPseudoInstruction(Instruction):
	def __init__(self, destination, source):
		super(LoadParameterPseudoInstruction, self).__init__('<LOAD-PARAMETER>')
		if destination.is_general_purpose_register64():
			self.destination = destination
		else:
			raise ValueError('Load parameter pseudo-instruction expects a general-purpose 32-bit register as a destination')
		if isinstance(source, yeppp.codegen.c.Parameter):
			parameter = current_function.find_parameter(source)
			if parameter is not None:
				self.parameter = parameter
			else:
				raise ValueError('{0} is not an argument of the currently active function'.format(source))
		else:
			raise TypeError('Load parameter pseudo-instruction expects a Parameter object as a source')
		# MOV r32/m32, r32: 8B /r
		self.size = 1 + 2 + 1 + 1 # REX prefix + 1-byte opcode + ModR/M + SIB (rsp addressing) + 8-bit offset

	def __str__(self):
		return "LOAD.PARAMETER {0} = {1}".format(self.destination, self.parameter)

	def get_input_registers_list(self):
		if self.parameter.register_location:
			return [self.parameter.register_location]
		else:
			return [rsp]

	def get_output_registers_list(self):
		return [self.destination.register]

	def get_constant(self):
		return None

	def get_local_variable(self):
		return None

class ArithmeticInstruction(BinaryInstruction):
	def __init__(self, name, destination, source):
		super(ArithmeticInstruction, self).__init__(name, destination, source)
		allowed_instructions = ['ADD', 'ADC', 'SUB', 'SBB', 'CMP', 'TEST', 'AND', 'OR', 'XOR']
		if not name in allowed_instructions:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})', name, ", ".join(allowed_instructions))
		if name != 'TEST' and destination.is_general_purpose_register64() and (source.is_general_purpose_register64() or source.is_memory_address64()):
			# ADD r64, r64/m64: REX.W 03 /r
			# ADC r64, r64/m64: REX.W 13 /r
			# SUB r64, r64/m64: REX.W 2B /r
			# SBB r64, r64/m64: REX.W 1B /r
			# CMP r64, r64/m64: REX.W 3B /r
			# AND r64, r64/m64: REX.W 23 /r
			# OR  r64, r64/m64: REX.W 0B /r
			# XOR r64, r64/m64: REX.W 33 /r
			self.size = 1 + 2 + source.get_modrm_extra_length()
			self.always_has_rex_prefix = True
		elif destination.is_general_purpose_register32() and (source.is_general_purpose_register32() or source.is_memory_address32()):
			# ADD r32, r32/m32: 03 /r
			# ADC r32, r32/m32: 13 /r
			# SUB r32, r32/m32: 2B /r
			# SBB r32, r32/m32: 1B /r
			# CMP r32, r32/m32: 3B /r
			# AND r32, r32/m32: 23 /r
			# OR  r32, r32/m32: 0B /r
			# XOR r32, r32/m32: 33 /r
			self.size = 2 + source.get_modrm_extra_length()
		elif destination.is_general_purpose_register16() and (source.is_general_purpose_register16() or source.is_memory_address16()):
			# ADD r16, r16/m16: 66 03 /r
			# ADC r16, r16/m16: 66 13 /r
			# SUB r16, r16/m16: 66 2B /r
			# SBB r16, r16/m16: 66 1B /r
			# CMP r16, r16/m16: 66 3B /r
			# AND r16, r16/m16: 66 23 /r
			# OR  r16, r16/m16: 66 0B /r
			# XOR r16, r16/m16: 66 33 /r
			self.size = 1 + 2 + source.get_modrm_extra_length()
		elif destination.is_general_purpose_register8() and (source.is_general_purpose_register8() or source.is_memory_address8()):
			# ADD r8, r8/m8: 03 /r
			# ADC r8, r8/m8: 13 /r
			# SUB r8, r8/m8: 2B /r
			# SBB r8, r8/m8: 1B /r
			# CMP r8, r8/m8: 3B /r
			# AND r8, r8/m8: 22 /r
			# OR  r8, r8/m8: 0A /r
			# XOR r8, r8/m8: 32 /r
			self.size = 2 + source.get_modrm_extra_length()
		elif destination.is_general_purpose_register64() and source.is_sint32():
			if name != 'TEST' and source.is_sint8():
				# ADD r64/m64, imm8: REX.W 83 /0 ib
				# ADC r64/m64, imm8: REX.W 83 /2 ib
				# SUB r64/m64, imm8: REX.W 83 /5 ib
				# SBB r64/m64, imm8: REX.W 83 /3 ib
				# CMP r64/m64, imm8: REX.W 83 /7 ib
				# AND r64/m64, imm8: REX.W 83 /4 ib
				# OR  r64/m64, imm8: REX.W 83 /1 ib
				# XOR r64/m64, imm8: REX.W 83 /6 ib
				self.size = 1 + 2 + 1 + destination.get_modrm_extra_length()
			else:
				if destination == rax:
					# ADD  rax, imm32: REX.W 05 id
					# ADC  rax, imm32: REX.W 15 id
					# SUB  rax, imm32: REX.W 2D id
					# SBB  rax, imm32: REX.W 1D id
					# CMP  rax, imm32: REX.W 3D id
					# AND  rax, imm32: REX.W 25 id
					# OR   rax, imm32: REX.W 0D id
					# XOR  rax, imm32: REX.W 35 id
					# TEST rax, imm32: REX.W A9 id
					self.size = 1 + 1 + 4
				else:
					# ADD  r64/m64, imm32: REX.W 81 /0 id
					# ADC  r64/m64, imm32: REX.W 81 /2 id
					# SUB  r64/m64, imm32: REX.W 81 /5 id
					# SBB  r64/m64, imm32: REX.W 81 /3 id
					# CMP  r64/m64, imm32: REX.W 81 /7 id
					# AND  r64/m64, imm32: REX.W 81 /4 id
					# OR   r64/m64, imm32: REX.W 81 /1 id
					# XOR  r64/m64, imm32: REX.W 81 /6 id
					# TEST r64/m64, imm32: REX.W F7 /0 id
					self.size = 1 + 2 + 4
			self.always_has_rex_prefix = True
			self.short_eax_form = name != 'TEST'
		elif destination.is_general_purpose_register32() and source.is_int32():
			if name != 'TEST' and source.is_sint8():
				# ADD r32/m32, imm8: 83 /0 ib
				# ADC r32/m32, imm8: 83 /2 ib
				# SUB r32/m32, imm8: 83 /5 ib
				# SBB r32/m32, imm8: 83 /3 ib
				# CMP r32/m32, imm8: 83 /7 ib
				# AND r32/m32, imm8: 83 /4 ib
				# OR  r32/m32, imm8: 83 /1 ib
				# XOR r32/m32, imm8: 83 /6 ib
				self.size = 2 + 1 + destination.get_modrm_extra_length()
			else:
				if destination == eax:
					# ADD  eax, imm32: 05 id
					# ADC  eax, imm32: 15 id
					# SUB  eax, imm32: 2D id
					# SBB  eax, imm32: 1D id
					# CMP  eax, imm32: 3D id
					# AND  eax, imm32: 25 id
					# OR   eax, imm32: 0D id
					# XOR  eax, imm32: 35 id
					# TEST eax, imm32: A9 id
					self.size = 1 + 4
				else:
					# ADD  r32/m32, imm32: 81 /0 id
					# ADC  r32/m32, imm32: 81 /2 id
					# SUB  r32/m32, imm32: 81 /5 id
					# SBB  r32/m32, imm32: 81 /3 id
					# CMP  r32/m32, imm32: 81 /7 id
					# AND  r32/m32, imm32: 81 /4 id
					# OR   r32/m32, imm32: 81 /1 id
					# XOR  r32/m32, imm32: 81 /6 id
					# TEST r32/m32, imm32: F7 /0 id
					self.size = 2 + 4
			self.short_eax_form = name != 'TEST'
		elif destination.is_general_purpose_register16() and source.is_int16():
			if name != 'TEST' and source.is_sint8():
				# ADD r16/m16, imm8: 66 83 /0 ib
				# ADC r16/m16, imm8: 66 83 /2 ib
				# SUB r16/m16, imm8: 66 83 /5 ib
				# SBB r16/m16, imm8: 66 83 /3 ib
				# CMP r16/m16, imm8: 66 83 /7 ib
				# AND r16/m16, imm8: 66 83 /4 ib
				# OR  r16/m16, imm8: 66 83 /1 ib
				# XOR r16/m16, imm8: 66 83 /6 ib
				self.size = 1 + 2 + 1 + destination.get_modrm_extra_length()
			else:
				if destination == ax:
					# ADD  ax, imm16: 66 05 iw
					# ADC  ax, imm16: 66 15 iw
					# SUB  ax, imm16: 66 2D iw
					# SBB  ax, imm16: 66 1D iw
					# CMP  ax, imm16: 66 3D iw
					# AND  ax, imm16: 66 25 iw
					# OR   ax, imm16: 66 0D iw
					# XOR  ax, imm16: 66 35 iw
					# TEST ax, imm16: 66 A9 iw
					self.size = 1 + 1 + 2
				else:
					# ADD  r16/m16, imm16: 66 81 /0 iw
					# ADC  r16/m16, imm16: 66 81 /2 iw
					# SUB  r16/m16, imm16: 66 81 /5 iw
					# SBB  r16/m16, imm16: 66 81 /3 iw
					# CMP  r16/m16, imm16: 66 81 /7 iw
					# AND  r16/m16, imm16: 66 81 /4 iw
					# OR   r16/m16, imm16: 66 81 /1 iw
					# XOR  r16/m16, imm16: 66 81 /6 iw
					# TEST r16/m16, imm16: 66 F7 /0 iw
					self.size = 1 + 2 + 2
			self.short_eax_form = name != 'TEST'
		elif destination.is_general_purpose_register8() and source.is_int8():
			if destination == al:
				# ADD  al, imm8: 05 ib
				# ADC  al, imm8: 15 ib
				# SUB  al, imm8: 2D ib
				# SBB  al, imm8: 1D ib
				# CMP  al, imm8: 3D ib
				# AND  al, imm8: 25 ib
				# OR   al, imm8: 0D ib
				# XOR  al, imm8: 34 ib
				# TEST al, imm8: A8 ib
				self.size = 1 + 1
			else:
				# ADD  r8/m8, imm8: 80 /0 ib
				# ADC  r8/m8, imm8: 80 /2 ib
				# SUB  r8/m8, imm8: 80 /5 ib
				# SBB  r8/m8, imm8: 80 /3 ib
				# CMP  r8/m8, imm8: 80 /7 ib
				# AND  r8/m8, imm8: 80 /4 ib
				# OR   r8/m8, imm8: 80 /1 ib
				# XOR  r8/m8, imm8: 80 /6 ib
				# TEST r8/m8, imm8: F6 /0 ib
				self.size = 2 + 1
			self.short_eax_form = True
		else:
			raise ValueError('Invalid operands in instruction {0} {1}, {2}'.format(name, destination, source))

	def get_input_registers_list(self):
		if self.name in ['XOR', 'SUB'] and self.source == self.destination:
			return list()
		else:
			input_registers_list = list()
			input_registers_list.extend(self.destination.get_registers_list())
			input_registers_list.extend(self.source.get_registers_list())
			return input_registers_list

	def get_output_registers_list(self):
		if self.name in ['CMP', 'TEST']:
			return list()
		else:
			if self.destination.is_register():
				return [self.destination.register]
			else:
				return list()

	def get_constant(self):
		if self.source.is_constant():
			return self.source.constant
		else:
			return None

	def get_local_variable(self):
		if self.destination.is_local_variable():
			return self.destination.variable
		elif self.source.is_local_variable():
			return self.source.variable
		else:
			return None

class ShiftInstruction(BinaryInstruction):
	def __init__(self, name, destination, source):
		super(ShiftInstruction, self).__init__(name, destination, source)
		allowed_instructions = ['SHR', 'SAR', 'SHL', 'ROR', 'ROL', 'ROR', 'RCL', 'RCR']
		if not self.name in allowed_instructions:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})', name, ", ".join(allowed_instructions))
		if (destination.is_general_purpose_register8() or destination.is_memory_address8()) and source.is_int8():
			if source.immediate == 1:
				# ROL r8/m8, 1: D0 /0
				# ROR r8/m8, 1: D0 /1
				# RCL r8/m8, 1: D0 /2
				# RCR r8/m8, 1: D0 /3
				# SHL r8/m8, 1: D0 /4
				# SHR r8/m8, 1: D0 /5
				# SAR r8/m8, 1: D0 /7
				self.size = 2 + destination.get_modrm_extra_length()
			else:
				# ROL r8/m8, imm8: C0 /0 ib
				# ROR r8/m8, imm8: C0 /1 ib
				# RCL r8/m8, imm8: C0 /2 ib
				# RCR r8/m8, imm8: C0 /3 ib
				# SHL r8/m8, imm8: C0 /4 ib
				# SHR r8/m8, imm8: C0 /5 ib
				# SAR r8/m8, imm8: C0 /7 ib
				self.size = 2 + 1 + destination.get_modrm_extra_length()
		elif (destination.is_general_purpose_register16() or destination.is_memory_address16()) and source.is_int8():
			if source.immediate == 1:
				# ROL r16/m16, 1: 66 D1 /0
				# ROR r16/m16, 1: 66 D1 /1
				# RCL r16/m16, 1: 66 D1 /2
				# RCR r16/m16, 1: 66 D1 /3
				# SHL r16/m16, 1: 66 D1 /4
				# SHR r16/m16, 1: 66 D1 /5
				# SAR r16/m16, 1: 66 D1 /7
				self.size = 1 + 2 + destination.get_modrm_extra_length()
			else:
				# ROL r16/m16, imm8: 66 C1 /0 ib
				# ROR r16/m16, imm8: 66 C1 /1 ib
				# RCL r16/m16, imm8: 66 C1 /2 ib
				# RCR r16/m16, imm8: 66 C1 /3 ib
				# SHL r16/m16, imm8: 66 C1 /4 ib
				# SHR r16/m16, imm8: 66 C1 /5 ib
				# SAR r16/m16, imm8: 66 C1 /7 ib
				self.size = 1 + 2 + 1 + destination.get_modrm_extra_length()
		elif (destination.is_general_purpose_register32() or destination.is_memory_address32()) and source.is_int8():
			if source.immediate == 1:
				# ROL r32/m32, 1: D1 /0
				# ROR r32/m32, 1: D1 /1
				# RCL r32/m32, 1: D1 /2
				# RCR r32/m32, 1: D1 /3
				# SHL r32/m32, 1: D1 /4
				# SHR r32/m32, 1: D1 /5
				# SAR r32/m32, 1: D1 /7
				self.size = 2 + destination.get_modrm_extra_length()
			else:
				# ROL r32/m32, imm8: C1 /0 ib
				# ROR r32/m32, imm8: C1 /1 ib
				# RCL r32/m32, imm8: C1 /2 ib
				# RCR r32/m32, imm8: C1 /3 ib
				# SHL r32/m32, imm8: C1 /4 ib
				# SHR r32/m32, imm8: C1 /5 ib
				# SAR r32/m32, imm8: C1 /7 ib
				self.size = 2 + 1 + destination.get_modrm_extra_length()
		elif (destination.is_general_purpose_register8() or destination.is_memory_address8()) and source.is_general_purpose_register():
			if source.register == rcx or source.register == ecx or source.register == cx or source.register == cl or source.register.is_virtual():
				# ROL r8/m8, cl: D2 /0 ib
				# ROR r8/m8, cl: D2 /1 ib
				# RCL r8/m8, cl: D2 /2 ib
				# RCR r8/m8, cl: D2 /3 ib
				# SHL r8/m8, cl: D2 /4 ib
				# SHR r8/m8, cl: D2 /5 ib
				# SAR r8/m8, cl: D2 /7 ib
				self.size = 2 + 1 + destination.get_modrm_extra_length()
			else:
				raise ValueError('Invalid operands in instruction {0} {1}, {2}'.format(name, destination, source))
		elif (destination.is_general_purpose_register16() or destination.is_memory_address16()) and source.is_general_purpose_register():
			if source.register == rcx or source.register == ecx or source.register == cx or source.register == cl or source.register.is_virtual():
				# ROL r16/m16, cl: 66 D3 /0
				# ROR r16/m16, cl: 66 D3 /1
				# RCL r16/m16, cl: 66 D3 /2
				# RCR r16/m16, cl: 66 D3 /3
				# SHL r16/m16, cl: 66 D3 /4
				# SHR r16/m16, cl: 66 D3 /5
				# SAR r16/m16, cl: 66 D3 /7
				self.size = 1 + 2 + destination.get_modrm_extra_length()
			else:
				raise ValueError('Invalid operands in instruction {0} {1}, {2}'.format(name, destination, source))
		elif (destination.is_general_purpose_register32() or destination.is_memory_address32()) and source.is_general_purpose_register():
			if source.register == rcx or source.register == ecx or source.register == cx or source.register == cl or source.register.is_virtual():
				# RCL r32/m32, cl: D3 /2
				# RCR r32/m32, cl: D3 /3
				# SHL r32/m32, cl: D3 /4
				# SHR r32/m32, cl: D3 /5
				# SAR r32/m32, cl: D3 /7
				self.size = 2 + destination.get_modrm_extra_length()
			else:
				raise ValueError('Invalid operands in instruction {0} {1}, {2}'.format(name, destination, source))
		else:
			raise ValueError('Invalid operands in instruction {0} {1}, {2}'.format(name, destination, source))

	def get_input_registers_list(self):
		return self.destination.get_registers_list() + self.source.get_registers_list()

	def get_output_registers_list(self):
		if self.destination.is_register():
			return [self.destination.register]
		else:
			return list()

	def get_constant(self):
		return None

	def get_local_variable(self):
		if self.destination.is_local_variable():
			return self.destination.variable
		else:
			return None

class BitTestInstruction(BinaryInstruction):
	def __init__(self, name, destination, source):
		super(BitTestInstruction, self).__init__(name,  destination, source)
		allowed_instructions = ['BT', 'BTS', 'BTR', 'BTC']
		if not name in allowed_instructions:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})', name, ", ".join(allowed_instructions))
		if (destination.is_general_purpose_register32() or destination.is_memory_address32()) and source.is_general_purpose_register32():
			#  BT m32/r32, r32: 0F A3 /r
			# BTS m32/r32, r32: 0F AB /r
			# BTR m32/r32, r32: 0F B3 /r
			# BTC m32/r32, r32: 0F BB /r
			self.size = 3 + destination.get_modrm_extra_length()
		elif (destination.is_general_purpose_register16() or destination.is_memory_address16()) and source.is_general_purpose_register16():
			#  BT m16/r16, r16: 66 0F A3 /r
			# BTS m16/r16, r16: 66 0F AB /r
			# BTR m16/r16, r16: 66 0F B3 /r
			# BTC m16/r16, r16: 66 0F BB /r
			self.size = 1 + 3 + destination.get_modrm_extra_length()
		elif (destination.is_general_purpose_register32() or destination.is_memory_address32()) and source.is_uint8():
			#  BT m32/r32, imm8: 0F BA /4 ib
			# BTS m32/r32, imm8: 0F BA /5 ib
			# BTR m32/r32, imm8: 0F BA /6 ib
			# BTC m32/r32, imm8: 0F BA /7 ib
			self.size = 3 + 1 + destination.get_modrm_extra_length()
		elif (destination.is_general_purpose_register16() or destination.is_memory_address16()) and source.is_uint8():
			#  BT m16/r16, imm8: 66 0F BA /4 ib
			# BTS m16/r16, imm8: 66 0F BA /5 ib
			# BTR m16/r16, imm8: 66 0F BA /6 ib
			# BTC m16/r16, imm8: 66 0F BA /7 ib
			self.size = 1 + 3 + 1 + destination.get_modrm_extra_length()
		else:
			raise ValueError('Invalid operands in instruction {0} {1}, {2}'.format(name, destination, source))

	def get_input_registers_list(self):
		return self.destination.get_registers_list() + self.source.get_registers_list()

	def get_output_registers_list(self):
		if self.destination.is_general_purpose_register() and self.name != "BT":
			return [self.destination.register]
		else:
			return list()

	def get_constant(self):
		return None

	def get_local_variable(self):
		if self.destination.is_local_variable():
			return self.destination.variable
		else:
			return None

class ImulInstruction(Instruction):
	def __init__(self, destination, source, immediate):
		super(ImulInstruction, self).__init__("IMUL")
		if (destination.is_general_purpose_register8() or destination.is_memory_address8(strict_size = True)) and source.is_none() and immediate.is_none():
			# IMUL r8/m8: F6 /5
			self.operands = 1
			self.source = destination
			self.size = 2 + self.source.get_modrm_extra_length()
			self.output_registers_list = [ax]
			self.input_registers_list = [al]
			self.input_registers_list.append(self.source.get_registers_list())
		elif (destination.is_general_purpose_register16() or destination.is_memory_address16(strict_size = True)) and source.is_none() and immediate.is_none():
			# IMUL r16/m16: 66 F7 /5
			self.operands = 1
			self.source = destination
			self.size = 1 + 2 + self.source.get_modrm_extra_length()
			self.output_registers_list = [dx, ax]
			self.input_registers_list = [ax]
			self.input_registers_list.append(self.source.get_registers_list())
		elif (destination.is_general_purpose_register32() or destination.is_memory_address32(strict_size = True)) and source.is_none() and immediate.is_none():
			# IMUL r32/m32: F7 /5
			self.operands = 1
			self.source = destination
			self.size = 2 + self.source.get_modrm_extra_length()
			self.output_registers_list = [edx, eax]
			self.input_registers_list = [eax]
			self.input_registers_list.append(self.source.get_registers_list())
		elif (destination.is_general_purpose_register64() or destination.is_memory_address64(strict_size = True)) and source.is_none() and immediate.is_none():
			# IMUL r64/m64: REX.W F7 /5
			self.operands = 1
			self.source = destination
			self.size = 1 + 2 + self.source.get_modrm_extra_length()
			self.output_registers_list = [rdx, rax]
			self.input_registers_list = [rax]
			self.input_registers_list.append(self.source.get_registers_list())
		elif destination.is_general_purpose_register16() and (source.is_general_purpose_register16() or source.is_memory_address16()) and immediate.is_none():
			# IMUL r16, r16/m16: 66 0F AF /r
			self.operands = 2
			self.source = source
			self.destination = destination
			self.size = 1 + 3 + source.get_modrm_extra_length()
			self.output_registers_list = destination.get_registers_list()
			self.input_registers_list = destination.get_registers_list() + source.get_registers_list()
		elif destination.is_general_purpose_register32() and (source.is_general_purpose_register32() or source.is_memory_address32()) and immediate.is_none():
			# IMUL r32, r32/m32: 0F AF /r
			self.operands = 2
			self.source = source
			self.destination = destination
			self.size = 3 + source.get_modrm_extra_length()
			self.output_registers_list = destination.get_registers_list()
			self.input_registers_list = destination.get_registers_list() + source.get_registers_list()
		elif destination.is_general_purpose_register64() and (source.is_general_purpose_register64() or source.is_memory_address64()) and immediate.is_none():
			# IMUL r64, r64/m64: REX.W 0F AF /r
			self.operands = 2
			self.source = source
			self.destination = destination
			self.size = 1 + 3 + source.get_modrm_extra_length()
			self.output_registers_list = destination.get_registers_list()
			self.input_registers_list = destination.get_registers_list() + source.get_registers_list()
		elif destination.is_general_purpose_register16() and (source.is_general_purpose_register16() or source.is_memory_address16()) and immediate.is_int16():
			# IMUL r16, r16/m16,  imm8: 66 6B /r ib
			# IMUL r16, r16/m16, imm16: 66 69 /r iw
			self.operands = 3
			self.source = source
			self.destination = destination
			self.immediate = immediate
			if immediate.is_sint8():
				self.size = 1 + 2 + source.get_modrm_extra_length() + 1
			else:
				self.size = 1 + 2 + source.get_modrm_extra_length() + 2
			self.output_registers_list = destination.get_registers_list()
			self.input_registers_list = source.get_registers_list()
		elif destination.is_general_purpose_register32() and (source.is_general_purpose_register32() or source.is_memory_address32()) and immediate.is_int32():
			# IMUL r32, r32/m32,  imm8: 6B /r ib
			# IMUL r32, r32/m32, imm32: 69 /r id
			self.operands = 3
			self.source = source
			self.destination = destination
			self.immediate = immediate
			if immediate.is_sint8():
				self.size = 2 + source.get_modrm_extra_length() + 1
			else:
				self.size = 2 + source.get_modrm_extra_length() + 4
			self.output_registers_list = destination.get_registers_list()
			self.input_registers_list = source.get_registers_list()
		elif destination.is_general_purpose_register64() and (source.is_general_purpose_register64() or source.is_memory_address64()) and immediate.is_int32():
			# IMUL r64, r64/m64,  imm8: REX.W 6B /r ib
			# IMUL r64, r64/m64, imm32: REX.W 69 /r id
			self.operands = 3
			self.source = source
			self.destination = destination
			self.immediate = immediate
			if immediate.is_sint8():
				self.size = 1 + 2 + source.get_modrm_extra_length() + 1
			else:
				self.size = 1 + 2 + source.get_modrm_extra_length() + 4
			self.output_registers_list = destination.get_registers_list()
			self.input_registers_list = source.get_registers_list()
		else:
			raise ValueError('Invalid operands in instruction IMUL {0}, {1}, {2}'.format(destination, source, immediate))

	def __str__(self):
		if self.operands == 1:
			return "IMUL {0}".format(self.source)
		elif self.operands == 2:
			return "IMUL {0}, {1}".format(self.destination, self.source)
		else:
			return "IMUL {0}, {1}, {2}".format(self.destination, self.source, self.immediate)

	def get_input_registers_list(self):
		return self.input_registers_list

	def get_output_registers_list(self):
		return self.output_registers_list

	def get_constant(self):
		if self.operands == 1:
			if self.source.is_constant():
				return self.source.constant
			else:
				return None
		else:
			if self.source.is_constant():
				return self.source.constant
			else:
				return None

	def get_local_variable(self):
		if self.operands == 1:
			if self.destination.is_local_variable():
				return self.destination.variable
			else:
				return None
		else:
			if self.source.is_local_variable():
				return self.source.variable
			else:
				return None

class MulInstruction(Instruction):
	def __init__(self, source):
		super(MulInstruction, self).__init__("MUL")
		if source.is_general_purpose_register8() or source.is_memory_address8():
			# MUL r8/m8: F6 /4
			self.source = source
			self.input_registers_list = self.source.get_registers_list() + [al]
			self.output_registers_list = [ax]
		elif source.is_general_purpuse_register16() or source.is_memory_address16():
			# MUL r16/m16: 66 F7 /4
			self.source = source
			self.input_registers_list = self.source.get_registers_list() + [ax]
			self.output_registers_list = [ax, dx]
		elif source.is_general_purpose_register32() or source.is_memory_address32():
			# MUL r32/m32: F7 /4
			self.source = source
			self.input_registers_list = self.source.get_registers_list() + [eax]
			self.output_registers_list = [eax, edx]
		else:
			raise ValueError('Invalid operands in instruction MUL {0}'.format(source))

	def __str__(self):
		return "MUL {0}".format(self.source)

	def get_input_registers_list(self):
		return self.input_registers_list

	def get_output_registers_list(self):
		return self.output_registers_list

	def get_constant(self):
		if self.source.is_constant():
			return self.source.constant
		else:
			return None

	def get_local_variable(self):
		if self.source.is_local_variable():
			return self.source.variable
		else:
			return None

class MoveExtendInstruction(BinaryInstruction):
	def __init__(self, name, destination, source):
		super(MoveExtendInstruction, self).__init__(name, destination, source)
		allowed_instructions = ['MOVZX', 'MOVSX']
		if not name in allowed_instructions:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})', name, ", ".join(allowed_instructions))
		if destination.is_general_purpose_register64() and (source.is_general_purpose_register8() or source.is_memory_address8(strict_size=True)):
			# MOVZX r64, r8/m8: REX.W 0F B6 /r
			# MOVSX r64, r8/m8: REX.W 0F BE /r
			self.size = 1 + 3 + source.get_modrm_extra_length()
			self.always_has_rex_prefix = True
		elif destination.is_general_purpose_register32() and (source.is_general_purpose_register8() or source.is_memory_address8(strict_size=True)):
			# MOVZX r32, r8/m8: 0F B6 /r
			# MOVSX r32, r8/m8: 0F BE /r
			self.size = 3 + source.get_modrm_extra_length()
		elif destination.is_general_purpose_register16() and (source.is_general_purpose_register8() or source.is_memory_address8(strict_size=True)):
			# MOVZX r16, r8/m8: 66 0F B6 /r
			# MOVSX r16, r8/m8: 66 0F BE /r
			self.size = 3 + 1 + source.get_modrm_extra_length()
		elif destination.is_general_purpose_register64() and (source.is_general_purpose_register16() or source.is_memory_address16(strict_size=True)):
			# MOVZX r64, r16/m16: REX.W 0F B7 /r
			# MOVSX r64, r16/m16: REX.W 0F BF /r
			self.size = 1 + 3 + source.get_modrm_extra_length()
			self.always_has_rex_prefix = True
		elif destination.is_general_purpose_register32() and (source.is_general_purpose_register16() or source.is_memory_address16(strict_size=True)):
			# MOVZX r32, r16/m16: 0F B7 /r
			# MOVSX r32, r16/m16: 0F BF /r
			self.size = 3 + source.get_modrm_extra_length()
		elif name == "MOVSX" and destination.is_general_purpose_register64() and (source.is_general_purpose_register32() or source.is_memory_address32(strict_size=True)):
			# MOVSX r64, r32/m32: REX.W 63 /r
			self.size = 1 + 2 + source.get_modrm_extra_length()
			self.always_has_rex_prefix = True
		else:
			raise ValueError('Invalid operands in instruction {0} {1}, {2}'.format(name, destination, source))

	def get_input_registers_list(self):
		input_registers_list = list()
		input_registers_list.extend(self.source.get_registers_list())
		if self.destination.is_memory_address():
			input_registers_list.extend(self.destination.get_registers_list())
		return input_registers_list

	def get_output_registers_list(self):
		if self.destination.is_register():
			return [self.destination.register]
		else:
			return list()

	def get_constant(self):
		if self.source.is_constant():
			return self.source.constant
		else:
			return None

	def get_local_variable(self):
		if self.source.is_local_variable():
			return self.source.variable
		else:
			return None

class MovInstruction(BinaryInstruction):
	def __init__(self, destination, source):
		super(MovInstruction, self).__init__('MOV', destination, source)
		if destination.is_general_purpose_register64() and (source.is_general_purpose_register64() or source.is_memory_address64()):
			# MOV r64, r64/m64: REX.W 8B /r
			self.size = 1 + 2 + source.get_modrm_extra_length()
		elif destination.is_general_purpose_register32() and (source.is_general_purpose_register32() or source.is_memory_address32()):
			# MOV r32, r32/m32: 8B /r
			self.size = 2 + source.get_modrm_extra_length()
		elif destination.is_general_purpose_register16() and (source.is_general_purpose_register16() or source.is_memory_address16()):
			# MOV r16, r16/m16: 66 8B /r
			self.size = 1 + 2 + source.get_modrm_extra_length()
		elif destination.is_general_purpose_register8() and (source.is_general_purpose_register8() or source.is_memory_address8()):
			# MOV r8, r8/m8: 8A /r
			self.size = 2 + source.get_modrm_extra_length()
		elif (destination.is_general_purpose_register64() or destination.is_memory_address64()) and source.is_general_purpose_register64():
			# MOV r64/m64, r64: REX.W 89 /r
			self.size = 1 + 2 + destination.get_modrm_extra_length()
		elif (destination.is_general_purpose_register32() or destination.is_memory_address32()) and source.is_general_purpose_register32():
			# MOV r32/m32, r32: 89 /r
			self.size = 2 + destination.get_modrm_extra_length()
		elif (destination.is_general_purpose_register16() or destination.is_memory_address16()) and source.is_general_purpose_register16():
			# MOV r16/m16, r16: 66 89 /r
			self.size = 1 + 2 + destination.get_modrm_extra_length()
		elif (destination.is_general_purpose_register8() or destination.is_memory_address8()) and source.is_general_purpose_register8():
			# MOV r8/m8, r8: 88 /r
			self.size = 2 + destination.get_modrm_extra_length()
		elif destination.is_general_purpose_register64() and source.is_int64():
			# MOV r64, imm64: REX.W B8+ rq
			self.size = 1 + 1 + 8
		elif destination.is_general_purpose_register32() and source.is_int32():
			# MOV r32, imm32: B8 +rd id
			self.size = 1 + 4
		elif destination.is_general_purpose_register16() and source.is_int16():
			# MOV r16, imm16: 66 B8 +rw iw
			self.size = 1 + 1 + 2
		elif destination.is_general_purpose_register8() and source.is_int8():
			# MOV r8, imm8: B0 +rb ib
			self.size = 1 + 1
		elif (destination.is_general_purpose_register64() or destination.is_memory_address64(strict_size = True)) and source.is_sint32():
			# MOV r64/m64, imm32: REX.W C7 /0 id
			self.size = 1 + 2 + 4 + destination.get_modrm_extra_length()
		elif (destination.is_general_purpose_register32() or destination.is_memory_address32(strict_size = True)) and source.is_int32():
			# MOV r32/m32, imm32: C7 /0 id
			self.size = 2 + 4 + destination.get_modrm_extra_length()
		elif (destination.is_general_purpose_register16() or destination.is_memory_address16(strict_size = True)) and source.is_int16():
			# MOV r16/m16, imm16: 66 C7 /0 iw
			self.size = 1 + 2 + 2 + destination.get_modrm_extra_length()
		elif (destination.is_general_purpose_register8() or destination.is_memory_address8(strict_size = True)) and source.is_int8():
			# MOV r8/m8, imm8: C6 /0 ib
			self.size = 2 + 1 + destination.get_modrm_extra_length()
		else:
			raise ValueError('Invalid operands in instruction MOV {0}, {1}'.format(destination, source))

	def get_input_registers_list(self):
		input_registers_list = list()
		input_registers_list.extend(self.source.get_registers_list())
		if self.destination.is_memory_address():
			input_registers_list.extend(self.destination.get_registers_list())
		return input_registers_list

	def get_output_registers_list(self):
		if self.destination.is_register():
			return [self.destination.register]
		else:
			return list()

	def get_constant(self):
		if self.source.is_constant():
			return self.source.constant
		else:
			return None

	def get_local_variable(self):
		if self.source.is_local_variable():
			return self.source.variable
		elif self.destination.is_local_variable():
			return self.destination.variable
		else:
			return None

class ConditionalJumpInstruction(Instruction):
	def __init__(self, name, destination):
		super(ConditionalJumpInstruction, self).__init__(name)
		allowed_instructions = [ 'JA',  'JAE',  'JB',  'JBE',  'JC',  'JE',  'JG',  'JGE',  'JL',  'JLE',  'JO',  'JP',  'JPO',  'JS',  'JZ',
								'JNA', 'JNAE', 'JNB', 'JNBE', 'JNC', 'JNE', 'JNG', 'JNGE', 'JNL', 'JNLE', 'JNO', 'JNP', 'JNPO', 'JNS', 'JNZ']
		if not name in allowed_instructions:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})', name, ", ".join(allowed_instructions))
		if destination.is_label():
			self.destination = destination.label
			#   JA rel32: 0F 87 cd
			#  JAE rel32: 0F 83 cd
			#   JB rel32: 0F 82 cd
			#  JBE rel32: 0F 86 cd
			#   JC rel32: 0F 82 cd
			#   JE rel32: 0F 84 cd
			#   JZ rel32: 0F 84 cd
			self.size = 2 + 4
			self.is_visited = False
		else:
			raise TypeError('Only string names are accepted as destination labels')

	def __str__(self):
		return "{0} .{1}".format(self.name, self.destination)

	def get_input_registers_list(self):
		return list()

	def get_output_registers_list(self):
		return list()

	def get_constant(self):
		return None

	def get_local_variable(self):
		return None

class CmovInstruction(BinaryInstruction):
	def __init__(self, name, destination, source):
		super(CmovInstruction, self).__init__(name,  destination, source, 'CMOV')
		allowed_instructions = [ 'CMOVA',  'CMOVAE',  'CMOVB',  'CMOVBE',  'CMOVC',  'CMOVE',  'CMOVG',  'CMOVGE',  'CMOVL',  'CMOVLE',  'CMOVO',  'CMOVP',  'CMOVPO',  'CMOVS',  'CMOVZ',
								'CMOVNA', 'CMOVNAE', 'CMOVNB', 'CMOVNBE', 'CMOVNC', 'CMOVNE', 'CMOVNG', 'CMOVNGE', 'CMOVNL', 'CMOVNLE', 'CMOVNO', 'CMOVNP', 'CMOVNPO', 'CMOVNS', 'CMOVNZ']
		if not name in allowed_instructions:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})', name, ", ".join(allowed_instructions))
		if destination.is_general_purpose_register32() and (source.is_general_purpose_register32() or source.is_memory_address32()):
			#   CMOVA r32, r32/m32: 0F 47 /r
			#  CMOVAE r32, r32/m32: 0F 43 /r
			#   CMOVB r32, r32/m32: 0F 42 /r
			#  CMOVBE r32, r32/m32: 0F 46 /r
			#   CMOVC r32, r32/m32: 0F 42 /r
			#   CMOVE r32, r32/m32: 0F 44 /r
			#   CMOVG r32, r32/m32: 0F 4F /r
			#  CMOVGE r32, r32/m32: 0F 4D /r
			#   CMOVL r32, r32/m32: 0F 4C /r
			#  CMOVLE r32, r32/m32: 0F 4E /r
			#  CMOVNA r32, r32/m32: 0F 46 /r
			# CMOVNAE r32, r32/m32: 0F 42 /r
			#  CMOVNB r32, r32/m32: 0F 43 /r
			# CMOVNBE r32, r32/m32: 0F 47 /r
			#  CMOVNC r32, r32/m32: 0F 43 /r
			#  CMOVNE r32, r32/m32: 0F 45 /r
			#  CMOVNG r32, r32/m32: 0F 4E /r
			# CMOVNGE r32, r32/m32: 0F 4C /r
			#  CMOVNL r32, r32/m32: 0F 4D /r
			# CMOVNLE r32, r32/m32: 0F 4F /r
			#  CMOVNO r32, r32/m32: 0F 41 /r
			#  CMOVNP r32, r32/m32: 0F 4B /r
			#  CMOVNS r32, r32/m32: 0F 49 /r
			#  CMOVNZ r32, r32/m32: 0F 45 /r
			#   CMOVO r32, r32/m32: 0F 40 /r
			#   CMOVP r32, r32/m32: 0F 4A /r
			#  CMOVPE r32, r32/m32: 0F 4A /r
			#  CMOVPO r32, r32/m32: 0F 4B /r
			#   CMOVS r32, r32/m32: 0F 48 /r
			#   CMOVZ r32, r32/m32: 0F 44 /r
			self.destination = destination
			self.source = source
			self.size = 3 + source.get_modrm_extra_length()
		elif destination.is_general_purpose_register16() and (source.is_general_purpose_register16() or source.is_memory_address16()):
			#   CMOVA r16, r16/m16: 66 0F 47 /r
			#  CMOVAE r16, r16/m16: 66 0F 43 /r
			#   CMOVB r16, r16/m16: 66 0F 42 /r
			#  CMOVBE r16, r16/m16: 66 0F 46 /r
			#   CMOVC r16, r16/m16: 66 0F 42 /r
			#   CMOVE r16, r16/m16: 66 0F 44 /r
			#   CMOVG r16, r16/m16: 66 0F 4F /r
			#  CMOVGE r16, r16/m16: 66 0F 4D /r
			#   CMOVL r16, r16/m16: 66 0F 4C /r
			#  CMOVLE r16, r16/m16: 66 0F 4E /r
			#  CMOVNA r16, r16/m16: 66 0F 46 /r
			# CMOVNAE r16, r16/m16: 66 0F 42 /r
			#  CMOVNB r16, r16/m16: 66 0F 43 /r
			# CMOVNBE r16, r16/m16: 66 0F 47 /r
			#  CMOVNC r16, r16/m16: 66 0F 43 /r
			#  CMOVNE r16, r16/m16: 66 0F 45 /r
			#  CMOVNG r16, r16/m16: 66 0F 4E /r
			# CMOVNGE r16, r16/m16: 66 0F 4C /r
			#  CMOVNL r16, r16/m16: 66 0F 4D /r
			# CMOVNLE r16, r16/m16: 66 0F 4F /r
			#  CMOVNO r16, r16/m16: 66 0F 41 /r
			#  CMOVNP r16, r16/m16: 66 0F 4B /r
			#  CMOVNS r16, r16/m16: 66 0F 49 /r
			#  CMOVNZ r16, r16/m16: 66 0F 45 /r
			#   CMOVO r16, r16/m16: 66 0F 40 /r
			#   CMOVP r16, r16/m16: 66 0F 4A /r
			#  CMOVPE r16, r16/m16: 66 0F 4A /r
			#  CMOVPO r16, r16/m16: 66 0F 4B /r
			#   CMOVS r16, r16/m16: 66 0F 48 /r
			#   CMOVZ r16, r16/m16: 66 0F 44 /r
			self.destination = destination
			self.source = source
			self.size = 1 + 3 + source.get_modrm_extra_length()
		else:
			raise ValueError('Invalid operands in instruction {0} {1}, {2}'.format(name, destination, source))

	def get_input_registers_list(self):
		input_registers_list = [self.destination]
		input_registers_list.extend(self.source.get_registers_list)
		return input_registers_list

	def get_output_registers_list(self):
		return [self.destination.register]

	def get_constant(self):
		if self.source.is_constant():
			return self.source.constant
		else:
			return None

	def get_local_variable(self):
		if self.source.is_local_variable():
			return self.source.variable
		else:
			return None

class SetInstruction(Instruction):
	def __init__(self, name, destination):
		super(SetInstruction, self).__init__(name)
		allowed_instructions = [ 'SETA',  'SETAE',  'SETB',  'SETBE',  'SETC',  'SETE',  'SETG',  'SETGE',  'SETL',  'SETLE',  'SETO',  'SETP',  'SETPO',  'SETS',  'SETZ',
								'SETNA', 'SETNAE', 'SETNB', 'SETNBE', 'SETNC', 'SETNE', 'SETNG', 'SETNGE', 'SETNL', 'SETNLE', 'SETNO', 'SETNP', 'SETNPO', 'SETNS', 'SETNZ']
		if not name in allowed_instructions:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})', name, ", ".join(allowed_instructions))
		if destination.is_general_purpose_register8() or destination.is_memory_address8():
			#   SETA r8/m8: 0F 97 /r
			#  SETAE r8/m8: 0F 93 /r
			#   SETB r8/m8: 0F 92 /r
			#  SETBE r8/m8: 0F 96 /r
			#   SETC r8/m8: 0F 92 /r
			#   SETE r8/m8: 0F 94 /r
			#   SETG r8/m8: 0F 9F /r
			#  SETGE r8/m8: 0F 9D /r
			#   SETL r8/m8: 0F 9C /r
			#  SETLE r8/m8: 0F 9E /r
			#  SETNA r8/m8: 0F 96 /r
			# SETNAE r8/m8: 0F 92 /r
			#  SETNB r8/m8: 0F 93 /r
			# SETNBE r8/m8: 0F 97 /r
			#  SETNC r8/m8: 0F 93 /r
			#  SETNE r8/m8: 0F 95 /r
			#  SETNG r8/m8: 0F 9E /r
			# SETNGE r8/m8: 0F 9C /r
			#  SETNL r8/m8: 0F 9D /r
			# SETNLE r8/m8: 0F 9F /r
			#  SETNO r8/m8: 0F 91 /r
			#  SETNP r8/m8: 0F 9B /r
			#  SETNS r8/m8: 0F 99 /r
			#  SETNZ r8/m8: 0F 95 /r
			#   SETO r8/m8: 0F 90 /r
			#   SETP r8/m8: 0F 9A /r
			#  SETPE r8/m8: 0F 9A /r
			#  SETPO r8/m8: 0F 9B /r
			#   SETS r8/m8: 0F 98 /r
			#   SETZ r8/m8: 0F 94 /r
			self.destination = destination
			self.size = 3 + destination.get_modrm_extra_length()
		else:
			raise ValueError('Invalid operands in instruction {0} {1}}'.format(name, destination))

	def __str__(self):
		return "{0} {1}".format(self.name, self.destination)

	def get_input_registers_list(self):
		if self.destination.is_general_purpose_register8():
			return list()
		else:
			return self.destination.get_register_list()

	def get_output_registers_list(self):
		if self.destination.is_general_purpose_register8():
			return [self.destination.register]
		else:
			return list()

	def get_constant(self):
		return None

	def get_local_variable(self):
		if self.destination.is_local_variable():
			return self.destination.variable
		else:
			return None

class AvxZeroInstruction(Instruction):
	def __init__(self, name):
		allowed_instructions = [ 'VZEROALL', 'VZEROUPPER' ]
		if name in allowed_instructions:
			super(AvxZeroInstruction, self).__init__(name, 'AVX')
			# VZEROUPPER: VEX.128.0F.WIG 77
			#   VZEROALL: VEX.256.0F.WIG 77

			self.size = 2 + 1
		else:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})', name, ", ".join(allowed_instructions))

	def __str__(self):
		return self.name

	def get_input_registers_list(self):
		if self.name == 'VZEROALL':
			return []
		else:
			return [ymm0, ymm1, ymm2, ymm3, ymm4,
			        ymm5, ymm6, ymm7, ymm8, ymm9,
			        ymm10, ymm11, ymm12, ymm13, ymm14, ymm15]

	def get_output_registers_list(self):
		return [ymm0, ymm1, ymm2, ymm3, ymm4,
		        ymm5, ymm6, ymm7, ymm8, ymm9,
		        ymm10, ymm11, ymm12, ymm13, ymm14, ymm15]

	def get_constant(self):
		return None

	def get_local_variable(self):
		return None

class SseMovInstruction(BinaryInstruction):
	def __init__(self, name, destination, source):
		allowed_instructions = [ 'MOVAPS', 'MOVUPS', 'MOVAPD', 'MOVUPD', 'MOVDQA', 'MOVDQU', 'LDDQU' ]
		if not name in allowed_instructions:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})', name, ", ".join(allowed_instructions))
		is_ps = name in ['MOVAPS', 'MOVUPS']
		if name == 'LDDQU':
			super(SseMovInstruction, self).__init__(name, destination, source, 'SSE3')
		elif is_ps:
			super(SseMovInstruction, self).__init__(name, destination, source, 'SSE')
		else:
			super(SseMovInstruction, self).__init__(name, destination, source, 'SSE2')
		if name == 'LDDQU':
			if destination.is_sse_register() and source.is_memory_address128():
				# LDDQU xmm, m128: F2 0F F0 /r
				self.size = 1 + 3 + source.get_modrm_extra_length()
			else:
				raise ValueError('Invalid operands in instruction {0} {1}, {2}'.format(name, destination, source))
		else:
			if destination.is_sse_register() and (source.is_sse_register() or source.is_memory_address128()):
				# MOVAPS xmm, xmm/m128: 0F 28 /r
				# MOVUPS xmm, xmm/m128: 0F 10 /r
				# MOVAPD xmm, xmm/m128: 66 0F 28 /r
				# MOVUPD xmm, xmm/m128: 66 0F 10 /r
				# MOVDQA xmm, xmm/m128: 66 0F 6F /r
				# MOVDQU xmm, xmm/m128: F3 0F 6F /r
				if is_ps:
					self.size = 3 + source.get_modrm_extra_length()
				else:
					self.size = 1 + 3 + source.get_modrm_extra_length()
			elif (destination.is_sse_register() or destination.is_memory_address128()) and source.is_sse_register():
				# MOVAPS xmm/m128, xmm: 0F 29 /r
				# MOVUPS xmm/m128, xmm: 0F 11 /r
				# MOVAPD xmm/m128, xmm: 66 0F 29 /r
				# MOVUPD xmm/m128, xmm: 66 0F 11 /r
				# MOVDQA xmm/m128, xmm: 66 0F 7F /r
				# MOVDQU xmm/m128, xmm: F3 0F 7F /r
				if is_ps:
					self.size = 3 + destination.get_modrm_extra_length()
				else:
					self.size = 1 + 3 + destination.get_modrm_extra_length()
			else:
				raise ValueError('Invalid operands in instruction {0} {1}, {2}'.format(name, destination, source))

	def get_input_registers_list(self):
		input_registers_list = list()
		if self.destination.is_memory_address():
			input_registers_list.extend(self.destination.get_registers_list())
		input_registers_list.extend(self.source.get_registers_list())
		return input_registers_list

	def get_output_registers_list(self):
		if self.destination.is_sse_register():
			return [self.destination.register]
		else:
			return list()

	def get_constant(self):
		if self.source.is_constant():
			return self.source.constant
		else:
			return None

	def get_local_variable(self):
		if self.destination.is_local_variable():
			return self.destination.variable
		elif self.source.is_local_variable():
			return self.source.variable
		else:
			return None

class AvxMovInstruction(BinaryInstruction):
	def __init__(self, name, destination, source):
		allowed_instructions = [ 'VMOVAPS', 'VMOVUPS', 'VMOVAPD', 'VMOVUPD', 'VMOVDQA', 'VMOVDQU' ]
		if not name in allowed_instructions:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})'.format(name, ", ".join(allowed_instructions)))
		super(AvxMovInstruction, self).__init__(name, destination, source, 'AVX')
		if destination.is_sse_register() and (source.is_sse_register() or source.is_memory_address128()):
			# VMOVAPS xmm, xmm/m128: VEX.128.0F.WIG 28 /r
			# VMOVUPS xmm, xmm/m128: VEX.128.0F.WIG 10 /r
			# VMOVAPD xmm, xmm/m128: VEX.128.66.0F.WIG 28 /r
			# VMOVUPD xmm, xmm/m128: VEX.128.66.0F.WIG 10 /r
			# VMOVDQA xmm, xmm/m128: VEX.128.66.0F.WIG 6F /r
			# VMOVDQU xmm, xmm/m128: VEX.128.F3.0F.WIG 6F /r
			self.size = source.get_vex_extra_length() + 1 + source.get_modrm_extra_length()
		elif destination.is_avx_register() and (source.is_avx_register() or source.is_memory_address256()):
			# VMOVAPS ymm, ymm/m256: VEX.256.0F.WIG 28 /r
			# VMOVUPS ymm, ymm/m256: VEX.256.0F.WIG 10 /r
			# VMOVAPD ymm, ymm/m256: VEX.256.66.0F.WIG 28 /r
			# VMOVUPD ymm, ymm/m256: VEX.256.66.0F.WIG 10 /r
			# VMOVDQA ymm, ymm/m256: VEX.256.66.0F.WIG 6F /r
			# VMOVDQU ymm, ymm/m256: VEX.256.F3.0F.WIG 6F /r
			self.size = source.get_vex_extra_length() + 1 + source.get_modrm_extra_length()
		elif (destination.is_sse_register() or destination.is_memory_address128()) and source.is_sse_register():
			# VMOVAPS xmm/m128, xmm: VEX.128.0F.WIG 29 /r
			# VMOVUPS xmm/m128, xmm: VEX.128.0F.WIG 11 /r
			# VMOVAPD xmm/m128, xmm: VEX.128.66.0F.WIG 29 /r
			# VMOVUPD xmm/m128, xmm: VEX.128.66.0F.WIG 11 /r
			# VMOVDQA xmm/m128, xmm: VEX.128.66.0F.WIG 7F /r
			# VMOVDQU xmm/m128, xmm: VEX.128.F3.0F.WIG 7F /r
			self.size = destination.get_vex_extra_length() + 1 + destination.get_modrm_extra_length()
		elif (destination.is_avx_register() or destination.is_memory_address256()) and source.is_avx_register():
			# VMOVAPS ymm/m256, ymm: VEX.256.0F.WIG 29 /r
			# VMOVUPS ymm/m256, ymm: VEX.256.0F.WIG 11 /r
			# VMOVAPD ymm/m256, ymm: VEX.256.66.0F.WIG 29 /r
			# VMOVUPD ymm/m256, ymm: VEX.256.66.0F.WIG 11 /r
			# VMOVDQA ymm/m256, ymm: VEX.256.66.0F.WIG 7F /r
			# VMOVDQU ymm/m256, ymm: VEX.256.F3.0F.WIG 7F /r
			self.size = destination.get_vex_extra_length() + 1 + destination.get_modrm_extra_length()
		else:
			raise ValueError('Invalid operands in instruction {0} {1}, {2}'.format(name, destination, source))

	def get_input_registers_list(self):
		input_registers_list = list()
		if self.destination.is_memory_address():
			input_registers_list.extend(self.destination.get_registers_list())
		input_registers_list.extend(self.source.get_registers_list())
		return input_registers_list

	def get_output_registers_list(self):
		if self.destination.is_sse_register():
			return [self.destination.register]
		else:
			return list()

	def get_constant(self):
		if self.source.is_constant():
			return self.source.constant
		else:
			return None

	def get_local_variable(self):
		if self.destination.is_local_variable():
			return self.destination.variable
		elif self.source.is_local_variable():
			return self.source.variable
		else:
			return None

class SseMovWordInstruction(BinaryInstruction):
	def __init__(self, name, destination, source):
		allowed_instructions = [ 'MOVD', 'MOVQ' ]
		if not name in allowed_instructions:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})', name, ", ".join(allowed_instructions))
		super(SseMovWordInstruction, self).__init__(name, destination, source, 'SSE2')
		if destination.is_sse_register() and (source.is_general_purpose_register32() or source.is_memory_address32()):
			# MOVD xmm, r32/m32: 66 0F 6E /r
			self.size = 1 + 3 + source.get_modrm_extra_length()
		elif destination.is_sse_register() and (source.is_general_purpose_register64() or source.is_memory_address64()):
			# MOVQ xmm, r64/m64: 66 REX.W 0F 6E /r
			self.size = 1 + 1 + 3 + source.get_modrm_extra_length()
		elif (destination.is_general_purpose_register32() or destination.is_memory_address32()) or source.is_sse_register():
			# MOVD r32/m32, xmm: 66 0F 7E /r
			self.size = 1 + 3 + destination.get_modrm_extra_length()
		elif (destination.is_general_purpose_register64() or destination.is_memory_address64()) or source.is_sse_register():
			# MOVD r64/m64, xmm: 66 REX.W 0F 7E /r
			self.size = 1 + 1 + 3 + destination.get_modrm_extra_length()
		else:
			raise ValueError('Invalid operands in instruction {0} {1}, {2}'.format(name, destination, source))

	def get_input_registers_list(self):
		input_registers_list = list()
		if self.destination.is_memory_address():
			input_registers_list.extend(self.destination.get_registers_list())
		input_registers_list.extend(self.source.get_registers_list())
		return input_registers_list

	def get_output_registers_list(self):
		if self.destination.is_register():
			return [self.destination.register]
		else:
			return list()

	def get_constant(self):
		if self.source.is_constant():
			return self.source.constant
		else:
			return None

	def get_local_variable(self):
		if self.source.is_local_variable():
			return self.source.variable
		elif self.destination.is_local_variable():
			return self.destination.variable
		else:
			return None

class MxcsrControlInstruction(Instruction):
	def __init__(self, name, operand):
		allowed_instructions = ['LDMXCSR', 'STMXCSR']
		if name in allowed_instructions:
			super(MxcsrControlInstruction, self).__init__(name, 'SSE')
		else:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})', name, ", ".join(allowed_instructions))
		if operand.is_memory_address32():
			# LDMXCSR m32: 0F AE /2
			# STMXCSR m32: 0F AE /3
			self.operand = operand
			self.size = 3 + operand.get_modrm_extra_length()
		else:
			raise ValueError('Invalid operands in instruction {0} {1}'.format(name, operand))

	def __str__(self):
		return "{0} {1}".format(self.name, self.operand)

	def get_input_registers_list(self):
		return self.operand.get_registers_list()

	def get_output_registers_list(self):
		return list()

	def get_constant(self):
		if self.name == 'LDMXCSR' and self.operand.is_constant():
			return self.operand.constant
		else:
			return None

	def get_local_variable(self):
		if self.operand.is_local_variable():
			return self.operand.variable
		else:
			return None

class AvxExtract128Instruction(Instruction):
	def __init__(self, name, destination, source, immediate):
		allowed_instructions = [ 'VEXTRACTF128' ]
		if not name in allowed_instructions:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})'.format(name, ", ".join(allowed_instructions)))
		super(AvxExtract128Instruction, self).__init__(name, 'AVX')
		self.destination = destination
		self.source = source
		self.immediate = immediate
		if (destination.is_sse_register() or destination.is_memory_address128()) and source.is_avx_register() and immediate.is_uint8():
			# VEXTRACTF128 xmm/m128, ymm, imm8: VEX.256.66.0F3A.W0 19 /r ib

			self.size = 3 + 1 + destination.get_modrm_extra_length()
		else:
			raise ValueError('Invalid operands in instruction {0} {1}, {2}'.format(name, destination, source, immediate))

	def __str__(self):
		return "{0} {1}, {2}, {3}".format(self.name, self.destination, self.source, self.immediate)

	def get_input_registers_list(self):
		input_registers_list = self.source.get_registers_list()
		if self.destination.is_memory_address():
			input_registers_list.extend(self.destination)
		return input_registers_list

	def get_output_registers_list(self):
		if self.destination.is_register():
			return [self.destination.register]
		else:
			return list()

	def get_constant(self):
		return None

	def get_local_variable(self):
		if self.destination.is_local_variable():
			return self.destination.variable
		else:
			return None

class SseScalarFloatingPointMovInstruction(BinaryInstruction):
	def __init__(self, name, destination, source):
		allowed_instructions = [ 'MOVSS', 'MOVSD' ]
		if not name in allowed_instructions:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})'.format(name, ", ".join(allowed_instructions)))
		if name == 'MOVSS':
			super(SseScalarFloatingPointMovInstruction, self).__init__(name, destination, source, 'SSE')
		else:
			super(SseScalarFloatingPointMovInstruction, self).__init__(name, destination, source, 'SSE2')
		if name == 'MOVSS' and (destination.is_sse_register() or destination.is_memory_address32()) and source.is_sse_register():
			# MOVSS xmm/m32, xmm: F3 0F 11 /r
			self.size = 4 + destination.get_modrm_extra_length()
		elif name == 'MOVSS' and destination.is_sse_register() and (source.is_sse_register() or source.is_memory_address32()):
			# MOVSS xmm, xmm/m32: F3 0F 10 /r
			self.size = 4 + source.get_modrm_extra_length()
		elif name == 'MOVSD' and (destination.is_sse_register() or destination.is_memory_address64()) and source.is_sse_register():
			# MOVSD xmm/m64, xmm: F2 0F 11 /r
			self.size = 4 + destination.get_modrm_extra_length()
		elif name == 'MOVSD' and destination.is_sse_register() and (source.is_sse_register() or source.is_memory_address64()):
			# MOVSD xmm, xmm/m64: F2 0F 10 /r
			self.size = 4 + source.get_modrm_extra_length()
		else:
			raise ValueError('Invalid operands in instruction {0} {1}, {2}'.format(name, destination, source))

	def get_input_registers_list(self):
		if self.source.is_memory_address():
			# MOVSx xmm, [mem]
			return self.source.get_registers_list()
		else:
			# MOVSx xmm, xmm or MOVSx [mem], xmm
			return self.destination.get_registers_list() + self.source.get_registers_list()

	def get_output_registers_list(self):
		if self.destination.is_sse_register():
			return [self.destination.register]
		else:
			return list()

	def get_constant(self):
		if self.source.is_constant():
			return self.source.constant
		else:
			return None

	def get_local_variable(self):
		if self.source.is_local_variable():
			return self.source.variable
		elif self.destination.is_local_variable():
			return self.destination.variable
		else:
			return None

class AvxScalarFloatingPointMovInstruction(Instruction):
	def __init__(self, name, destination, source_x, source_y):
		allowed_instructions = [ 'VMOVSS', 'VMOVSD' ]
		if not name in allowed_instructions:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})', name, ", ".join(allowed_instructions))
		super(AvxScalarFloatingPointMovInstruction, self).__init__(name, 'AVX')
		self.destination = destination
		self.source_x = source_x
		self.source_y = source_y
		if name == 'VMOVSS' and destination.is_sse_register() and source_x.is_sse_register() and source_y.is_sse_register():
			# VMOVSS xmm, xmm, xmm: VEX.NDS.LIG.F3.0F.WIG 10 /r
			# VMOVSS xmm, xmm, xmm: VEX.NDS.LIG.F3.0F.WIG 11 /r

			self.size = destination.get_vex_extra_length() + 1 + destination.get_modrm_extra_length()
		elif name == 'VMOVSS' and destination.is_sse_register() and source_x.is_memory_address32() and source_y.is_none():
			# VMOVSS xmm, m32: VEX.LIG.F3.0F.WIG 10 /r

			self.size = source_x.get_vex_extra_length() + 1 + source_x.get_modrm_extra_length()
		elif name == 'VMOVSS' and destination.is_memory_address32() and source_x.is_sse_register() and source_y.is_none():
			# VMOVSS m32, xmm: VEX.LIG.F3.0F.WIG 11 /r

			self.size = destination.get_vex_extra_length() + 1 + destination.get_modrm_extra_length()
		elif name == 'VMOVSD' and destination.is_sse_register() and source_x.is_sse_register() and source_y.is_sse_register():
			# VMOVSD xmm, xmm, xmm: VEX.NDS.LIG.F2.0F.WIG 10 /r
			# VMOVSD xmm, xmm, xmm: VEX.NDS.LIG.F2.0F.WIG 11 /r

			self.size = destination.get_vex_extra_length() + 1 + destination.get_modrm_extra_length()
		elif name == 'VMOVSD' and destination.is_sse_register() and source_x.is_memory_address64() and source_y.is_none():
			# VMOVSD xmm, m64: VEX.LIG.F2.0F.WIG 10 /r

			self.size = source_x.get_vex_extra_length() + 1 + source_x.get_modrm_extra_length()
		elif name == 'VMOVSD' and destination.is_memory_address64() and source_x.is_sse_register() and source_y.is_none():
			# VMOVSD m64, xmm: VEX.LIG.F2.0F.WIG 11 /r

			self.size = destination.get_vex_extra_length() + 1 + destination.get_modrm_extra_length()
		else:
			raise ValueError('Invalid operands in instruction {0} {1}, {2}'.format(name, destination, source))

	def __str__(self):
		if self.source_y.is_none():
			return "{0} {1}, {2}".format(self.name, self.destination, self.source_x)
		else:
			return "{0} {1}, {2}, {3}".format(self.name, self.destination, self.source_x, self.source_y)

	def get_input_registers_list(self):
		input_registers_list = self.source_x.get_registers_list()
		if self.destination.is_memory_address():
			input_registers_list.extend(self.destination.get_registers_list())
		if self.source_y.is_sse_register():
			input_registers_list.extend(self.source_y.get_registers_list())
		return input_registers_list

	def get_output_registers_list(self):
		if self.destination.is_sse_register():
			return [self.destination.register]
		else:
			return list()

	def get_constant(self):
		if self.source_x.is_constant():
			return self.source_x.constant
		else:
			return None

	def get_local_variable(self):
		if self.source_x.is_local_variable():
			return self.source_x.variable
		elif self.destination.is_local_variable():
			return self.destination.variable
		else:
			return None

class SseFloatingPointBinaryInstruction(BinaryInstruction):
	def __init__(self, name, destination, source):
		sse_scalar_instructions = [ 'ADDSS', 'SUBSS', 'MULSS', 'DIVSS', 'MINSS', 'MAXSS',
									'CMPEQSS',  'CMPLTSS',  'CMPLESS',  'CMPUNORDSS', 
									'CMPNEQSS', 'CMPNLTSS', 'CMPNLESS', 'CMPORDSS']
		sse_vector_instructions = [ 'ADDPS', 'SUBPS', 'MULPS', 'DIVPS', 'MINPS', 'MAXPS',
									'ANDPS', 'ANDNPS', 'ORPS', 'XORPS',  'UNPCKLPS', 'UNPCKHPS',
									'CMPEQPS',  'CMPLTPS',  'CMPLEPS',  'CMPUNORDPS', 
									'CMPNEQPS', 'CMPNLTPS', 'CMPNLEPS', 'CMPORDPS']
		sse_instructions = sse_scalar_instructions + sse_vector_instructions
		sse2_scalar_instructions = ['ADDSD', 'SUBSD', 'MULSD', 'DIVSD', 'MINSD', 'MAXSD',
									'CMPEQSD',  'CMPLTSD',  'CMPLESD',  'CMPUNORDSD', 
									'CMPNEQSD', 'CMPNLTSD', 'CMPNLESD', 'CMPORDSD']
		sse2_vector_instructions = ['ADDPD', 'SUBPD', 'MULPD', 'DIVPD', 'MINPD', 'MAXPD',
									'ANDPD', 'ANDNPD', 'ORPD', 'XORPD',  'UNPCKLPD', 'UNPCKHPD',
									'CMPEQPD',  'CMPLTPD',  'CMPLEPD',  'CMPUNORDPD', 
									'CMPNEQPD', 'CMPNLTPD', 'CMPNLEPD', 'CMPORDPD']
		sse3_vector_instructions = ['ADDSUBPS', 'ADDSUBPD', 'HADDPS', 'HADDPD', 'HSUBPS', 'HSUBPD']
		sse2_instructions =  sse2_scalar_instructions + sse2_vector_instructions
		vector_instructions = sse_vector_instructions + sse2_vector_instructions + sse3_vector_instructions
		if name in sse_instructions:
			super(SseFloatingPointBinaryInstruction, self).__init__(name, destination, source, 'SSE')
		elif name in sse2_instructions:
			super(SseFloatingPointBinaryInstruction, self).__init__(name, destination, source, 'SSE2')
		elif name in sse3_vector_instructions:
			super(SseFloatingPointBinaryInstruction, self).__init__(name, destination, source, 'SSE3')
		else:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})', name, ", ".join(sse_instructions + sse2_instructions))
		if name in vector_instructions and destination.is_sse_register() and (source.is_sse_register() or source.is_memory_address128()):
			# ADDPS xmm, xmm/m128: 0F 58 /r
			# ADDPD xmm, xmm/m128: 66 0F 58 /r

			# SUBPS xmm, xmm/m128: 0F 5C /r
			# SUBPD xmm, xmm/m128: 66 0F 5C /r

			# MULPS xmm, xmm/m128: 0F 59 /r
			# MULPD xmm, xmm/m128: 66 0F 59 /r

			# DIVPS xmm, xmm/m128: 0F 5E /r
			# DIVPD xmm, xmm/m128: 66 0F 5E /r

			# CMPPS xmm, xmm/m128: 0F C2 /r
			# CMPPD xmm, xmm/m128: 66 0F C2 /r

			# MINPS xmm, xmm/m128: 0F 5D /r
			# MINPD xmm, xmm/m128: 66 0F 5D /r

			# MAXPS xmm, xmm/m128: 0F 5F /r
			# MAXPD xmm, xmm/m128: 66 0F 5F /r

			# ANDPS xmm, xmm/m128: 0F 54 /r
			# ANDPD xmm, xmm/m128: 66 0F 54 /r

			# ANDNPS xmm, xmm/m128: 0F 55 /r
			# ANDNPD xmm, xmm/m128: 66 0F 55 /r

			# ORPS xmm, xmm/m128: 0F 56 /r
			# ORPD xmm, xmm/m128: 66 0F 56 /r

			# XORPS xmm, xmm/m128: 0F 57 /r
			# XORPD xmm, xmm/m128: 66 0F 57 /r

			# UNPCKLPS xmm, xmm/m128: 0F 14 /r
			# UNPCKLPD xmm, xmm/m128: 66 0F 14 /r

			# UNPCKHPS xmm, xmm/m128: 0F 15 /r
			# UNPCKHPD xmm, xmm/m128: 66 0F 15 /r

			# ADDSUBPS xmm, xmm/m128: F2 0F D0 /r
			# ADDSUBPD xmm, xmm/m128: 66 0F D0 /r
			#   HADDPS xmm, xmm/m128: F2 0F 7C /r
			#   HADDPD xmm, xmm/m128: 66 0F 7C /r
			#   HSUBPS xmm, xmm/m128: F2 0F 7D /r
			#   HSUBPD xmm, xmm/m128: 66 0F 7D /r

			if name in sse_vector_instructions:
				self.size = 3 + source.get_modrm_extra_length()
			elif name in sse2_vector_instructions:
				self.size = 4 + source.get_modrm_extra_length()
			elif name in sse3_vector_instructions:
				self.size = 4 + source.get_modrm_extra_length()
		elif name in sse_scalar_instructions and destination.is_sse_register() and (source.is_sse_register() or source.is_memory_address32()):
			# ADDSS xmm, xmm/m128: F3 0F 58 /r
			# SUBSS xmm, xmm/m128: F3 0F 5C /r
			# MULSS xmm, xmm/m128: F3 0F 59 /r
			# DIVSS xmm, xmm/m128: F3 0F 5E /r
			# CMPSS xmm, xmm/m128: F3 0F C2 /r
			# MINSS xmm, xmm/m128: F3 0F 5D /r
			# MAXSS xmm, xmm/m128: F3 0F 5F /r

			self.size = 4 + source.get_modrm_extra_length()
		elif name in sse2_scalar_instructions and destination.is_sse_register() and (source.is_sse_register() or source.is_memory_address64()):
			# ADDSD xmm, xmm/m128: F2 0F 58 /r
			# SUBSD xmm, xmm/m128: F2 0F 5C /r
			# MULSD xmm, xmm/m128: F2 0F 59 /r
			# DIVSD xmm, xmm/m128: F2 0F 5E /r
			# CMPSD xmm, xmm/m128: F2 0F C2 /r
			# MINSD xmm, xmm/m128: F2 0F 5D /r
			# MAXSD xmm, xmm/m128: F2 0F 5F /r

			self.size = 4 + source.get_modrm_extra_length()
		else:
			raise ValueError('Invalid operands in instruction {0} {1}, {2}'.format(name, destination, source))

	def get_input_registers_list(self):
		if self.name in [ 'XORPS', 'XORPD', 'ANDNPS', 'ANDNPD' ] and self.source == self.destination:
			return list()
		else:
			return self.source.get_registers_list() + self.destination.get_registers_list()

	def get_output_registers_list(self):
		return [self.destination.register]

	def get_constant(self):
		if self.source.is_constant():
			return self.source.constant
		else:
			return None

	def get_local_variable(self):
		if self.source.is_local_variable():
			return self.source.variable
		else:
			return None

class AvxFloatingPointScalarBinaryInstruction(Instruction):
	def __init__(self, name, destination, source_x, source_y):
		single_instructions = [ 'VADDSS', 'VSUBSS', 'VMINSS', 'VMAXSS', 'VMULSS', 'VDIVSS']
		double_instructions = [ 'VADDSD', 'VSUBSD', 'VMINSD', 'VMAXSD', 'VMULSD', 'VDIVSD']
		allowed_instructions = single_instructions + double_instructions
		self.destination = destination
		self.source_x = source_x
		self.source_y = source_y
		if name in allowed_instructions:
			super(AvxFloatingPointScalarBinaryInstruction, self).__init__(name, 'AVX')
		else:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})'.format(name, ", ".join(allowed_instructions)))
		if name in single_instructions and destination.is_sse_register() and source_x.is_sse_register() and (source_y.is_sse_register() or source_y.is_memory_address32()):
			# VADDSS xmm, xmm, xmm/m32: VEX.NDS.LIG.F3.0F.WIG 58 /r
			# VSUBSS xmm, xmm, xmm/m32: VEX.NDS.LIG.F3.0F.WIG 5C /r
			# VMULSS xmm, xmm, xmm/m32: VEX.NDS.LIG.F3.0F.WIG 59 /r
			# VDIVSS xmm, xmm, xmm/m32: VEX.NDS.LIG.F3.0F.WIG 5E /r
			# VMINSS xmm, xmm, xmm/m32: VEX.NDS.LIG.F3.0F.WIG 5D /r
			# VMAXSS xmm, xmm, xmm/m32: VEX.NDS.LIG.F3.0F.WIG 5F /r

			self.size = source_y.get_vex_extra_length() + 1 + source_y.get_modrm_extra_length()
		elif name in double_instructions and destination.is_sse_register() and source_x.is_sse_register() and (source_y.is_sse_register() or source_y.is_memory_address64()):
			# VADDSD xmm, xmm, xmm/m64: VEX.NDS.LIG.F2.0F.WIG 58 /r
			# VSUBSD xmm, xmm, xmm/m64: VEX.NDS.LIG.F2.0F.WIG 5C /r
			# VMULSD xmm, xmm, xmm/m64: VEX.NDS.LIG.F2.0F.WIG 59 /r
			# VDIVSD xmm, xmm, xmm/m64: VEX.NDS.LIG.F2.0F.WIG 5E /r
			# VMINSD xmm, xmm, xmm/m64: VEX.NDS.LIG.F2.0F.WIG 5D /r
			# VMAXSD xmm, xmm, xmm/m64: VEX.NDS.LIG.F2.0F.WIG 5F /r

			self.size = source_y.get_vex_extra_length() + 1 + source_y.get_modrm_extra_length()
		else:
			raise ValueError('Invalid operands in instruction {0} {1}, {2}, {3}'.format(name, destination, source_x, source_y))

	def __str__(self):
		return "{0} {1}, {2}, {3}".format(self.name, self.destination, self.source_x, self.source_y)

	def get_input_registers_list(self):
		return self.source_x.get_registers_list() + self.source_y.get_registers_list() + self.destination.get_registers_list()

	def get_output_registers_list(self):
		return [self.destination.register]

	def get_constant(self):
		if self.source_y.is_constant():
			return self.source_y.constant
		else:
			return None

	def get_local_variable(self):
		if self.source_y.is_local_variable():
			return self.source_y.variable
		else:
			return None

class AvxFloatingPointVectorBinaryInstruction(Instruction):
	def __init__(self, name, destination, source_x, source_y):
		allowed_instructions = [ 'VADDPS', 'VADDPD', 'VSUBPS', 'VSUBPD',
		                         'VMINPS', 'VMINPD', 'VMAXPS', 'VMAXPD',
		                         'VMULPS', 'VMULPD', 'VDIVPS', 'VDIVPD',
		                         'VANDPS', 'VANDPD', 'VANDNPS', 'VANDNPD',
		                         'VORPS', 'VORPD', 'VXORPS', 'VXORPD',
		                         'VUNPCKLPS', 'VUNPCKLPD', 'VUNPCKHPS', 'VUNPCKHPD',
		                         'VHADDPS', 'VHADDPD', 'VHSUBPS', 'VHSUBPD',
		                         'VADDSUBPS', 'VADDSUBPD']
		self.destination = destination
		self.source_x = source_x
		self.source_y = source_y
		if name in allowed_instructions:
			super(AvxFloatingPointVectorBinaryInstruction, self).__init__(name, 'AVX')
		else:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})'.format(name, ", ".join(allowed_instructions)))
		if destination.is_sse_register() and source_x.is_sse_register() and (source_y.is_sse_register() or source_y.is_memory_address128()):
			# VADDPS xmm, xmm, xmm/m128: VEX.NDS.128.0F.WIG 58 /r
			# VADDPD xmm, xmm, xmm/m128: VEX.NDS.128.66.0F.WIG 58 /r

			# VSUBPS xmm, xmm, xmm/m128: VEX.NDS.128.0F.WIG 5C /r
			# VSUBPD xmm, xmm, xmm/m128: VEX.NDS.128.66.0F.WIG 5C /r

			# VMULPS xmm, xmm, xmm/m128: VEX.NDS.128.0F.WIG 59 /r
			# VMULPD xmm, xmm, xmm/m128: VEX.NDS.128.66.0F.WIG 59 /r

			# VDIVPS xmm, xmm, xmm/m128: VEX.NDS.128.0F.WIG 5E /r
			# VDIVPD xmm, xmm, xmm/m128: VEX.NDS.128.66.0F.WIG 5E /r

			# VMINPS xmm, xmm, xmm/m128: VEX.NDS.128.0F.WIG 5D /r
			# VMINPD xmm, xmm, xmm/m128: VEX.NDS.128.66.0F.WIG 5D /r

			# VMAXPS xmm, xmm, xmm/m128: VEX.NDS.128.0F.WIG 5F /r
			# VMAXPD xmm, xmm, xmm/m128: VEX.NDS.128.66.0F.WIG 5F /r

			# VANDPS xmm, xmm, xmm/m128: VEX.NDS.128.0F.WIG 54 /r
			# VANDPD xmm, xmm, xmm/m128: VEX.NDS.128.66.0F.WIG 54 /r

			# VANDNPS xmm, xmm, xmm/m128: VEX.NDS.128.0F.WIG 55 /r
			# VANDNPD xmm, xmm, xmm/m128: VEX.NDS.128.66.0F.WIG 55 /r

			# VORPS xmm, xmm, xmm/m128: VEX.NDS.128.0F.WIG 56 /r
			# VORPD xmm, xmm, xmm/m128: VEX.NDS.128.66.0F.WIG 56 /r

			# VXORPS xmm, xmm, xmm/m128: VEX.NDS.128.0F.WIG 57 /r
			# VXORPD xmm, xmm, xmm/m128: VEX.NDS.128.66.0F.WIG 57 /r

			# VUNPCKLPS xmm, xmm, xmm/m128: VEX.NDS.128.0F.WIG 14 /r
			# VUNPCKLPD xmm, xmm, xmm/m128: VEX.NDS.128.66.0F.WIG 14 /r

			# VUNPCKHPS xmm, xmm, xmm/m128: VEX.NDS.128.0F.WIG 15 /r
			# VUNPCKHPD xmm, xmm, xmm/m128: VEX.NDS.128.66.0F.WIG 15 /r

			# VADDSUBPS xmm, xmm, xmm/m128: VEX.NDS.128.F2.0F.WIG D0 /r
			# VADDSUBPD xmm, xmm, xmm/m128: VEX.NDS.128.66.0F.WIG D0 /r
			#   VHADDPS xmm, xmm, xmm/m128: VEX.NDS.128.F2.0F.WIG 7C /r
			#   VHADDPD xmm, xmm, xmm/m128: VEX.NDS.128.66.0F.WIG 7C /r
			#   VHSUBPS xmm, xmm, xmm/m128: VEX.NDS.128.F2.0F.WIG 7D /r
			#   VHSUBPD xmm, xmm, xmm/m128: VEX.NDS.128.66.0F.WIG 7D /r

			self.size = source_y.get_vex_extra_length() + 1 + source_y.get_modrm_extra_length()
		elif destination.is_avx_register() and source_x.is_avx_register() and (source_y.is_avx_register() or source_y.is_memory_address256()):
			# VADDPS ymm, ymm, ymm/m256: VEX.NDS.256.0F.WIG 58 /r
			# VADDPD ymm, ymm, ymm/m256: VEX.NDS.256.66.0F.WIG 58 /r

			# VSUBPS ymm, ymm, ymm/m256: VEX.NDS.256.0F.WIG 5C /r
			# VSUBPD ymm, ymm, ymm/m256: VEX.NDS.256.66.0F.WIG 5C /r

			# VMULPS ymm, ymm, ymm/m256: VEX.NDS.256.0F.WIG 59 /r
			# VMULPD ymm, ymm, ymm/m256: VEX.NDS.256.66.0F.WIG 59 /r

			# VDIVPS ymm, ymm, ymm/m256: VEX.NDS.256.0F.WIG 5E /r
			# VDIVPD ymm, ymm, ymm/m256: VEX.NDS.256.66.0F.WIG 5E /r

			# VMINPS ymm, ymm, ymm/m256: VEX.NDS.256.0F.WIG 5D /r
			# VMINPD ymm, ymm, ymm/m256: VEX.NDS.256.66.0F.WIG 5D /r

			# VMAXPS ymm, ymm, ymm/m256: VEX.NDS.256.0F.WIG 5F /r
			# VMAXPD ymm, ymm, ymm/m256: VEX.NDS.256.66.0F.WIG 5F /r

			# VANDPS ymm, ymm, ymm/m256: VEX.NDS.256.0F.WIG 54 /r
			# VANDPD ymm, ymm, ymm/m256: VEX.NDS.256.66.0F.WIG 54 /r

			# VANDNPS ymm, ymm, ymm/m256: VEX.NDS.256.0F.WIG 55 /r
			# VANDNPD ymm, ymm, ymm/m256: VEX.NDS.256.66.0F.WIG 55 /r

			# VORPS ymm, ymm, ymm/m256: VEX.NDS.256.0F.WIG 56 /r
			# VORPD ymm, ymm, ymm/m256: VEX.NDS.256.66.0F.WIG 56 /r

			# VXORPS ymm, ymm, ymm/m256: VEX.NDS.256.0F.WIG 57 /r
			# VXORPD ymm, ymm, ymm/m256: VEX.NDS.256.66.0F.WIG 57 /r

			# VUNPCKLPS ymm, ymm, ymm/m256: VEX.NDS.256.0F.WIG 14 /r
			# VUNPCKLPD ymm, ymm, ymm/m256: VEX.NDS.256.66.0F.WIG 14 /r

			# VUNPCKHPS ymm, ymm, ymm/m256: VEX.NDS.256.0F.WIG 15 /r
			# VUNPCKHPD ymm, ymm, ymm/m256: VEX.NDS.256.66.0F.WIG 15 /r

			# VADDSUBPS ymm, ymm, ymm/m256: VEX.NDS.256.F2.0F.WIG D0 /r
			# VADDSUBPD ymm, ymm, ymm/m256: VEX.NDS.256.66.0F.WIG D0 /r
			#   VHADDPS ymm, ymm, ymm/m256: VEX.NDS.256.F2.0F.WIG 7C /r
			#   VHADDPD ymm, ymm, ymm/m256: VEX.NDS.256.66.0F.WIG 7C /r
			#   VHSUBPS ymm, ymm, ymm/m256: VEX.NDS.256.F2.0F.WIG 7D /r
			#   VHSUBPD ymm, ymm, ymm/m256: VEX.NDS.256.66.0F.WIG 7D /r

			self.size = source_y.get_vex_extra_length() + 1 + source_y.get_modrm_extra_length()
		else:
			raise ValueError('Invalid operands in instruction {0} {1}, {2}, {3}'.format(name, destination, source_x, source_y))

	def __str__(self):
		return "{0} {1}, {2}, {3}".format(self.name, self.destination, self.source_x, self.source_y)

	def get_input_registers_list(self):
		if self.name in ['VXORPS', 'VXORPD', 'VANDNPS', 'VANDNPD']:
			return list()
		else:
			return self.source_x.get_registers_list() + self.source_y.get_registers_list()

	def get_output_registers_list(self):
		return [self.destination.register]

	def get_constant(self):
		if self.source_y.is_constant():
			return self.source_y.constant
		else:
			return None

	def get_local_variable(self):
		if self.source_y.is_local_variable():
			return self.source_y.variable
		else:
			return None

class Fma4ScalarInstruction(Instruction):
	def __init__(self, name, destination, source_x, source_y, source_z):
		single_precision_instructions = [ 'VFMADDSS', 'VFMSUBSS', 'VFNMADDSS', 'VFNMSUBSS' ]
		double_precision_instructions = [ 'VFMADDSD', 'VFMSUBSD', 'VFNMADDSD', 'VFNMSUBSD' ]
		allowed_instructions = single_precision_instructions + double_precision_instructions
		self.destination = destination
		self.source_x = source_x
		self.source_y = source_y
		self.source_z = source_z
		if name in allowed_instructions:
			super(Fma4ScalarInstruction, self).__init__(name, 'FMA4')
		else:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})'.format(name, ", ".join(allowed_instructions)))
		if name in single_precision_instructions and destination.is_sse_register() and source_x.is_sse_register() and (source_y.is_sse_register() or source_y.is_memory_address32()) and source_z.is_sse_register():
			#  VFMADDSS xmm, xmm, xmm/m128, xmm: VEX.NDS.128.66.0F3A.W0 6A /r /is4
			#  VFMSUBSS xmm, xmm, xmm/m128, xmm: VEX.NDS.128.66.0F3A.W0 6E /r /is4
			# VFNMADDSS xmm, xmm, xmm/m128, xmm: VEX.NDS.128.66.0F3A.W0 7A /r /is4
			# VFNMSUBSS xmm, xmm, xmm/m128, xmm: VEX.NDS.128.66.0F3A.W0 7E /r /is4

			self.size = 3 + 1 + source_y.get_modrm_extra_length()
		elif name in single_precision_instructions and destination.is_sse_register() and source_x.is_sse_register() and source_y.is_sse_register() and (source_z.is_sse_register() or source_z.is_memory_address32()):
			#  VFMADDSS xmm, xmm, xmm, xmm/m128: VEX.NDS.128.66.0F3A.W1 6A /r /is4
			#  VFMSUBSS xmm, xmm, xmm, xmm/m128: VEX.NDS.128.66.0F3A.W1 6E /r /is4
			# VFNMADDSS xmm, xmm, xmm, xmm/m128: VEX.NDS.128.66.0F3A.W1 7A /r /is4
			# VFNMSUBSS xmm, xmm, xmm, xmm/m128: VEX.NDS.128.66.0F3A.W1 7E /r /is4

			self.size = 3 + 1 + source_z.get_modrm_extra_length()
		elif name in double_precision_instructions and destination.is_sse_register() and source_x.is_sse_register() and (source_y.is_sse_register() or source_y.is_memory_address64()) and source_z.is_sse_register():
			#  VFMADDSD xmm, xmm, xmm/m128, xmm: VEX.NDS.128.66.0F3A.W0 6B /r /is4
			#  VFMSUBSD xmm, xmm, xmm/m128, xmm: VEX.NDS.128.66.0F3A.W0 6F /r /is4
			# VFNMADDSD xmm, xmm, xmm/m128, xmm: VEX.NDS.128.66.0F3A.W0 7B /r /is4
			# VFNMSUBSD xmm, xmm, xmm/m128, xmm: VEX.NDS.128.66.0F3A.W0 7F /r /is4

			self.size = 3 + 1 + source_y.get_modrm_extra_length()
		elif name in double_precision_instructions and destination.is_sse_register() and source_x.is_sse_register() and source_y.is_sse_register() and (source_z.is_sse_register() or source_z.is_memory_address64()):
			#  VFMADDSD xmm, xmm, xmm, xmm/m128: VEX.NDS.128.66.0F3A.W1 6B /r /is4
			#  VFMSUBSD xmm, xmm, xmm, xmm/m128: VEX.NDS.128.66.0F3A.W1 6F /r /is4
			# VFNMADDSD xmm, xmm, xmm, xmm/m128: VEX.NDS.128.66.0F3A.W1 7B /r /is4
			# VFNMSUBSD xmm, xmm, xmm, xmm/m128: VEX.NDS.128.66.0F3A.W1 7F /r /is4

			self.size = 3 + 1 + source_z.get_modrm_extra_length()
		else:
			raise ValueError('Invalid operands in instruction {0} {1}, {2}, {3}'.format(name, destination, source_x, source_y, source_z))

	def __str__(self):
		return "{0} {1}, {2}, {3}, {4}".format(self.name, self.destination, self.source_x, self.source_y, self.source_z)

	def get_input_registers_list(self):
		return self.source_x.get_registers_list() + self.source_y.get_registers_list() + self.source_z.get_registers_list()

	def get_output_registers_list(self):
		return [self.destination.register]

	def get_constant(self):
		if self.source_y.is_constant():
			return self.source_y.constant
		elif self.source_z.is_constant():
			return self.source_z.constant
		else:
			return None

	def get_local_variable(self):
		if self.source_y.is_local_variable():
			return self.source_y.variable
		elif self.source_z.is_local_variable():
			return self.source_z.variable
		else:
			return None

class Fma4VectorInstruction(Instruction):
	def __init__(self, name, destination, source_x, source_y, source_z):
		allowed_instructions = [ 'VFMADDPS', 'VFMADDPD', 'VFMSUBPS', 'VFMSUBPD',
		                         'VFNMADDPS', 'VFNMADDPD', 'VFNMSUBPS', 'VFNMSUBPD',
		                         'VFMADDSUBPS', 'VFMADDSUBPD', 'VFMSUBADDPS', 'VFMSUBADDPS']
		self.destination = destination
		self.source_x = source_x
		self.source_y = source_y
		self.source_z = source_z
		if name in allowed_instructions:
			super(Fma4VectorInstruction, self).__init__(name, 'FMA4')
		else:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})'.format(name, ", ".join(allowed_instructions)))
		if destination.is_sse_register() and source_x.is_sse_register() and (source_y.is_sse_register() or source_y.is_memory_address128()) and source_z.is_sse_register():
			# VFMADDPS xmm, xmm, xmm/m128, xmm: VEX.NDS.128.66.0F3A.W0 68 /r /is4
			# VFMADDPD xmm, xmm, xmm/m128, xmm: VEX.NDS.128.66.0F3A.W0 69 /r /is4

			# VFMSUBPS xmm, xmm, xmm/m128, xmm: VEX.NDS.128.66.0F3A.W0 6C /r /is4
			# VFMSUBPD xmm, xmm, xmm/m128, xmm: VEX.NDS.128.66.0F3A.W0 6D /r /is4

			# VFNMADDPS xmm, xmm, xmm/m128, xmm: VEX.NDS.128.66.0F3A.W0 78 /r /is4
			# VFNMADDPD xmm, xmm, xmm/m128, xmm: VEX.NDS.128.66.0F3A.W0 79 /r /is4

			# VFNMSUBPS xmm, xmm, xmm/m128, xmm: VEX.NDS.128.66.0F3A.W0 7C /r /is4
			# VFNMSUBPD xmm, xmm, xmm/m128, xmm: VEX.NDS.128.66.0F3A.W0 7D /r /is4

			# VFMADDSUBPS xmm, xmm, xmm/m128, xmm: VEX.NDS.128.66.0F3A.W0 5C /r /is4
			# VFMADDSUBPD xmm, xmm, xmm/m128, xmm: VEX.NDS.128.66.0F3A.W0 5D /r /is4

			# VFMSUBADDPS xmm, xmm, xmm/m128, xmm: VEX.NDS.128.66.0F3A.W0 5E /r /is4
			# VFMSUBADDPD xmm, xmm, xmm/m128, xmm: VEX.NDS.128.66.0F3A.W0 5F /r /is4

			self.size = 3 + 1 + source_y.get_modrm_extra_length()
		elif destination.is_sse_register() and source_x.is_sse_register() and source_y.is_sse_register() and (source_z.is_sse_register() or source_z.is_memory_address128()):
			# VFMADDPS xmm, xmm, xmm, xmm/m128: VEX.NDS.128.66.0F3A.W1 68 /r /is4
			# VFMADDPD xmm, xmm, xmm, xmm/m128: VEX.NDS.128.66.0F3A.W1 69 /r /is4

			# VFMSUBPS xmm, xmm, xmm, xmm/m128: VEX.NDS.128.66.0F3A.W1 6C /r /is4
			# VFMSUBPD xmm, xmm, xmm, xmm/m128: VEX.NDS.128.66.0F3A.W1 6D /r /is4

			# VFNMADDPS xmm, xmm, xmm, xmm/m128: VEX.NDS.128.66.0F3A.W1 78 /r /is4
			# VFNMADDPD xmm, xmm, xmm, xmm/m128: VEX.NDS.128.66.0F3A.W1 79 /r /is4

			# VFNMSUBPS xmm, xmm, xmm, xmm/m128: VEX.NDS.128.66.0F3A.W1 7C /r /is4
			# VFNMSUBPD xmm, xmm, xmm, xmm/m128: VEX.NDS.128.66.0F3A.W1 7D /r /is4

			# VFMADDSUBPS xmm, xmm, xmm, xmm/m128: VEX.NDS.128.66.0F3A.W1 5C /r /is4
			# VFMADDSUBPD xmm, xmm, xmm, xmm/m128: VEX.NDS.128.66.0F3A.W1 5D /r /is4

			# VFMSUBADDPS xmm, xmm, xmm, xmm/m128: VEX.NDS.128.66.0F3A.W1 5E /r /is4
			# VFMSUBADDPD xmm, xmm, xmm, xmm/m128: VEX.NDS.128.66.0F3A.W1 5F /r /is4

			self.size = 3 + 1 + source_z.get_modrm_extra_length()
		elif destination.is_avx_register() and source_x.is_avx_register() and (source_y.is_avx_register() or source_y.is_memory_address256()) and source_z.is_avx_register():
			# VFMADDPS ymm, ymm, ymm/m256, ymm: VEX.NDS.256.66.0F3A.W0 68 /r /is4
			# VFMADDPD ymm, ymm, ymm/m256, ymm: VEX.NDS.256.66.0F3A.W0 69 /r /is4

			# VFMSUBPS ymm, ymm, ymm/m256, ymm: VEX.NDS.256.66.0F3A.W0 6C /r /is4
			# VFMSUBPD ymm, ymm, ymm/m256, ymm: VEX.NDS.256.66.0F3A.W0 6D /r /is4

			# VFNMADDPS ymm, ymm, ymm/m256, ymm: VEX.NDS.256.66.0F3A.W0 78 /r /is4
			# VFNMADDPD ymm, ymm, ymm/m256, ymm: VEX.NDS.256.66.0F3A.W0 79 /r /is4

			# VFNMSUBPS ymm, ymm, ymm/m256, ymm: VEX.NDS.256.66.0F3A.W0 7C /r /is4
			# VFNMSUBPD ymm, ymm, ymm/m256, ymm: VEX.NDS.256.66.0F3A.W0 7D /r /is4

			# VFMADDSUBPS ymm, ymm, ymm/m256, ymm: VEX.NDS.256.66.0F3A.W0 5C /r /is4
			# VFMADDSUBPD ymm, ymm, ymm/m256, ymm: VEX.NDS.256.66.0F3A.W0 5D /r /is4

			# VFMSUBADDPS ymm, ymm, ymm/m256, ymm: VEX.NDS.256.66.0F3A.W0 5E /r /is4
			# VFMSUBADDPD ymm, ymm, ymm/m256, ymm: VEX.NDS.256.66.0F3A.W0 5F /r /is4

			self.size = 3 + 1 + source_y.get_modrm_extra_length()
		elif destination.is_avx_register() and source_x.is_avx_register() and source_y.is_avx_register() and (source_z.is_avx_register() or source_z.is_memory_address256()):
			# VFMADDPS ymm, ymm, ymm, ymm/m256: VEX.NDS.256.66.0F3A.W1 68 /r /is4
			# VFMADDPD ymm, ymm, ymm, ymm/m256: VEX.NDS.256.66.0F3A.W1 69 /r /is4

			# VFMSUBPS ymm, ymm, ymm, ymm/m256: VEX.NDS.256.66.0F3A.W1 6C /r /is4
			# VFMSUBPD ymm, ymm, ymm, ymm/m256: VEX.NDS.256.66.0F3A.W1 6D /r /is4

			# VFNMADDPS ymm, ymm, ymm, ymm/m256: VEX.NDS.256.66.0F3A.W1 78 /r /is4
			# VFNMADDPD ymm, ymm, ymm, ymm/m256: VEX.NDS.256.66.0F3A.W1 79 /r /is4

			# VFNMSUBPS ymm, ymm, ymm, ymm/m256: VEX.NDS.256.66.0F3A.W1 7C /r /is4
			# VFNMSUBPD ymm, ymm, ymm, ymm/m256: VEX.NDS.256.66.0F3A.W1 7D /r /is4

			# VFMADDSUBPS ymm, ymm, ymm, ymm/m256: VEX.NDS.256.66.0F3A.W1 5C /r /is4
			# VFMADDSUBPD ymm, ymm, ymm, ymm/m256: VEX.NDS.256.66.0F3A.W1 5D /r /is4

			# VFMSUBADDPS ymm, ymm, ymm, ymm/m256: VEX.NDS.256.66.0F3A.W1 5E /r /is4
			# VFMSUBADDPD ymm, ymm, ymm, ymm/m256: VEX.NDS.256.66.0F3A.W1 5F /r /is4

			self.size = 3 + 1 + source_z.get_modrm_extra_length()
		else:
			raise ValueError('Invalid operands in instruction {0} {1}, {2}, {3}'.format(name, destination, source_x, source_y, source_z))

	def __str__(self):
		return "{0} {1}, {2}, {3}, {4}".format(self.name, self.destination, self.source_x, self.source_y, self.source_z)

	def get_input_registers_list(self):
		return self.source_x.get_registers_list() + self.source_y.get_registers_list() + self.source_z.get_registers_list()

	def get_output_registers_list(self):
		return [self.destination.register]

	def get_constant(self):
		if self.source_y.is_constant():
			return self.source_y.constant
		elif self.source_z.is_constant():
			return self.source_z.constant
		else:
			return None

	def get_local_variable(self):
		if self.source_y.is_local_variable():
			return self.source_y.variable
		elif self.source_z.is_local_variable():
			return self.source_z.variable
		else:
			return None

class SseIntegerBinaryInstruction(BinaryInstruction):
	def __init__(self, name, destination, source):
		sse2_instructions = [ 'PADDB', 'PADDW', 'PADDD', 'PADDQ', 'PSUBB', 'PSUBW', 'PSUBD', 'PSUBQ',
							  'PADDSB', 'PADDSW', 'PADDUSB', 'PADDUSW',
							  'PSUBSB', 'PSUBSW', 'PSUBUSB', 'PSUBUSW',
							  'PAVGB', 'PAVGW', 'PMAXUB', 'PMAXSW', 'PMINUB', 'PMINSW', 'PSADBW',
							  'PAND', 'PANDN', 'POR', 'PXOR',
							  'PCMPEQB', 'PCMPEQW', 'PCMPEQD', 'PCMPGTB', 'PCMPGTW', 'PCMPGTD',
							  'PMULUDQ', 'PMULLW', 'PMULHW', 'PMULHUW', 'PMADDWD',
							  'PUNPCKLBW', 'PUNPCKHBW', 'PUNPCKLWD', 'PUNPCKHWD', 'PUNPCKLDQ', 'PUNPCKHDQ',
							  'PUNPCKLQDQ', 'PUNPCKHQDQ', 'PACKSSWB', 'PACKSSDW', 'PACKUSWB']
		ssse3_instructions = [ 'PSIGNB', 'PSIGNW', 'PSIGND', 'PABSB', 'PABSW', 'PABSD',
							   'PHADDW', 'PHADDD', 'PHSUBW', 'PHSUBD', 'PHADDSW', 'PHSUBSW',
							   'PSHUFB', 'PMULHRSW', 'PMADDUBSW']
		sse41_instructions = [ 'PCMPEQQ', 'PACKUSDW', 'PTEST',
							   'PMULLD', 'PMULDQ',
							   'PMAXSB', 'PMAXSD', 'PMAXUW', 'PMAXUD',
							   'PMINSB', 'PMINSD', 'PMINUW', 'PMINUD', 'PHMINPOSUW']
		sse42_instructions = [ 'PCMPGTQ' ]
		aes_instructions = [ 'AESIMC', 'AESENC', 'AESENCLAST', 'AESDEC', 'AESDECLAST' ]
		if name in sse2_instructions:
			super(SseIntegerBinaryInstruction, self).__init__(name, destination, source, 'SSE2')
		elif name in ssse3_instructions:
			super(SseIntegerBinaryInstruction, self).__init__(name, destination, source, 'SSSE3')
		elif name in sse41_instructions:
			super(SseIntegerBinaryInstruction, self).__init__(name, destination, source, 'SSE4.1')
		elif name in sse42_instructions:
			super(SseIntegerBinaryInstruction, self).__init__(name, destination, source, 'SSE4.2')
		elif name in aes_instructions:
			super(SseIntegerBinaryInstruction, self).__init__(name, destination, source, 'AES')
		else:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})'.format(name, ", ".join(sse2_instructions + ssse3_instructions)))
		if destination.is_sse_register() and (source.is_sse_register() or source.is_memory_address128()):
			# PADDB xmm, xmm/m128: 66 0F FC /r
			# PADDW xmm, xmm/m128: 66 0F FD /r
			# PADDD xmm, xmm/m128: 66 0F FE /r
			# PADDQ xmm, xmm/m128: 66 0F D4 /r

			# PSUBB xmm, xmm/m128: 66 0F F8 /r
			# PSUBW xmm, xmm/m128: 66 0F F9 /r
			# PSUBD xmm, xmm/m128: 66 0F FA /r
			# PSUBQ xmm, xmm/m128: 66 0F FB /r

			#  PADDSB xmm, xmm/m128: 66 0F EC /r
			#  PADDSW xmm, xmm/m128: 66 0F ED /r
			# PADDUSB xmm, xmm/m128: 66 0F DC /r
			# PADDUSW xmm, xmm/m128: 66 0F DD /r

			#  PSUBSB xmm, xmm/m128: 66 0F E8 /r
			#  PSUBSW xmm, xmm/m128: 66 0F E9 /r
			# PSUBUSB xmm, xmm/m128: 66 0F D8 /r
			# PSUBUSW xmm, xmm/m128: 66 0F D9 /r

			#  PAVGB xmm, xmm/m128: 66 0F E0 /r
			#  PAVGW xmm, xmm/m128: 66 0F E3 /r
			# PMAXUB xmm, xmm/m128: 66 0F DE /r
			# PMAXSW xmm, xmm/m128: 66 0F EE /r
			# PMINUB xmm, xmm/m128: 66 0F DA /r
			# PMINSW xmm, xmm/m128: 66 0F EA /r

			#  PAND xmm, xmm/m128: 66 0F DB /r
			# PANDN xmm, xmm/m128: 66 0F DF /r
			#   POR xmm, xmm/m128: 66 0F EB /r
			#  PXOR xmm, xmm/m128: 66 0F EF /r

			# PCMPEQB xmm, xmm/m128: 66 0F 74 /r
			# PCMPEQW xmm, xmm/m128: 66 0F 75 /r
			# PCMPEQD xmm, xmm/m128: 66 0F 76 /r
			# PCMPGTB xmm, xmm/m128: 66 0F 64 /r
			# PCMPGTW xmm, xmm/m128: 66 0F 65 /r
			# PCMPGTD xmm, xmm/m128: 66 0F 66 /r

			# PMULUDQ xmm, xmm/m128: 66 0F F4 /r
			#  PMULLW xmm, xmm/m128: 66 0F D5 /r
			#  PMULHW xmm, xmm/m128: 66 0F E5 /r
			# PMULHUW xmm, xmm/m128: 66 0F E4 /r

			# PUNPCKLBW xmm, xmm/m128: 66 0F 60 /r
			# PUNPCKHBW xmm, xmm/m128: 66 0F 68 /r
			# PUNPCKLWD xmm, xmm/m128: 66 0F 61 /r
			# PUNPCKHWD xmm, xmm/m128: 66 0F 69 /r
			# PUNPCKLDQ xmm, xmm/m128: 66 0F 62 /r
			# PUNPCKHDQ xmm, xmm/m128: 66 0F 6A /r

			# PUNPCKLQDQ xmm, xmm/m128: 66 0F 6C /r
			# PUNPCKHQDQ xmm, xmm/m128: 66 0F 6D /r
			#   PACKSSWB xmm, xmm/m128: 66 0F 63 /r
			#   PACKSSDW xmm, xmm/m128: 66 0F 6B /r
			#   PACKUSWB xmm, xmm/m128: 66 0F 67 /r

			# PSIGNB xmm, xmm/m128: 66 0F 38 08 /r
			# PSIGNW xmm, xmm/m128: 66 0F 38 09 /r
			# PSIGND xmm, xmm/m128: 66 0F 38 0A /r
			#  PABSB xmm, xmm/m128: 66 0F 38 1C /r
			#  PABSW xmm, xmm/m128: 66 0F 38 1D /r
			#  PABSD xmm, xmm/m128: 66 0F 38 1E /r

			#  PHADDW xmm, xmm/m128: 66 0F 38 01 /r
			#  PHADDD xmm, xmm/m128: 66 0F 38 02 /r
			# PHADDSW xmm, xmm/m128: 66 0F 38 03 /r
			#  PHSUBW xmm, xmm/m128: 66 0F 38 05 /r
			#  PHSUBD xmm, xmm/m128: 66 0F 38 06 /r
			# PHSUBSW xmm, xmm/m128: 66 0F 38 07 /r

			#    PSHUFB xmm, xmm/m128: 66 0F 38 00 /r
			#  PMULHRSW xmm, xmm/m128: 66 0F 38 0B /r
			# PMADDUBSW xmm, xmm/m128: 66 0F 38 04 /r

			#   PCMPEQQ xmm, xmm/m128: 66 0F 38 29 /r
			#  PACKUSDW xmm, xmm/m128: 66 0F 38 2B /r
			#     PTEST xmm, xmm/m128: 66 0F 38 17 /r
			#    PMULLD xmm, xmm/m128: 66 0F 38 40 /r
			#    PMULDQ xmm, xmm/m128: 66 0F 38 28 /r

			# PMAXSB xmm, xmm/m128: 66 0F 38 3C /r
			# PMAXSD xmm, xmm/m128: 66 0F 38 3D /r
			# PMAXUW xmm, xmm/m128: 66 0F 38 3E /r
			# PMAXUD xmm, xmm/m128: 66 0F 38 3F /r

			#     PMINSB xmm, xmm/m128: 66 0F 38 38 /r
			#     PMINSD xmm, xmm/m128: 66 0F 38 39 /r
			#     PMINUW xmm, xmm/m128: 66 0F 38 3A /r
			#     PMINUD xmm, xmm/m128: 66 0F 38 3B /r
			# PHMINPOSUW xmm, xmm/m128: 66 0F 38 41 /r

			# PCMPGTQ xmm, xmm/m128: 66 0F 38 37 /r

			#     AESIMC xmm, xmm/m128: 66 0F 38 DB /r
			#     AESENC xmm, xmm/m128: 66 0F 38 DC /r
			# AESENCLAST xmm, xmm/m128: 66 0F 38 DD /r
			#     AESDEC xmm, xmm/m128: 66 0F 38 DE /r
			# AESDECLAST xmm, xmm/m128: 66 0F 38 DF /r

			if name in sse2_instructions:
				self.size = 4 + source.get_modrm_extra_length()
			else:
				self.size = 5 + source.get_modrm_extra_length()
		else:
			raise ValueError('Invalid operands in instruction {0} {1}, {2}'.format(name, destination, source))

	def get_input_registers_list(self):
		if self.destination.is_register():
			if self.name in ['PANDN', 'PXOR',
			                 'PCMPEQB', 'PCMPEQW', 'PCMPEQD', 'PCMPEQQ',
			                 'PCMPGTB', 'PCMPGTW', 'PCMPGTD', 'PCMPGTQ',
			                 'AESIMC', 'AESENC', 'AESENCLAST', 'AESDEC', 'AESDECLAST']:
				return self.source.get_registers_list()
			else:
				return self.destination.get_registers_list() + self.source.get_registers_list()
		else:
			return self.source.get_registers_list()

	def get_output_registers_list(self):
		return [self.destination.register]

	def get_constant(self):
		if self.source.is_constant():
			return self.source.constant
		else:
			return None

	def get_local_variable(self):
		if self.source.is_local_variable():
			return self.source.variable
		else:
			return None

class MmxSseShiftInstruction(BinaryInstruction):
	def __init__(self, name, destination, source):
		shift_instructions = ['PSLLW', 'PSLLD', 'PSLLQ',
		                      'PSRLW', 'PSRLD', 'PSRLQ',
		                      'PSRAW', 'PSRAD']
		if name in shift_instructions:
			super(MmxSseShiftInstruction, self).__init__(name, destination, source, "SSE2")
		else:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})'.format(name, ", ".join(shift_instructions)))
		if destination.is_sse_register() and (source.is_sse_register() or source.is_memory_address128()):
			# PSLLW xmm, xmm/m128: 66 0F F1 /r
			# PSLLD xmm, xmm/m128: 66 0F F2 /r
			# PSLLQ xmm, xmm/m128: 66 0F F3 /r
			# PSRLW xmm, xmm/m128: 66 0F D1 /r
			# PSRLD xmm, xmm/m128: 66 0F D2 /r
			# PSRLQ xmm, xmm/m128: 66 0F D3 /r
			# PSRAW xmm, xmm/m128: 66 0F E1 /r
			# PSRAD xmm, xmm/m128: 66 0F E2 /r
			self.size = 4 + source.get_modrm_extra_length(source)
		elif destination.is_sse_register() and source.is_int8():
			# PSLLW xmm, imm8: 66 0F 71 /6 ib
			# PSLLD xmm, imm8: 66 0F 72 /6 ib
			# PSLLQ xmm, imm8: 66 0F 73 /6 ib
			# PSRLW xmm, imm8: 66 0F 71 /2 ib
			# PSRLD xmm, imm8: 66 0F 72 /2 ib
			# PSRLQ xmm, imm8: 66 0F 73 /2 ib
			# PSRAW xmm, imm8: 66 0F 71 /4 ib
			# PSRAD xmm, imm8: 66 0F 72 /4 ib
			self.size = 4 + 1
		else:
			raise ValueError('Invalid operands in instruction {0} {1}, {2}'.format(name, destination, source))

	def get_input_registers_list(self):
		return self.destination.get_registers_list() + self.source.get_registers_list()
		
	def get_output_registers_list(self):
		return [self.destination.register]

	def get_constant(self):
		if self.source.is_constant():
			return self.source.constant
		else:
			return None

	def get_local_variable(self):
		if self.source.is_local_variable():
			return self.source.variable
		else:
			return None

class MmxSseShuffleInstruction(Instruction):
	def __init__(self, name, destination, source, immediate):
		mmx_shuffle_instructions = ['PSHUFW']
		sse_shuffle_instructions = ['SHUFPS']
		sse2_shuffle_instructions = ['SHUFPD', 'PSHUFD', 'PSHUFLW', 'PSHUFHW']
		if name in mmx_shuffle_instructions:
			super(MmxSseShuffleInstruction, self).__init__(name, "MMX+")
		elif name in sse_shuffle_instructions:
			super(MmxSseShuffleInstruction, self).__init__(name, "SSE2")
		else:
			raise ValueError('Instruction {0} is not one of the MMX/SSE shuffle instructions ({1})'.format(name, ", ".join(mmx_shuffle_instructions + sse_shuffle_instructions + sse2_shuffle_instructions)))
		if name in mmx_shuffle_instructions and destination.is_mmx_register() and (source.is_mmx_register() or source.is_memory_address64()) and immediate.is_uint8():
			# PSHUFW mm, mm/m64, imm8: 0F 70 /r ib
			self.size = 3 + 1 + source.get_modrm_extra_length()
		elif name in (sse_shuffle_instructions + sse2_shuffle_instructions) and destination.is_sse_register() and (source.is_sse_register() or source.is_memory_address128()) and immediate.is_uint8():
			# PSHUFLW xmm, xmm/m128, imm8: F2 0F 70 /r ib
			# PSHUFHW xmm, xmm/m128, imm8: F3 0F 70 /r ib
			# PSHUFD  xmm, xmm/m128, imm8: 66 0F 70 /r ib
			self.size = 1 + 3 + 1 + source.get_modrm_extra_length()
		else:
			raise ValueError('Invalid operands in instruction {0} {1}, {2}, {3}'.format(name, destination, source, immediate))

	def __str__(self):
		return "{0} {1}, {2}, {3}".format(self.name, self.destination, self.source, self.immediate)

	def get_input_registers_list(self):
		input_registers_list = self.source.get_registers_list()
		if self.name in ['SHUFPS', 'SHUFPD']:
			input_registers_list += self.destination.get_registers_list()
		return input_registers_list

	def get_output_registers_list(self):
		return self.destination.get_registers_list()

	def get_constant(self):
		if self.source.is_constant():
			return self.source.constant
		else:
			return None

	def get_local_variable(self):
		if self.source.is_local_variable():
			return self.source.variable
		else:
			return None

class SseOctowordShiftInstruction(BinaryInstruction):
	def __init__(self, name, destination, source):
		shift_instructions = ['PSLLDQ', 'PSRLDQ']
		if name in shift_instructions:
			super(SseOctowordShiftInstruction, self).__init__(name, destination, source, "SSE2")
		else:
			raise ValueError('Instruction {0} is not one of the octoword shift instructions ({1})'.format(name, ", ".join(shift_instructions)))
		if destination.is_sse_register() and source.is_int8():
			# PSLLDQ xmm, imm8: 66 0F 73 /7 ib
			# PSRLDQ xmm, imm8: 66 0F 73 /3 ib
			self.size = 4 + 1
		else:
			raise ValueError('Invalid operands in instruction {0} {1}, {2}'.format(name, destination, source))

	def get_constant(self):
		return None

	def get_local_variable(self):
		return None

class SseMovExtendInstruction(BinaryInstruction):
	def __init__(self, name, destination, source):
		m16_loads = [ 'PMOVSXBQ', 'PMOVZXBQ' ]
		m32_loads = [ 'PMOVSXBD', 'PMOVZXBD', 'PMOVSXWQ', 'PMOVZXWQ' ]
		m64_loads = [ 'PMOVSXBW', 'PMOVZXBW', 'PMOVSXWD', 'PMOVZXWD', 'PMOVSXDQ', 'PMOVZXDQ' ]
		mov_extend_instructions = m16_loads + m32_loads + m64_loads
		if name in mov_extend_instructions:
			super(SseMovExtendInstruction, self).__init__(name, destination, source, 'SSE4.1')
		else:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})', name, ", ".join(mov_extend_instructions))
		if destination.is_sse_register() and (source.is_sse_register() or source.is_memory_address16()) and name in m16_loads:
			# PMOVSXBQ xmm, xmm/m16: 66 0F 38 22 /r
			# PMOVZXBQ xmm, xmm/m16: 66 0F 38 32 /r
			self.size = 5 + source.get_modrm_extra_length()
		elif destination.is_sse_register() and (source.is_sse_register() or source.is_memory_address32()) and name in m32_loads:
			# PMOVSXBD xmm, xmm/m32: 66 0F 38 21 /r
			# PMOVZXBD xmm, xmm/m32: 66 0F 38 31 /r
			# PMOVSXWQ xmm, xmm/m32: 66 0F 38 24 /r
			# PMOVZXWQ xmm, xmm/m32: 66 0F 38 34 /r
			self.size = 5 + source.get_modrm_extra_length()
		elif destination.is_sse_register() and (source.is_sse_register() or source.is_memory_address64()) and name in m64_loads:
			# PMOVSXBW xmm, xmm/m64: 66 0F 38 20 /r
			# PMOVZXBW xmm, xmm/m64: 66 0F 38 30 /r
			# PMOVSXWD xmm, xmm/m64: 66 0F 38 23 /r
			# PMOVZXWD xmm, xmm/m64: 66 0F 38 33 /r
			# PMOVSXDQ xmm, xmm/m64: 66 0F 38 25 /r
			# PMOVZXDQ xmm, xmm/m64: 66 0F 38 35 /r
			self.size = 5 + source.get_modrm_extra_length()
		else:
			raise ValueError('Invalid operands in instruction {0} {1}, {2}'.format(name, destination, source))

	def get_input_registers_list(self):
		return self.source.get_registers_list()

	def get_output_registers_list(self):
		return [self.destination.register]

	def get_constant(self):
		if self.source.is_constant():
			return self.source.constant
		else:
			return None

	def get_local_variable(self):
		if self.source.is_local_variable():
			return self.source.variable
		else:
			return None

class SseBlendMaskedInstruction(Instruction):
	def __init__(self, name, destination, source, mask):
		allowed_instructions = ['BLENDVPS', 'BLENDVPD', 'PBLENDV']
		if name in allowed_instructions:
			super(SseBlendMaskedInstruction, self).__init__(name, 'SSE4.1')
		else:
			raise ValueError('Instruction {0} is not one of the allowed instructions ({1})', name, ", ".join(allowed_instructions))
		if destination.is_sse_register() and (source.is_sse_register() or source.is_memory_address16()) and (mask.is_sse_register() and (mask.register.is_virtual() or mask.register == xmm0)):
			# BLENDVPS xmm, xmm/m128: 66 0F 38 14 /r
			# BLENDVPD xmm, xmm/m128: 66 0F 38 15 /r
			# PBLENDVB xmm, xmm/m128: 66 0F 38 10 /r

			self.destination = destination
			self.source = source
			self.mask = mask
			self.size = 5 + source.get_modrm_extra_length()
		else:
			raise ValueError('Invalid operands in instruction {0} {1}, {2}, {3}'.format(name, destination, source, mask))

	def __str__(self):
		return "{0} {1}, {2}, {3}".format(self.name, self.destination, self.source, self.mask)

	def get_input_registers_list(self):
		return self.destination.get_registers_list() + self.source.get_registers_list() + self.mask.get_registers_list()

	def get_output_registers_list(self):
		return [self.destination.register]

	def get_constant(self):
		if self.source.is_constant():
			return self.source.constant
		else:
			return None

	def get_local_variable(self):
		if self.source.is_local_variable():
			return self.source.variable
		else:
			return None

class JmpInstruction(Instruction):
	def __init__(self, destination):
		super(JmpInstruction, self).__init__("JMP")
		if destination.is_label():
			self.destination = destination.label
			#   JA rel32: 0F 87 cd
			#  JAE rel32: 0F 83 cd
			#   JB rel32: 0F 82 cd
			#  JBE rel32: 0F 86 cd
			#   JC rel32: 0F 82 cd
			#   JE rel32: 0F 84 cd
			#   JZ rel32: 0F 84 cd
			self.size = 2 + 4
			self.is_visited = False
		else:
			raise TypeError('Only string names are accepted as destination labels')

	def __str__(self):
		return "{0} .{1}".format(self.name, self.destination)

	def get_input_registers_list(self):
		return list()

	def get_output_registers_list(self):
		return list()

	def get_constant(self):
		return None

	def get_local_variable(self):
		return None

class PushInstruction(Instruction):
	def __init__(self, source):
		super(PushInstruction, self).__init__('PUSH')
		self.source = source
		if source.is_general_purpose_register64():
			self.size = 1 + 1
		else:
			raise ValueError('Invalid operand in instruction PUSH {0}'.format(source))

	def __str__(self):
		return "PUSH {0}".format(self.source)

	def get_input_registers_list(self):
		input_registers_list = [rsp]
		input_registers_list.extend(self.source.get_registers_list())
		return input_registers_list

	def get_output_registers_list(self):
		return [rsp]

	def get_constant(self):
		return None

	def get_local_variable(self):
		return None

class PopInstruction(Instruction):
	def __init__(self, destination):
		super(PopInstruction, self).__init__('POP')
		self.destination = destination
		if destination.is_general_purpose_register64():
			self.size = 1 + 1
		else:
			raise ValueError('Invalid operand in instruction POP {0}'.format(destination))

	def __str__(self):
		return "POP {0}".format(self.destination)

	def get_input_registers_list(self):
		input_registers_list = [rsp]
		if self.destination.is_memory_address():
			input_registers_list.extend(self.destination.get_registers_list())
		return input_registers_list

	def get_output_registers_list(self):
		output_registers_list = [rsp]
		if self.destination.is_register():
			output_registers_list.append(self.destination.register)
		return output_registers_list

	def get_constant(self):
		return None

	def get_local_variable(self):
		return None

class RetInstruction(Instruction):
	def __init__(self, frame_size):
		super(RetInstruction, self).__init__('RET')
		if not frame_size.is_none and not frame_size.is_uint16():
			raise ValueError('Invalid operand in instruction RET {0}'.format(frame_size))
		if frame_size.is_none():
			self.immediate = 0
		else:
			self.immediate = frame_size.immediate
		if self.immediate == 0:
			self.size = 1
		else:
			self.size = 3

	def __str__(self):
		if self.immediate == 0:
			return "RET"
		else:
			return "RET {0}".format(self.immediate)

	def get_input_registers_list(self):
		return [rsp]

	def get_output_registers_list(self):
		return [rsp]

	def get_constant(self):
		return None

	def get_local_variable(self):
		return None

class ReturnInstruction(QuasiInstruction):
	def __init__(self, return_value = None):
		super(ReturnInstruction, self).__init__('RETURN')
		if return_value is None:
			self.return_value = None
		elif return_value.is_int32():
			self.return_value = return_value.immediate
		else:
			raise ValueError('Return value is not a 32-bit integer')

	def to_instruction_list(self):
		instruction_list = list()
		if self.return_value is None:
			pass
		if self.return_value == 0:
			instruction_list.append(ArithmeticInstruction('XOR', Operand(eax), Operand(eax)))
		else:
			instruction_list.append(MovInstruction(Operand(eax), Operand(self.return_value)))
		instruction_list.append(RetInstruction(Operand(None)))
		return instruction_list

	def __str__(self):
		return "RETURN {0}".format(self.return_value)

	def get_input_registers_list(self):
		return [rsp]

	def get_output_registers_list(self):
		return [rsp]

	def get_constant(self):
		return None

	def get_local_variable(self):
		return None

class Int3Instruction(Instruction):
	def __init__(self):
		super(Int3Instruction, self).__init__('INT')
		# INT 3: CC
		self.size = 1

	def __str__(self):
		return "INT 3"

	def get_input_registers_list(self):
		return []

	def get_output_registers_list(self):
		return []

	def get_constant(self):
		return None

	def get_local_variable(self):
		return None

def LABEL(name):
	current_function.add_instruction(Label(name))

def ALIGN(alignment):
	current_function.add_instruction(AlignQuasiInstruction(alignment))

def RETURN(return_value = None):
	current_function.add_instruction(ReturnInstruction(Operand(return_value)))

def INT3():
	current_function.add_instruction(Int3Instruction())

class LOAD:
	@staticmethod
	def CONSTANT(destination, source):
		current_function.add_instruction(LoadConstantPseudoInstruction(Operand(destination), Operand(source)))

	@staticmethod
	def PARAMETER(destination, source):
		current_function.add_instruction(LoadParameterPseudoInstruction(Operand(destination), source))

	@staticmethod
	def ELEMENT(destination, source, ctype):
		if isinstance(ctype, yeppp.codegen.c.Type):
			source = Operand(source)
			destination = Operand(destination)
			if destination.is_register():
				if source.is_memory_address():
					memory_size = ctype.get_size(current_function.abi)
					if destination.is_general_purpose_register64():
						if ctype.is_unsigned_integer() or ctype.is_pointer():
							if memory_size == 8:
								current_function.add_instruction(MovInstruction(destination, source))
							elif memory_size == 4:
								destination = Operand(destination.register.get_dword())
								current_function.add_instruction(MovInstruction(destination, source))
							elif memory_size == 2:
								destination = Operand(destination.register.get_dword())
								source.size = 16
								current_function.add_instruction(MoveExtendInstruction('MOVZX', destination, source))
							elif memory_size == 1:
								destination = Operand(destination.register.get_dword())
								source.size = 8
								current_function.add_instruction(MoveExtendInstruction('MOVZX', destination, source))
							else:
								raise ValueError("Invalid memory operand size {0}".format(memory_size))
						elif ctype.is_signed_integer():
							if memory_size == 8:
								current_function.add_instruction(MovInstruction(destination, source))
							elif memory_size == 4:
								source.size = 32
								current_function.add_instruction(MoveExtendInstruction('MOVSX', destination, source))
							elif memory_size == 2:
								source.size = 16
								current_function.add_instruction(MoveExtendInstruction('MOVSX', destination, source))
							elif memory_size == 1:
								source.size = 8
								current_function.add_instruction(MoveExtendInstruction('MOVSX', destination, source))
							else:
								raise ValueError("Invalid memory operand size {0}".format(memory_size))
						else:
							raise TypeError("Invalid memory operand type")
					elif destination.is_general_purpose_register32():
						if ctype.is_unsigned_integer() or (ctype.is_pointer() and current_function.abi.name == "x32"):
							if memory_size == 4:
								current_function.add_instruction(MovInstruction(destination, source))
							elif memory_size == 2:
								source.size = 16
								current_function.add_instruction(MoveExtendInstruction('MOVZX', destination, source))
							elif memory_size == 1:
								source.size = 8
								current_function.add_instruction(MoveExtendInstruction('MOVZX', destination, source))
							else:
								raise ValueError("Invalid memory operand size {0}".format(memory_size))
						elif ctype.is_signed_integer():
							if memory_size == 4:
								source.size = 32
								current_function.add_instruction(MovInstruction(destination, source))
							elif memory_size == 2:
								source.size = 16
								current_function.add_instruction(MoveExtendInstruction('MOVSX', destination, source))
							elif memory_size == 1:
								source.size = 8
								current_function.add_instruction(MoveExtendInstruction('MOVSX', destination, source))
							else:
								raise ValueError("Invalid memory operand size {0}".format(memory_size))
						else:
							raise TypeError("Invalid memory operand type")
					elif destination.is_general_purpose_register16():
						if ctype.is_unsigned_integer():
							if memory_size == 2:
								current_function.add_instruction(MovInstruction(destination, source))
							elif memory_size == 1:
								source.size = 8
								current_function.add_instruction(MoveExtendInstruction('MOVZX', destination, source))
							else:
								raise ValueError("Invalid memory operand size {0}".format(memory_size))
						elif ctype.is_signed_integer():
							if memory_size == 2:
								current_function.add_instruction(MovInstruction(destination, source))
							elif memory_size == 1:
								source.size = 8
								current_function.add_instruction(MoveExtendInstruction('MOVSX', destination, source))
							else:
								raise ValueError("Invalid memory operand size {0}".format(memory_size))
						else:
							raise TypeError("Invalid memory operand type")
					elif destination.is_general_purpose_register8():
						if ctype.is_unsigned_integer():
							if memory_size == 1:
								current_function.add_instruction(MovInstruction(destination, source))
							else:
								raise ValueError("Invalid memory operand size {0}".format(memory_size))
						elif ctype.is_signed_integer():
							if memory_size == 1:
								current_function.add_instruction(MovInstruction(destination, source))
							else:
								raise ValueError("Invalid memory operand size {0}".format(memory_size))
						else:
							raise TypeError("Invalid memory operand type")
					else:
						raise TypeError("general-purpose register expected")
				else:
					raise TypeError("Source must be a memory operand")
			else:
				raise TypeError("Destination must be a register")
		else:
			raise TypeError("Type must be a C type")

def ADD(destination, source):
	instruction = ArithmeticInstruction('ADD', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def ADC(destination, source):
	instruction = ArithmeticInstruction('ADC', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def SUB(destination, source):
	instruction = ArithmeticInstruction('SUB', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def SBB(destination, source):
	instruction = ArithmeticInstruction('SBB', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def CMP(destination, source):
	instruction = ArithmeticInstruction('CMP', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def AND(destination, source):
	instruction = ArithmeticInstruction('AND', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def OR(destination, source):
	instruction = ArithmeticInstruction('OR', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def XOR(destination, source):
	instruction = ArithmeticInstruction('XOR', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def TEST(destination, source):
	instruction = ArithmeticInstruction('TEST', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def SHL(destination, source):
	instruction = ShiftInstruction('SHL', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def SHR(destination, source):
	instruction = ShiftInstruction('SHR', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def SAR(destination, source):
	instruction = ShiftInstruction('SAR', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def ROL(destination, source):
	instruction = ShiftInstruction('ROL', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def ROR(destination, source):
	instruction = ShiftInstruction('ROR', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def RCL(destination, source):
	instruction = ShiftInstruction('RCL', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def RCR(destination, source):
	instruction = ShiftInstruction('RCR', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def BT(destination, source):
	instruction = BitTestInstruction('BT', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def BTS(destination, source):
	instruction = BitTestInstruction('BTS', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def BTR(destination, source):
	instruction = BitTestInstruction('BTR', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def BTC(destination, source):
	instruction = BitTestInstruction('BTC', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def IMUL(destination = None, source = None, immediate = None):
	instruction = ImulInstruction(Operand(destination), Operand(source), Operand(immediate))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def MUL(source):
	instruction = MulInstruction(Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def RET(immediate = None):
	instruction = RetInstruction(Operand(immediate))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def MOV(destination, source):
	instruction = MovInstruction(Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def MOVZX(destination, source):
	instruction = MoveExtendInstruction('MOVZX', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def MOVSX(destination, source):
	instruction = MoveExtendInstruction('MOVSX', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def PUSH(source):
	instruction = PushInstruction(Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def POP(source):
	instruction = PopInstruction(Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def CMOVA(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVAE', Operand(destination), Operand(source)))

def CMOVAE(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVAE', Operand(destination), Operand(source)))

def CMOVB(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVB', Operand(destination), Operand(source)))

def CMOVBE(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVBE', Operand(destination), Operand(source)))

def CMOVC(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVC', Operand(destination), Operand(source)))

def CMOVE(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVE', Operand(destination), Operand(source)))

def CMOVG(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVG', Operand(destination), Operand(source)))

def CMOVGE(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVGE', Operand(destination), Operand(source)))

def CMOVL(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVL', Operand(destination), Operand(source)))

def CMOVLE(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVLE', Operand(destination), Operand(source)))

def CMOVO(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVO', Operand(destination), Operand(source)))

def CMOVP(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVP', Operand(destination), Operand(source)))

def CMOVPO(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVPO', Operand(destination), Operand(source)))

def CMOVS(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVS', Operand(destination), Operand(source)))

def CMOVS(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVS', Operand(destination), Operand(source)))

def CMOVZ(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVZ', Operand(destination), Operand(source)))

def CMOVNA(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVNA', Operand(destination), Operand(source)))

def CMOVNAE(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVNAE', Operand(destination), Operand(source)))

def CMOVNB(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVNB', Operand(destination), Operand(source)))

def CMOVNBE(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVNBE', Operand(destination), Operand(source)))

def CMOVNC(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVNC', Operand(destination), Operand(source)))

def CMOVNE(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVNE', Operand(destination), Operand(source)))

def CMOVNG(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVNG', Operand(destination), Operand(source)))

def CMOVNGE(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVNGE', Operand(destination), Operand(source)))

def CMOVNL(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVNL', Operand(destination), Operand(source)))

def CMOVNLE(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVNLE', Operand(destination), Operand(source)))

def CMOVNO(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVNO', Operand(destination), Operand(source)))

def CMOVNP(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVNP', Operand(destination), Operand(source)))

def CMOVNPO(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVNPO', Operand(destination), Operand(source)))

def CMOVNS(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVNS', Operand(destination), Operand(source)))

def CMOVNS(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVNS', Operand(destination), Operand(source)))

def CMOVNZ(destination, source):
	current_function.add_instruction(CmovInstruction('CMOVNZ', Operand(destination), Operand(source)))

def SETA(destination):
	current_function.add_instruction(SetInstruction('SETAE', Operand(destination)))

def SETAE(destination):
	current_function.add_instruction(SetInstruction('SETAE', Operand(destination)))

def SETB(destination):
	current_function.add_instruction(SetInstruction('SETB', Operand(destination)))

def SETBE(destination):
	current_function.add_instruction(SetInstruction('SETBE', Operand(destination)))

def SETC(destination):
	current_function.add_instruction(SetInstruction('SETC', Operand(destination)))

def SETE(destination):
	current_function.add_instruction(SetInstruction('SETE', Operand(destination)))

def SETG(destination):
	current_function.add_instruction(SetInstruction('SETG', Operand(destination)))

def SETGE(destination):
	current_function.add_instruction(SetInstruction('SETGE', Operand(destination)))

def SETL(destination):
	current_function.add_instruction(SetInstruction('SETL', Operand(destination)))

def SETLE(destination):
	current_function.add_instruction(SetInstruction('SETLE', Operand(destination)))

def SETO(destination):
	current_function.add_instruction(SetInstruction('SETO', Operand(destination)))

def SETP(destination):
	current_function.add_instruction(SetInstruction('SETP', Operand(destination)))

def SETPO(destination):
	current_function.add_instruction(SetInstruction('SETPO', Operand(destination)))

def SETS(destination):
	current_function.add_instruction(SetInstruction('SETS', Operand(destination)))

def SETS(destination):
	current_function.add_instruction(SetInstruction('SETS', Operand(destination)))

def SETZ(destination):
	current_function.add_instruction(SetInstruction('SETZ', Operand(destination)))

def SETNA(destination):
	current_function.add_instruction(SetInstruction('SETNA', Operand(destination)))

def SETNAE(destination):
	current_function.add_instruction(SetInstruction('SETNAE', Operand(destination)))

def SETNB(destination):
	current_function.add_instruction(SetInstruction('SETNB', Operand(destination)))

def SETNBE(destination):
	current_function.add_instruction(SetInstruction('SETNBE', Operand(destination)))

def SETNC(destination):
	current_function.add_instruction(SetInstruction('SETNC', Operand(destination)))

def SETNE(destination):
	current_function.add_instruction(SetInstruction('SETNE', Operand(destination)))

def SETNG(destination):
	current_function.add_instruction(SetInstruction('SETNG', Operand(destination)))

def SETNGE(destination):
	current_function.add_instruction(SetInstruction('SETNGE', Operand(destination)))

def SETNL(destination):
	current_function.add_instruction(SetInstruction('SETNL', Operand(destination)))

def SETNLE(destination):
	current_function.add_instruction(SetInstruction('SETNLE', Operand(destination)))

def SETNO(destination):
	current_function.add_instruction(SetInstruction('SETNO', Operand(destination)))

def SETNP(destination):
	current_function.add_instruction(SetInstruction('SETNP', Operand(destination)))

def SETNPO(destination):
	current_function.add_instruction(SetInstruction('SETNPO', Operand(destination)))

def SETNS(destination):
	current_function.add_instruction(SetInstruction('SETNS', Operand(destination)))

def SETNS(destination):
	current_function.add_instruction(SetInstruction('SETNS', Operand(destination)))

def SETNZ(destination):
	current_function.add_instruction(SetInstruction('SETNZ', Operand(destination)))

def JA(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JA', Operand(destination)))

def JAE(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JAE', Operand(destination)))

def JB(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JB', Operand(destination)))

def JBE(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JBE', Operand(destination)))

def JC(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JC', Operand(destination)))

def JE(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JE', Operand(destination)))

def JG(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JG', Operand(destination)))

def JGE(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JGE', Operand(destination)))

def JL(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JL', Operand(destination)))

def JLE(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JLE', Operand(destination)))

def JO(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JO', Operand(destination)))

def JP(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JP', Operand(destination)))

def JPO(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JPO', Operand(destination)))

def JS(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JS', Operand(destination)))

def JZ(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JZ', Operand(destination)))

def JNA(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JNA', Operand(destination)))

def JNAE(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JNAE', Operand(destination)))

def JNB(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JNB', Operand(destination)))

def JNBE(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JNBE', Operand(destination)))

def JNC(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JNC', Operand(destination)))

def JNE(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JNE', Operand(destination)))

def JNG(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JNG', Operand(destination)))

def JNGE(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JNGE', Operand(destination)))

def JNL(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JNL', Operand(destination)))

def JNLE(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JNLE', Operand(destination)))

def JNO(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JNO', Operand(destination)))

def JNP(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JNP', Operand(destination)))

def JNPO(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JNPO', Operand(destination)))

def JNS(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JNS', Operand(destination)))

def JNZ(destination):
	current_function.add_instruction(ConditionalJumpInstruction('JNZ', Operand(destination)))

def JMP(destination):
	instruction = JmpInstruction(Operand(destination))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def MOVAPS(destination, source):
	instruction = SseMovInstruction('MOVAPS', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def MOVUPS(destination, source):
	instruction = SseMovInstruction('MOVUPS', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def MOVAPD(destination, source):
	instruction = SseMovInstruction('MOVAPD', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def MOVUPD(destination, source):
	instruction = SseMovInstruction('MOVUPD', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def MOVDQA(destination, source):
	instruction = SseMovInstruction('MOVDQA', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def MOVDQU(destination, source):
	instruction = SseMovInstruction('MOVDQU', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def MOVSS(destination, source):
	instruction = SseScalarFloatingPointMovInstruction('MOVSS', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def MOVSD(destination, source):
	instruction = SseScalarFloatingPointMovInstruction('MOVSD', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def ADDSS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('ADDSS', Operand(destination), Operand(source)))

def ADDPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('ADDPS', Operand(destination), Operand(source)))

def ADDSD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('ADDSD', Operand(destination), Operand(source)))

def ADDPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('ADDPD', Operand(destination), Operand(source)))

def SUBSS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('SUBSS', Operand(destination), Operand(source)))

def SUBPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('SUBPS', Operand(destination), Operand(source)))

def SUBSD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('SUBSD', Operand(destination), Operand(source)))

def SUBPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('SUBPD', Operand(destination), Operand(source)))

def MULSS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('MULSS', Operand(destination), Operand(source)))

def MULPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('MULPS', Operand(destination), Operand(source)))

def MULSD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('MULSD', Operand(destination), Operand(source)))

def MULPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('MULPD', Operand(destination), Operand(source)))

def DIVSS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('DIVSS', Operand(destination), Operand(source)))

def DIVPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('DIVPS', Operand(destination), Operand(source)))

def DIVSD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('DIVSD', Operand(destination), Operand(source)))

def DIVPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('DIVPD', Operand(destination), Operand(source)))

def CMPSS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPSS', Operand(destination), Operand(source)))

def CMPPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPPS', Operand(destination), Operand(source)))

def CMPSD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPSD', Operand(destination), Operand(source)))

def CMPPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPPD', Operand(destination), Operand(source)))

def CMPEQSS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPEQSS', Operand(destination), Operand(source)))

def CMPLTSS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPLTSS', Operand(destination), Operand(source)))

def CMPLESS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPLESS', Operand(destination), Operand(source)))

def CMPUNORDSS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPUNORDSS', Operand(destination), Operand(source)))

def CMPNEQSS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPNEQSS', Operand(destination), Operand(source)))

def CMPNLTSS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPNLTSS', Operand(destination), Operand(source)))

def CMPNLESS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPNLESS', Operand(destination), Operand(source)))

def CMPORDSS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPORDSS', Operand(destination), Operand(source)))

def CMPEQPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPEQPS', Operand(destination), Operand(source)))

def CMPLTPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPLTPS', Operand(destination), Operand(source)))

def CMPLEPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPLEPS', Operand(destination), Operand(source)))

def CMPUNORDPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPUNORDPS', Operand(destination), Operand(source)))

def CMPNEQPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPNEQPS', Operand(destination), Operand(source)))

def CMPNLTPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPNLTPS', Operand(destination), Operand(source)))

def CMPNLEPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPNLEPS', Operand(destination), Operand(source)))

def CMPORDPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPORDPS', Operand(destination), Operand(source)))

def CMPEQSD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPEQSD', Operand(destination), Operand(source)))

def CMPLTSD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPLTSD', Operand(destination), Operand(source)))

def CMPLESD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPLESD', Operand(destination), Operand(source)))

def CMPUNORDSD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPUNORDSD', Operand(destination), Operand(source)))

def CMPNEQSD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPNEQSD', Operand(destination), Operand(source)))

def CMPNLTSD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPNLTSD', Operand(destination), Operand(source)))

def CMPNLESD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPNLESD', Operand(destination), Operand(source)))

def CMPORDSD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPORDSD', Operand(destination), Operand(source)))

def CMPEQPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPEQPD', Operand(destination), Operand(source)))

def CMPLTPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPLTPD', Operand(destination), Operand(source)))

def CMPLEPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPLEPD', Operand(destination), Operand(source)))

def CMPUNORDPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPUNORDPD', Operand(destination), Operand(source)))

def CMPNEQPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPNEQPD', Operand(destination), Operand(source)))

def CMPNLTPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPNLTPD', Operand(destination), Operand(source)))

def CMPNLEPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPNLEPD', Operand(destination), Operand(source)))

def CMPORDPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('CMPORDPD', Operand(destination), Operand(source)))

def MINSS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('MINSS', Operand(destination), Operand(source)))

def MINPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('MINPS', Operand(destination), Operand(source)))

def MINSD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('MINSD', Operand(destination), Operand(source)))

def MINPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('MINPD', Operand(destination), Operand(source)))

def MAXSS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('MAXSS', Operand(destination), Operand(source)))

def MAXPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('MAXPS', Operand(destination), Operand(source)))

def MAXSD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('MAXSD', Operand(destination), Operand(source)))

def MAXPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('MAXPD', Operand(destination), Operand(source)))

def HADDPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('HADDPS', Operand(destination), Operand(source)))

def HADDPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('HADDPD', Operand(destination), Operand(source)))

def HSUBPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('HSUBPS', Operand(destination), Operand(source)))

def HSUBPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('HSUBPD', Operand(destination), Operand(source)))

def ADDSUBPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('ADDSUBPS', Operand(destination), Operand(source)))

def ADDSUBPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('ADDSUBPD', Operand(destination), Operand(source)))

def ANDPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('ANDPS', Operand(destination), Operand(source)))

def ANDPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('ANDPD', Operand(destination), Operand(source)))

def ANDNPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('ANDNPS', Operand(destination), Operand(source)))

def ANDNPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('ANDNPD', Operand(destination), Operand(source)))

def ORPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('ORPS', Operand(destination), Operand(source)))

def ORPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('ORPD', Operand(destination), Operand(source)))

def XORPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('XORPS', Operand(destination), Operand(source)))

def XORPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('XORPD', Operand(destination), Operand(source)))

def UNPCKLPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('UNPCKLPS', Operand(destination), Operand(source)))

def UNPCKLPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('UNPCKLPD', Operand(destination), Operand(source)))

def UNPCKHPS(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('UNPCKHPS', Operand(destination), Operand(source)))

def UNPCKHPD(destination, source):
	current_function.add_instruction(SseFloatingPointBinaryInstruction('UNPCKHPD', Operand(destination), Operand(source)))

def LDMXCSR(source):
	current_function.add_instruction(MxcsrControlInstruction('LDMXCSR', Operand(source)))

def STMXCSR(destination):
	current_function.add_instruction(MxcsrControlInstruction('STMXCSR', Operand(destination)))

def PADDB(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PADDB', Operand(destination), Operand(source)))

def PADDW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PADDW', Operand(destination), Operand(source)))

def PADDD(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PADDD', Operand(destination), Operand(source)))

def PADDQ(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PADDQ', Operand(destination), Operand(source)))

def PSUBB(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PSUBB', Operand(destination), Operand(source)))

def PSUBW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PSUBW', Operand(destination), Operand(source)))

def PSUBD(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PSUBD', Operand(destination), Operand(source)))

def PSUBQ(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PSUBQ', Operand(destination), Operand(source)))

def PADDSB(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PADDSB', Operand(destination), Operand(source)))

def PADDSW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PADDSW', Operand(destination), Operand(source)))

def PADDUSB(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PADDUSB', Operand(destination), Operand(source)))

def PADDUSW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PADDUSW', Operand(destination), Operand(source)))

def PSUBSB(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PSUBSB', Operand(destination), Operand(source)))

def PSUBSW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PSUBSW', Operand(destination), Operand(source)))

def PSUBUSB(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PSUBUSB', Operand(destination), Operand(source)))

def PSUBUSW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PSUBUSW', Operand(destination), Operand(source)))

def PSLLW(destination, source):
	current_function.add_instruction(MmxSseShiftInstruction('PSLLW', Operand(destination), Operand(source)))

def PSLLD(destination, source):
	current_function.add_instruction(MmxSseShiftInstruction('PSLLD', Operand(destination), Operand(source)))

def PSLLQ(destination, source):
	current_function.add_instruction(MmxSseShiftInstruction('PSLLQ', Operand(destination), Operand(source)))

def PSRLW(destination, source):
	current_function.add_instruction(MmxSseShiftInstruction('PSRLW', Operand(destination), Operand(source)))

def PSRLD(destination, source):
	current_function.add_instruction(MmxSseShiftInstruction('PSRLD', Operand(destination), Operand(source)))

def PSRLQ(destination, source):
	current_function.add_instruction(MmxSseShiftInstruction('PSRLQ', Operand(destination), Operand(source)))

def PSRAW(destination, source):
	current_function.add_instruction(MmxSseShiftInstruction('PSRAW', Operand(destination), Operand(source)))

def PSRAD(destination, source):
	current_function.add_instruction(MmxSseShiftInstruction('PSRAD', Operand(destination), Operand(source)))

def PAVGB(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PAVGB', Operand(destination), Operand(source)))

def PSLLDQ(destination, source):
	current_function.add_instruction(SseOctowordShiftInstruction('PSLLDQ', Operand(destination), Operand(source)))

def PSRLDQ(destination, source):
	current_function.add_instruction(SseOctowordShiftInstruction('PSRLDQ', Operand(destination), Operand(source)))

def PAVGW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PAVGW', Operand(destination), Operand(source)))

def PMAXUB(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PMAXUB', Operand(destination), Operand(source)))

def PMAXSW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PMINUB', Operand(destination), Operand(source)))

def PMINUB(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PMINUB', Operand(destination), Operand(source)))

def PMINSW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PMINSW', Operand(destination), Operand(source)))

def PSADBW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PSADBW', Operand(destination), Operand(source)))

def PAND(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PAND', Operand(destination), Operand(source)))

def PANDN(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PANDN', Operand(destination), Operand(source)))

def POR(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('POR', Operand(destination), Operand(source)))

def PXOR(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PXOR', Operand(destination), Operand(source)))

def PCMPEQB(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PCMPEQB', Operand(destination), Operand(source)))

def PCMPEQW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PCMPEQW', Operand(destination), Operand(source)))

def PCMPEQD(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PCMPEQD', Operand(destination), Operand(source)))

def PCMPEQQ(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PCMPEQQ', Operand(destination), Operand(source)))

def PCMPGTB(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PCMPGTB', Operand(destination), Operand(source)))

def PCMPGTW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PCMPGTW', Operand(destination), Operand(source)))

def PCMPGTD(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PCMPGTD', Operand(destination), Operand(source)))

def PCMPGTQ(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PCMPGTQ', Operand(destination), Operand(source)))

def PMULLD(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PMULLD', Operand(destination), Operand(source)))

def PMULDQ(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PMULDQ', Operand(destination), Operand(source)))

def PMULUDQ(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PMULUDQ', Operand(destination), Operand(source)))

def PMULLW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PMULLW', Operand(destination), Operand(source)))

def PMULHW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PMULHW', Operand(destination), Operand(source)))

def PMULHUW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PMULHUW', Operand(destination), Operand(source)))

def PMULHRSW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PMULHRSW', Operand(destination), Operand(source)))

def PMADDWD(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PMADDWD', Operand(destination), Operand(source)))

def PMADDUBSW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PMADDUBSW', Operand(destination), Operand(source)))

def PUNPCKLBW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PUNPCKLBW', Operand(destination), Operand(source)))

def PUNPCKHBW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PUNPCKHBW', Operand(destination), Operand(source)))

def PUNPCKLWD(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PUNPCKLWD', Operand(destination), Operand(source)))

def PUNPCKHWD(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PUNPCKHWD', Operand(destination), Operand(source)))

def PUNPCKLDQ(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PUNPCKLDQ', Operand(destination), Operand(source)))

def PUNPCKHDQ(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PUNPCKHDQ', Operand(destination), Operand(source)))

def PUNPCKLQDQ(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PUNPCKLQDQ', Operand(destination), Operand(source)))

def PUNPCKHQDQ(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PUNPCKHQDQ', Operand(destination), Operand(source)))

def PSHUFW(destination, source, immediate):
	current_function.add_instruction(MmxSseShuffleInstruction('PSHUFW', Operand(destination), Operand(source), Operand(immediate)))

def PSHUFLW(destination, source, immediate):
	current_function.add_instruction(MmxSseShuffleInstruction('PSHUFLW', Operand(destination), Operand(source), Operand(immediate)))

def PSHUFHW(destination, source, immediate):
	current_function.add_instruction(MmxSseShuffleInstruction('PSHUFHW', Operand(destination), Operand(source), Operand(immediate)))

def PSHUFD(destination, source, immediate):
	current_function.add_instruction(MmxSseShuffleInstruction('PSHUFD', Operand(destination), Operand(source), Operand(immediate)))

def PACKSSWB(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PACKSSWB', Operand(destination), Operand(source)))

def PACKSSDW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PACKSSDW', Operand(destination), Operand(source)))

def PACKUSWB(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PACKUSWB', Operand(destination), Operand(source)))

def PACKUSDW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PACKUSDW', Operand(destination), Operand(source)))

def PSIGNB(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PSIGNB', Operand(destination), Operand(source)))

def PSIGNW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PSIGNW', Operand(destination), Operand(source)))

def PSIGND(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PSIGND', Operand(destination), Operand(source)))

def PABSB(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PABSB', Operand(destination), Operand(source)))

def PABSW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PABSW', Operand(destination), Operand(source)))

def PABSD(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PABSD', Operand(destination), Operand(source)))

def PHADDW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PHADDW', Operand(destination), Operand(source)))

def PHADDD(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PHADDD', Operand(destination), Operand(source)))

def PHSUBW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PHSUBW', Operand(destination), Operand(source)))

def PHSUBD(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PHSUBD', Operand(destination), Operand(source)))

def PHADDSW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PHADDSW', Operand(destination), Operand(source)))

def PHSUBSW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PHSUBSW', Operand(destination), Operand(source)))

def PSHUFB(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PSHUFB', Operand(destination), Operand(source)))

def PTEST(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PTEST', Operand(destination), Operand(source)))

def PMAXSB(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PMAXSB', Operand(destination), Operand(source)))

def PMAXSD(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PMAXSD', Operand(destination), Operand(source)))

def PMAXUW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PMAXUW', Operand(destination), Operand(source)))

def PMAXUD(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PMAXUD', Operand(destination), Operand(source)))

def PMINSB(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PMINSB', Operand(destination), Operand(source)))

def PMINSD(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PMINSD', Operand(destination), Operand(source)))

def PMINUW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PMINUW', Operand(destination), Operand(source)))

def PMINUD(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PMINUD', Operand(destination), Operand(source)))

def PHMINPOSUW(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('PHMINPOSUW', Operand(destination), Operand(source)))

def BLENDVPS(destination, source, mask):
	instruction = SseBlendMaskedInstruction('BLENDVPS', Operand(destination), Operand(source), Operand(mask))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def BLENDVPD(destination, source, mask):
	instruction = SseBlendMaskedInstruction('BLENDVPD', Operand(destination), Operand(source), Operand(mask))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def PBLENDVB(destination, source, mask):
	instruction = SseBlendMaskedInstruction('PBLENDVB', Operand(destination), Operand(source), Operand(mask))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def AESIMC(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('AESIMC', Operand(destination), Operand(source)))

def AESENC(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('AESENC', Operand(destination), Operand(source)))

def AESENCLAST(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('AESENCLAST', Operand(destination), Operand(source)))

def AESDEC(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('AESDEC', Operand(destination), Operand(source)))

def AESDECLAST(destination, source):
	current_function.add_instruction(SseIntegerBinaryInstruction('AESDECLAST', Operand(destination), Operand(source)))

def PMOVSXBW(destination, source):
	current_function.add_instruction(SseMovExtendInstruction('PMOVSXBW', Operand(destination), Operand(source)))

def PMOVZXBW(destination, source):
	current_function.add_instruction(SseMovExtendInstruction('PMOVZXBW', Operand(destination), Operand(source)))

def PMOVSXBD(destination, source):
	current_function.add_instruction(SseMovExtendInstruction('PMOVSXBD', Operand(destination), Operand(source)))

def PMOVZXBD(destination, source):
	current_function.add_instruction(SseMovExtendInstruction('PMOVZXBD', Operand(destination), Operand(source)))

def PMOVSXBQ(destination, source):
	current_function.add_instruction(SseMovExtendInstruction('PMOVSXBQ', Operand(destination), Operand(source)))

def PMOVZXBQ(destination, source):
	current_function.add_instruction(SseMovExtendInstruction('PMOVZXBQ', Operand(destination), Operand(source)))

def PMOVSXWD(destination, source):
	current_function.add_instruction(SseMovExtendInstruction('PMOVSXWD', Operand(destination), Operand(source)))

def PMOVZXWD(destination, source):
	current_function.add_instruction(SseMovExtendInstruction('PMOVZXWD', Operand(destination), Operand(source)))

def PMOVSXWQ(destination, source):
	current_function.add_instruction(SseMovExtendInstruction('PMOVSXWQ', Operand(destination), Operand(source)))

def PMOVZXWQ(destination, source):
	current_function.add_instruction(SseMovExtendInstruction('PMOVZXWQ', Operand(destination), Operand(source)))

def PMOVSXDQ(destination, source):
	current_function.add_instruction(SseMovExtendInstruction('PMOVSXDQ', Operand(destination), Operand(source)))

def PMOVZXDQ(destination, source):
	current_function.add_instruction(SseMovExtendInstruction('PMOVZXDQ', Operand(destination), Operand(source)))

def VZEROALL():
	current_function.add_instruction(AvxZeroInstruction('VZEROALL'))

def VZEROUPPER():
	current_function.add_instruction(AvxZeroInstruction('VZEROUPPER'))

def VEXTRACTF128(destination, source, immediate):
	current_function.add_instruction(AvxExtract128Instruction('VEXTRACTF128', Operand(destination), Operand(source), Operand(immediate)))

def VMOVSS(destination, source_x, source_y = None):
	current_function.add_instruction(AvxScalarFloatingPointMovInstruction('VMOVSS', Operand(destination), Operand(source_x), Operand(source_y)))

def VMOVSD(destination, source_x, source_y = None):
	current_function.add_instruction(AvxScalarFloatingPointMovInstruction('VMOVSD', Operand(destination), Operand(source_x), Operand(source_y)))

def VADDSS(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VADDSS', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VADDSS', Operand(destination), Operand(source_x), Operand(source_y)))

def VADDSD(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VADDSD', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VADDSD', Operand(destination), Operand(source_x), Operand(source_y)))

def VSUBSS(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VSUBSS', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VSUBSS', Operand(destination), Operand(source_x), Operand(source_y)))

def VSUBSD(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VSUBSD', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VSUBSD', Operand(destination), Operand(source_x), Operand(source_y)))

def VMAXSS(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VMAXSS', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VMAXSS', Operand(destination), Operand(source_x), Operand(source_y)))

def VMAXSD(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VMAXSD', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VMAXSD', Operand(destination), Operand(source_x), Operand(source_y)))

def VMINSS(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VMINSS', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VMINSS', Operand(destination), Operand(source_x), Operand(source_y)))

def VMINSD(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VMINSD', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VMINSD', Operand(destination), Operand(source_x), Operand(source_y)))

def VMULSS(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VMULSS', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VMULSS', Operand(destination), Operand(source_x), Operand(source_y)))

def VMULSD(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VMULSD', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VMULSD', Operand(destination), Operand(source_x), Operand(source_y)))

def VDIVSS(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VDIVSS', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VDIVSS', Operand(destination), Operand(source_x), Operand(source_y)))

def VDIVSD(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VDIVSD', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointScalarBinaryInstruction('VDIVSD', Operand(destination), Operand(source_x), Operand(source_y)))

def VMOVAPS(destination, source):
	instruction = AvxMovInstruction('VMOVAPS', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def VMOVUPS(destination, source):
	instruction = AvxMovInstruction('VMOVUPS', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def VMOVAPD(destination, source):
	instruction = AvxMovInstruction('VMOVAPD', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def VMOVUPD(destination, source):
	instruction = AvxMovInstruction('VMOVUPD', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def VMOVDQA(destination, source):
	instruction = AvxMovInstruction('VMOVDQA', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def VMOVDQU(destination, source):
	instruction = AvxMovInstruction('VMOVDQU', Operand(destination), Operand(source))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def VADDPS(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VADDPS', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VADDPS', Operand(destination), Operand(source_x), Operand(source_y)))

def VADDPD(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VADDPD', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VADDPD', Operand(destination), Operand(source_x), Operand(source_y)))

def VSUBPS(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VSUBPS', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VSUBPS', Operand(destination), Operand(source_x), Operand(source_y)))

def VSUBPD(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VSUBPD', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VSUBPD', Operand(destination), Operand(source_x), Operand(source_y)))

def VMAXPS(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VMAXPS', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VMAXPS', Operand(destination), Operand(source_x), Operand(source_y)))

def VMAXPD(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VMAXPD', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VMAXPD', Operand(destination), Operand(source_x), Operand(source_y)))

def VMINPS(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VMINPS', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VMINPS', Operand(destination), Operand(source_x), Operand(source_y)))

def VMINPD(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VMINPD', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VMINPD', Operand(destination), Operand(source_x), Operand(source_y)))

def VMULPS(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VMULPS', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VMULPS', Operand(destination), Operand(source_x), Operand(source_y)))

def VMULPD(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VMULPD', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VMULPD', Operand(destination), Operand(source_x), Operand(source_y)))

def VDIVPS(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VDIVPS', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VDIVPS', Operand(destination), Operand(source_x), Operand(source_y)))

def VDIVPD(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VDIVPD', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VDIVPD', Operand(destination), Operand(source_x), Operand(source_y)))

def VANDPS(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VANDPS', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VANDPS', Operand(destination), Operand(source_x), Operand(source_y)))

def VANDPD(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VANDPD', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VANDPD', Operand(destination), Operand(source_x), Operand(source_y)))

def VANDNPS(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VANDNPS', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VANDNPS', Operand(destination), Operand(source_x), Operand(source_y)))

def VANDNPD(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VANDNPD', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VANDNPD', Operand(destination), Operand(source_x), Operand(source_y)))

def VORPS(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VORPS', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VORPS', Operand(destination), Operand(source_x), Operand(source_y)))

def VORPD(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VORPD', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VORPD', Operand(destination), Operand(source_x), Operand(source_y)))

def VXORPS(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VXORPS', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VXORPS', Operand(destination), Operand(source_x), Operand(source_y)))

def VXORPD(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VXORPD', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VXORPD', Operand(destination), Operand(source_x), Operand(source_y)))

def VUNPCKLPS(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VUNPCKLPS', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VUNPCKLPS', Operand(destination), Operand(source_x), Operand(source_y)))

def VUNPCKLPD(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VUNPCKLPD', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VUNPCKLPD', Operand(destination), Operand(source_x), Operand(source_y)))

def VUNPCKHPS(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VUNPCKHPS', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VUNPCKHPS', Operand(destination), Operand(source_x), Operand(source_y)))

def VUNPCKHPD(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VUNPCKHPD', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VUNPCKHPD', Operand(destination), Operand(source_x), Operand(source_y)))

def VHADDPS(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VHADDPS', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VHADDPS', Operand(destination), Operand(source_x), Operand(source_y)))

def VHADDPD(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VHADDPD', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VHADDPD', Operand(destination), Operand(source_x), Operand(source_y)))

def VHSUBPS(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VHSUBPS', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VHSUBPS', Operand(destination), Operand(source_x), Operand(source_y)))

def VHSUBPD(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VHSUBPD', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VHSUBPD', Operand(destination), Operand(source_x), Operand(source_y)))

def VADDSUBPS(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VADDSUBPS', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VADDSUBPS', Operand(destination), Operand(source_x), Operand(source_y)))

def VADDSUBPD(destination, source_x, source_y = None):
	if source_y is None:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VADDSUBPD', Operand(destination), Operand(destination), Operand(source_x)))
	else:
		current_function.add_instruction(AvxFloatingPointVectorBinaryInstruction('VADDSUBPD', Operand(destination), Operand(source_x), Operand(source_y)))

def VFMADDSS(destination, source_x, source_y, source_z):
	instruction = Fma4ScalarInstruction('VFMADDSS', Operand(destination), Operand(source_x), Operand(source_y), Operand(source_z))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def VFMADDSD(destination, source_x, source_y, source_z):
	instruction = Fma4ScalarInstruction('VFMADDSD', Operand(destination), Operand(source_x), Operand(source_y), Operand(source_z))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def VFMSUBSS(destination, source_x, source_y, source_z):
	instruction = Fma4ScalarInstruction('VFMSUBSS', Operand(destination), Operand(source_x), Operand(source_y), Operand(source_z))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def VFMSUBSD(destination, source_x, source_y, source_z):
	instruction = Fma4ScalarInstruction('VFMSUBSD', Operand(destination), Operand(source_x), Operand(source_y), Operand(source_z))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def VFNMADDSS(destination, source_x, source_y, source_z):
	instruction = Fma4ScalarInstruction('VFNMADDSS', Operand(destination), Operand(source_x), Operand(source_y), Operand(source_z))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def VFNMADDSD(destination, source_x, source_y, source_z):
	instruction = Fma4ScalarInstruction('VFNMADDSD', Operand(destination), Operand(source_x), Operand(source_y), Operand(source_z))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def VFNMSUBSS(destination, source_x, source_y, source_z):
	instruction = Fma4ScalarInstruction('VFNMSUBSS', Operand(destination), Operand(source_x), Operand(source_y), Operand(source_z))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def VFNMSUBSD(destination, source_x, source_y, source_z):
	instruction = Fma4ScalarInstruction('VFNMSUBSD', Operand(destination), Operand(source_x), Operand(source_y), Operand(source_z))
	if current_function:
		current_function.add_instruction(instruction)
	return instruction

def VFMADDPS(destination, source_x, source_y, source_z):
	current_function.add_instruction(Fma4VectorInstruction('VFMADDPS', Operand(destination), Operand(source_x), Operand(source_y), Operand(source_z)))

def VFMADDPD(destination, source_x, source_y, source_z):
	current_function.add_instruction(Fma4VectorInstruction('VFMADDPD', Operand(destination), Operand(source_x), Operand(source_y), Operand(source_z)))

def VFMSUBPS(destination, source_x, source_y, source_z):
	current_function.add_instruction(Fma4VectorInstruction('VFMSUBPS', Operand(destination), Operand(source_x), Operand(source_y), Operand(source_z)))

def VFMSUBPD(destination, source_x, source_y, source_z):
	current_function.add_instruction(Fma4VectorInstruction('VFMSUBPD', Operand(destination), Operand(source_x), Operand(source_y), Operand(source_z)))

def VFNMADDPS(destination, source_x, source_y, source_z):
	current_function.add_instruction(Fma4VectorInstruction('VFNMADDPS', Operand(destination), Operand(source_x), Operand(source_y), Operand(source_z)))

def VFNMADDPD(destination, source_x, source_y, source_z):
	current_function.add_instruction(Fma4VectorInstruction('VFNMADDPD', Operand(destination), Operand(source_x), Operand(source_y), Operand(source_z)))

def VFNMSUBPS(destination, source_x, source_y, source_z):
	current_function.add_instruction(Fma4VectorInstruction('VFNMSUBPS', Operand(destination), Operand(source_x), Operand(source_y), Operand(source_z)))

def VFNMSUBPD(destination, source_x, source_y, source_z):
	current_function.add_instruction(Fma4VectorInstruction('VFNMSUBPD', Operand(destination), Operand(source_x), Operand(source_y), Operand(source_z)))

def VFMADDSUBPS(destination, source_x, source_y, source_z):
	current_function.add_instruction(Fma4VectorInstruction('VFMADDSUBPS', Operand(destination), Operand(source_x), Operand(source_y), Operand(source_z)))

def VFMADDSUBPD(destination, source_x, source_y, source_z):
	current_function.add_instruction(Fma4VectorInstruction('VFMADDSUBPD', Operand(destination), Operand(source_x), Operand(source_y), Operand(source_z)))

def VFMSUBADDPS(destination, source_x, source_y, source_z):
	current_function.add_instruction(Fma4VectorInstruction('VFMSUBADDPS', Operand(destination), Operand(source_x), Operand(source_y), Operand(source_z)))

def VFMSUBADDPD(destination, source_x, source_y, source_z):
	current_function.add_instruction(Fma4VectorInstruction('VFMSUBADDPD', Operand(destination), Operand(source_x), Operand(source_y), Operand(source_z)))

if __name__ == '__main__':
	import yeppp.codegen.codegen
	from yeppp.codegen.x64 import *

	abi = yeppp.codegen.c.ABI('x64-ms')
	assembler = Assembler(abi)
	x_argument = yeppp.codegen.c.Parameter("x", yeppp.codegen.c.Type("const double*"))
	y_argument = yeppp.codegen.c.Parameter("y", yeppp.codegen.c.Type("double*"))
	length_argument = yeppp.codegen.c.Parameter("length", yeppp.codegen.c.Type("size_t"))

	assembler.begin_function("yepMath_Exp_V64f_V64f", [x_argument, y_argument, length_argument], "Nehalem")

	signMask = Constant.float64x2(-0.0)
	magicBias = Constant.float64x2('0x1.00000000003FFp+52')
	specialMagicBias = Constant.float64x2('0x1.00000000007FEp+52')
	zeroCutoff = Constant.float64x2('-0x1.74385446D71C3p+9')
	infCutoff = Constant.float64x2('0x1.62E42FEFA39EFp+9')
	log2e = Constant.float64x2('0x1.71547652B82FEp+0')
	ln2_hi = Constant.float64x2('0x1.62E42FEFA3800p-1')
	ln2_lo = Constant.float64x2('0x1.EF35793C76730p-45')

	one = Constant.float64x2(1.0)
	c2 = Constant.float64x2('0x1.000000000000Cp-1')
	c3 = Constant.float64x2('0x1.5555555555522p-3')
	c4 = Constant.float64x2('0x1.555555554FE5Cp-5')
	c5 = Constant.float64x2('0x1.111111111DC0Ap-7')
	c6 = Constant.float64x2('0x1.6C16C185EC9E1p-10')
	c7 = Constant.float64x2('0x1.A01A016020E85p-13')
	c8 = Constant.float64x2('0x1.A019960BCB82Ap-16')
	c9 = Constant.float64x2('0x1.71DEADCE6F1F1p-19')
	c10 = Constant.float64x2('0x1.28B0A7EA5C532p-22')
	c11 = Constant.float64x2('0x1.AE3BB8191C002p-26')

	exponentCutDown = Constant.uint64x2(0x0000000000000010L)
	exponentCutUp   = Constant.uint64x2(0x0000000000007F00L)

	xPointer = GeneralPurposeRegister64()
	yPointer = GeneralPurposeRegister64()
	length = GeneralPurposeRegister64()

	LOAD.PARAMETER( xPointer, x_argument )
	LOAD.PARAMETER( yPointer, y_argument )
	LOAD.PARAMETER( length, length_argument )

	SUB( length, 20 )
	JB( 'restore' )
	ALIGN( 32 )
	LABEL( 'process_by_20' )
	xmm_ln2_lo = SSERegister()
	xmm_ln2_hi = SSERegister()
	xmm_magic_bias = SSERegister()
	xmm_log2e = SSERegister()
	LOAD.CONSTANT( xmm_ln2_lo, ln2_lo )
	LOAD.CONSTANT( xmm_ln2_hi, ln2_hi )
	LOAD.CONSTANT( xmm_magic_bias, magicBias )
	LOAD.CONSTANT( xmm_log2e, log2e )

	x = [SSERegister() for index in range(4)] + [LocalVariable(SSERegister) for index in range(6)]
	factors = [LocalVariable(SSERegister) for index in range(10)]

	for index in range(10):
		if isinstance(x[index], Register):
			xmm_x = x[index]
		else:
			xmm_x = SSERegister()

		xmm_temp = SSERegister()
		xmm_factor = SSERegister()
		xmm_copy = SSERegister()
		x_hi = GeneralPurposeRegister32()

		# Load two elements
		MOVAPS( xmm_x, [xPointer + index * 16] )
		# Duplicate loaded elements
		MOVAPS( xmm_temp, xmm_x )
		# Load the high part of the first element in the SSE register
		MOV( x_hi, [xPointer + index * 16 + 4] )
		# Multiply by log2e
		MULPD( xmm_temp, xmm_log2e )
		# Remove the sign of the high part of the first element in the SSE register
		ADD( x_hi, x_hi )
		# Round to integer
		ADDPD( xmm_temp, xmm_magic_bias )
		# Check the high part of the first element in the SSE register for special case
		CMP( x_hi, 0x810C4BE0 )
		JA( 'special' )
		# Extract the integer factor int(round(x * log2e)) + 0x3FF
		MOVAPS( xmm_factor, xmm_temp )
		# Move the scaling factor to the exponent field of double-precision number
		PSLLQ( xmm_factor, 52 )
		# Subtract the bias. Get round(x * log2e)
		SUBPD( xmm_temp, xmm_magic_bias )
		# Load the high part of the second element in the SSE register
		MOV( x_hi, [xPointer + index * 16 + 12] )
		# Save the scaling factor to a local variable
		MOVAPS( factors[index], xmm_factor )
		# Make a copy of the round(x * log2e) values
		MOVAPS( xmm_copy, xmm_temp )
		# Muliply round(x * log2e) by ln2 (high part)
		MULPD( xmm_temp, xmm_ln2_hi )
		# Remove the sign of the high part of the second element in the SSE register
		ADD( x_hi, x_hi )
		# Compute x - round(x * log2e) * ln2_hi
		SUBPD( xmm_x, xmm_temp )
		# Multiply round(x * log2e) by ln2 (low part)
		MULPD( xmm_copy, xmm_ln2_lo )
		# Check the high part of the first element in the SSE register for special case
		CMP( x_hi, 0x810C4BE0 )
		JA( 'special' )
		# Compute x - round(x * log2e) * ln2_hi - round(x * log2e) * ln2_lo
		SUBPD( xmm_x, xmm_temp )
		if isinstance(x[index], LocalVariable):
			MOVAPS(x[index], xmm_x)

	acc = [SSERegister() for index in range(10)]

	for index in range(10):
		if index % 4 == 0:
			LOAD.CONSTANT( acc[index], c11 )
		else:
			MOVAPS( acc[index], acc[index & -4] )

	def horner_step(coef, acc):
		xmm_c = SSERegister()
		LOAD.CONSTANT( xmm_c, coef )
		for index in range(10):
			if isinstance(x[index], Register):
				xmm_x = x[index]
			else:
				xmm_x = SSERegister()
				MOVAPS( xmm_x, x[index] )

			MULPD( acc[index], xmm_x )
			ADDPD( acc[index], xmm_c )

	horner_step(c10, acc)
	horner_step(c9, acc)
	horner_step(c8, acc)
	horner_step(c7, acc)
	horner_step(c6, acc)
	horner_step(c5, acc)
	horner_step(c4, acc)
	horner_step(c3, acc)
	horner_step(c2, acc)
	horner_step(one, acc)

	for index in range(10):
		xmm_factor = SSERegister()
		MOVAPS( xmm_factor, factors[index] )
		MULPD( acc[index], x[index] )
		MULPD( acc[index], xmm_factor )
		ADDPD( acc[index], xmm_factor )
		MOVUPS( [yPointer + index * 16], xmm_factor )

	ADD( xPointer, 160 )
	ADD( yPointer, 160 )
	SUB( length, 20 )
	JAE( 'process_by_20' )

	LABEL( 'restore' )
	ADD( length, 20 )
	JZ( 'returnOk' )

	LABEL( 'process_by_1' )
	INT3()
	ADD( xPointer, 8 )
	ADD( yPointer, 8 )
	SUB( length, 1 )
	JNZ( 'process_by_1' )

	LABEL( 'returnOk' )
	RETURN(0)

	LABEL( 'special' )
	RETURN(0)

	LOAD.CONSTANT( xmm_ln2_lo, ln2_lo )
	LOAD.CONSTANT( xmm_ln2_hi, ln2_hi )
	LOAD.CONSTANT( xmm_magic_bias, magicBias )
	LOAD.CONSTANT( xmm_log2e, log2e )

	for index in range(10):
		if isinstance(x[index], Register):
			xmm_x = x[index]
		else:
			xmm_x = SSERegister()

		xmm_temp = SSERegister()
		xmm_copy = SSERegister()
		x_hi = GeneralPurposeRegister32()

		# Load two elements
		MOVAPS( xmm_temp, [xPointer + index * 16] )
		# Duplicate loaded elements
		MOVAPS( xmm_x, xmm_temp )
		# Multiply by log2e
		MULPD( xmm_temp, xmm_log2e )
		# Round to integer
		ADDPD( xmm_temp, xmm_magic_bias )
		# Save the integer factor int(round(x * log2e)) + 0x3FF
		MOVAPS( factors[index], xmm_temp )
		# Subtract the bias. Get round(x * log2e)
		SUBPD( xmm_temp, xmm_magic_bias )
		# Make a copy of the round(x * log2e) values
		MOVAPS( xmm_copy, xmm_temp )
		# Muliply round(x * log2e) by ln2 (high part)
		MULPD( xmm_temp, xmm_ln2_hi )
		# Compute x - round(x * log2e) * ln2_hi
		SUBPD( xmm_x, xmm_temp )
		# Multiply round(x * log2e) by ln2 (low part)
		MULPD( xmm_copy, xmm_ln2_lo )
		# Compute x - round(x * log2e) * ln2_hi - round(x * log2e) * ln2_lo
		SUBPD( xmm_x, xmm_temp )
		if isinstance(x[index], LocalVariable):
			MOVAPS(x[index], xmm_x)

	xmm_exponent_cut_down = SSERegister()
	xmm_exponent_cut_up   = SSERegister()
	xmm_one               = SSERegister()

	LOAD.CONSTANT( xmm_exponent_cut_down, exponentCutDown )
	LOAD.CONSTANT( xmm_exponent_cut_up,   exponentCutUp )
	LOAD.CONSTANT( xmm_one,               one )

	factors_extra = [LocalVariable(SSERegister) for index in range(10)]

	for index in range(10):
		xmm_factor_extra = SSERegister()
		xmm_factor   = SSERegister()

		MOVDQA( xmm_factor_extra, factors[index] )
		PSLLQ( xmm_factor_extra, 52 )
		MOVDQA( xmm_factor, xmm_factor_extra )
		PMAXUW( xmm_factor, xmm_exponent_cut_down )
		PMINUW( xmm_factor, xmm_exponent_cut_up   )
		PSUBW( xmm_factor_extra, xmm_factor )
		PADDW( xmm_factor_extra, xmm_one )
		MOVDQA( factors[index], xmm_factor )
		MOVDQA( factors_extra[index], xmm_factor_extra )

	acc = [SSERegister() for index in range(10)]

	for index in range(10):
		if index % 4 == 0:
			LOAD.CONSTANT( acc[index], c11 )
		else:
			MOVAPS( acc[index], acc[index & -4] )

	horner_step(c10, acc)
	horner_step(c9, acc)
	horner_step(c8, acc)
	horner_step(c7, acc)
	horner_step(c6, acc)
	horner_step(c5, acc)
	horner_step(c4, acc)
	horner_step(c3, acc)
	horner_step(c2, acc)
	horner_step(one, acc)

	for index in range(10):
		xmm_factor_extra = SSERegister()
		MOVAPS( xmm_factor_extra, factors_extra[index] )
		MULPD( acc[index], x[index] )
		MULPD( acc[index], xmm_factor_extra )
		ADDPD( acc[index], xmm_factor_extra )
		MOVUPS( [yPointer + index * 16], xmm_factor_extra )

	xmm_zero_cutoff = SSERegister()
	xmm_inf_cutoff  = SSERegister()
	xmm_inf         = SSERegister()

	for index in range(10):
		xmm_factor = SSERegister()
		xmm_zero_mask = SSERegister()

		MOVAPS( xmm_factor, factors[index] )
		MULPD( acc[index], xmm_factor )
		MOVAPS( xmm_zero_mask, [xPointer + index * 16] )
		CMPNLTPD( xmm_zero_mask, xmm_zero_cutoff )
		ANDPD( acc[index], xmm_zero_mask )

		MOVAPS( xmm0, xmm_inf_cutoff )
		CMPLTPD( xmm0, [xPointer + index * 16] )
		BLENDVPD( acc[index], xmm_inf, xmm0 )
		MOVUPS( [yPointer + index * 16], acc[index] )

	ADD( xPointer, 160 )
	ADD( yPointer, 160 )
	SUB( length, 20 )
	JAE( 'process_by_20' )
	JMP( 'restore' )

	function = assembler.end_function()
	print assembler

# if __name__ == '__main__':
# 	import yeppp.codegen.codegen
# 	from yeppp.codegen.x64 import *
#
# 	abi = yeppp.codegen.c.ABI('x64-sysv')
# 	assembler = Assembler(abi)
# 	x_argument = yeppp.codegen.c.Parameter("x", yeppp.codegen.c.Type("const float*"))
# 	y_argument = yeppp.codegen.c.Parameter("y", yeppp.codegen.c.Type("const float*"))
# 	z_argument = yeppp.codegen.c.Parameter("z", yeppp.codegen.c.Type("float*"))
# 	length_argument = yeppp.codegen.c.Parameter("length", yeppp.codegen.c.Type("size_t"))
# 	assembler.begin_function("yepCore_DotProduct_V32f_S32f", [x_argument, y_argument, z_argument, length_argument], "Bulldozer")
#
# 	x_type = x_argument.get_type().get_primitive_type()
# 	y_type = y_argument.get_type().get_primitive_type()
# 	z_type = z_argument.get_type().get_primitive_type()
#
# 	x_size = x_type.get_size(abi)
# 	y_size = y_type.get_size(abi)
# 	z_size = z_type.get_size(abi)
#
# 	xPointer = GeneralPurposeRegister64()
# 	yPointer = GeneralPurposeRegister64()
# 	zPointer = GeneralPurposeRegister64()
# 	length = GeneralPurposeRegister64()
#
# 	xAligned16 = "xA16"
# 	xMisaligned16 = "xM16"
# 	returnOk = "returnOk"
# 	storeResult = "storeResult"
# 	restoreLength = "rstr"
# 	processRemainder = "processRemainder"
#
# 	LOAD.PARAMETER( xPointer, x_argument )
# 	LOAD.PARAMETER( yPointer, y_argument )
# 	LOAD.PARAMETER( zPointer, z_argument )
# 	LOAD.PARAMETER( length, length_argument )
#
# 	accumulators = [SSERegister() for i in range(8)]
# 	temporaries = [SSERegister() for i in range(8)]
#
# 	# Initialize accumulators to zero
# 	for accumulator in accumulators:
# 		VXORPS( accumulator, accumulator )
#
# 	def process_scalar():
# 		VMOVSS( temporaries[0], [xPointer] )
# 		VFMADDSS( accumulators[0], temporaries[0], [yPointer], accumulators[0] )
#
# 	TEST( xPointer, 15)
# 	JZ( xAligned16 )
#
# 	LABEL( xMisaligned16 )
# 	process_scalar()
# 	ADD( xPointer, x_size )
# 	ADD( yPointer, y_size )
# 	SUB( length, 1 )
# 	JZ( storeResult )
#
# 	TEST( xPointer, 15)
# 	JNZ( xMisaligned16 )
#
# 	LABEL( xAligned16 )
# 	SUB( length, 128 / x_size )
# 	JB( restoreLength )
#
# 	for index, accumulator, temporary in zip(range(len(accumulators)), accumulators, temporaries):
# 		VMOVUPS( temporary, [yPointer + index * 16])
# 		VFMADDPS( accumulator, temporary, [xPointer + index * 16], accumulator )
#
# 	SUB( xPointer, -128 )
# 	SUB( yPointer, -128 )
# 	SUB( length, 128 / x_size )
# 	JNZ( xAligned16 )
#
# 	LABEL( restoreLength )
# 	ADD( length, 128 / x_size )
# 	JZ( storeResult )
#
# 	LABEL( processRemainder )
# 	process_scalar()
# 	ADD( xPointer, x_size )
# 	ADD( yPointer, y_size )
# 	SUB( length, 1 )
# 	JNZ( processRemainder )
#
# 	LABEL( storeResult )
# 	for index in range(0, len(accumulators), 2):
# 		VADDPS( accumulators[index], accumulators[index + 1] )
# 	for index in range(0, len(accumulators), 4):
# 		VADDPS( accumulators[index], accumulators[index + 2] )
# 	for index in range(0, len(accumulators), 8):
# 		VADDPS( accumulators[index], accumulators[index + 4] )
# 	VHADDPS( accumulators[0], accumulators[0] )
# 	VHADDPS( accumulators[0], accumulators[0] )
#
# 	VMOVSS( [zPointer], accumulators[0] )
#
# 	LABEL( returnOk )
#
# 	RETURN(0)
#
